package jp.takayuki.swt.widgets;

import java.io.InputStream;

import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.ImageLoader;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Tray;
import org.eclipse.swt.widgets.TrayItem;

/**
 * Aj[VΉTrayItem
 * 
 * @author Takayuki Shoji
 */
public class AnimatedTrayItem {

	/**
	 * LbV
	 */
	private Image[] images;

	/**
	 * ^C~O
	 */
	private int[] delayTime;

	/**
	 * ΏۂTrayItem
	 */
	private TrayItem item;

	/**
	 * DisplayIuWFNg
	 */
	private Display display;

	/**
	 * Aj[V̌JԂ񐔁B0ŖB
	 */
	private int repeatCount;

	/**
	 * ݂̃t[
	 */
	private int currentFrame;

	/**
	 * ݂̎cĐ
	 */
	private int currentRepeatCount;

	/**
	 * RXgN^
	 * @param tray TrayIuWFNg
	 * @param style Style bit(ʏSWT.NONE)
	 */
	public AnimatedTrayItem(Tray tray, int style) {
		item = new TrayItem(tray, style);
		item.addDisposeListener(new DisposeListener() {
			public void widgetDisposed(DisposeEvent e) {
				disposeImages();
			}
		});
		display = tray.getDisplay();
	}

	/**
	 * TrayItem擾܂B
	 * CxgXi[ݒpB
	 * @return TrayItem
	 */
	public TrayItem getItem() {
		return this.item;
	}

	/**
	 * <p>摜ݒ肵܂B</p>
	 * @param is ̓Xg[
	 */
	public void setImage(InputStream is) {

		loadImages(is);

		if (images == null) {
			return;
		}

		// Đ
		currentFrame = 0;
		this.currentRepeatCount = this.repeatCount;

		display.timerExec(0, new DisplayImageAction());
	}

	/**
	 * ACR̕\s܂B
	 */
	private class DisplayImageAction implements Runnable {
		public void run() {
			if (item.isDisposed() || images == null) {
				return;
			}
			item.setImage(images[currentFrame]);

			// Đ̏I
			if (currentFrame + 1 == images.length) {
				currentRepeatCount--;
				if (repeatCount != 0 && currentRepeatCount <= 0) {
					return;
				}
			}
			//
			int delay = delayTime[currentFrame];
			// ̃t[ɐi߂
			currentFrame = (currentFrame + 1) % images.length;

			// ̃t[̕\\
			display.timerExec(delay * 10, this);
		}
	}

	/**
	 * 摜f[^ǂݍ݂܂B
	 * @param is ̓Xg[
	 */
	private void loadImages(InputStream is) {
		disposeImages();

		Image offScreenImage = null;
		Image image = null;
		GC offScreenImageGC = null;
		try {
			ImageLoader loader = new ImageLoader();
			ImageData[] data = loader.load(is);
			if (data.length < 1) {
				return;
			}
			this.repeatCount = loader.repeatCount;
			this.images = new Image[data.length];
			this.delayTime = new int[data.length];

			// GIFȊO
			if (loader.logicalScreenWidth == 0
					|| loader.logicalScreenHeight == 0) {
				this.images[0] = new Image(display, data[0]);
				return;
			}

			// ItXN[C[W̐
			offScreenImage = new Image(display, loader.logicalScreenWidth,
					loader.logicalScreenHeight);

			Color bgColor = display.getSystemColor(SWT.COLOR_WIDGET_BACKGROUND);

			// ItXN[`pGC𐶐
			offScreenImageGC = new GC(offScreenImage);
			// wiF̐ݒE`
			offScreenImageGC.setBackground(bgColor);
			offScreenImageGC.fillRectangle(0, 0, loader.logicalScreenWidth,
					loader.logicalScreenHeight);

			/*
			 * Image̍쐬
			 */

			int imageDataIndex = 0;

			// [_[擾ImageData̔z񂩂ImageData擾
			ImageData imageData = data[imageDataIndex];

			// ImageDataImage𐶐
			image = new Image(display, imageData);
			//System.out.println(":" + image);

			// ImageItXN[ɕ`
			offScreenImageGC.drawImage(image, 0, 0, imageData.width,
					imageData.height, imageData.x, imageData.y,
					imageData.width, imageData.height);

			for (int i = 1; i <= data.length; i++) {

				// ItXN[C[W𕡐
				Image imageForCache = new Image(display,
						loader.logicalScreenWidth, loader.logicalScreenHeight);
				GC imageForCacheGC = new GC(imageForCache);
				imageForCacheGC.drawImage(offScreenImage, 0, 0);
				imageForCacheGC.dispose();

				// LbVɕۊ
				this.images[i - 1] = imageForCache;
				this.delayTime[i - 1] = imageData.delayTime;

				//debug(offScreenImage, 500);
				//debug(this.images[i - 1], 1500);

				if (i == data.length) {
					break;
				}
				// disposalMethod̏
				switch (imageData.disposalMethod) {
				case SWT.DM_FILL_BACKGROUND:
					/* Fill with the background color before drawing. */
					offScreenImageGC.setBackground(bgColor);
					offScreenImageGC.fillRectangle(imageData.x, imageData.y,
							imageData.width, imageData.height);
					break;
				case SWT.DM_FILL_PREVIOUS:
					/* Restore the previous image before drawing. */
					offScreenImageGC.drawImage(image, 0, 0, imageData.width,
							imageData.height, imageData.x, imageData.y,
							imageData.width, imageData.height);
					break;
				}

				// ɕ`悷ImageData擾
				imageData = data[i];

				// ImageDataImage𐶐
				image.dispose();
				image = new Image(display, imageData);
				//System.out.println(":" + image);

				// ImageItXN[ɕ`
				offScreenImageGC.drawImage(image, 0, 0, imageData.width,
						imageData.height, imageData.x, imageData.y,
						imageData.width, imageData.height);

			}
		} catch (SWTException e) {
			e.printStackTrace();

			disposeImages();
		} finally {
			if (image != null && !image.isDisposed()) {
				image.dispose();
			}
			if (offScreenImageGC != null && !offScreenImageGC.isDisposed()) {
				offScreenImageGC.dispose();
			}
			if (offScreenImage != null && !offScreenImage.isDisposed()) {
				offScreenImage.dispose();
			}
		}
	}

	/**
	 * LbVj܂B
	 */
	private void disposeImages() {
		if (images == null) {
			return;
		}
		for (int i = 0; i < images.length; i++) {
			if (images[i] != null && !images[i].isDisposed()) {
				images[i].dispose();
				images[i] = null;
			}
		}
		images = null;
		delayTime = null;
	}
}
