package jp.takayuki.swt.widgets;

import jp.takayuki.swt.util.SWTUtils;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Monitor;
import org.eclipse.swt.widgets.Shell;

/**
 * <p>EoʒmEBhE</p>
 * <pre>
 * final NotifyWindow w = new NotifyWindow(display, parent);
 * w.setAutoHide(true);
 * //w.createShell(display, null);
 * w.getComposite().setSize(100, 100);
 * w.getComposite().addMouseListener(new MouseAdapter() {
 *     public void mouseUp(MouseEvent e) {
 *         w.hide();
 *         Shell parent = w.getParent();
 *         if (parent != null) {
 *             parent.setFocus();
 *         }
 *     }
 * });
 * </pre>
 * 
 * @author Takayuki Shoji
 */
public class NotifyWindow {

	/**
	 * ړ̈ʒuY
	 */
	private int targetY;

	/**
	 * Display
	 */
	private Display display;

	/**
	 * Shell
	 */
	private Shell shell;

	/**
	 * Composite
	 */
	private Composite composite;

	/**
	 * ړsNZ
	 */
	private int resolution = 2;

	/**
	 * ^C}[Ԋu(~b)
	 */
	private int timer = 30;

	/**
	 * ~tO
	 */
	private boolean stopShowing;

	/**
	 * ~tO
	 */
	private boolean stopHiding;

	/**
	 * show()̌ɎIhide()ꍇtrue
	 */
	private boolean autoHide;

	/**
	 * hide()̌ɎIclose()ꍇtrue
	 */
	private boolean autoClose;

	/**
	 * ڕW̌`
	 */
	private Rectangle targetBounds;

	/**
	 * RXgN^
	 * @param display
	 * @param parent
	 */
	public NotifyWindow(Display display, Shell parent) {
		this.display = display;

		if (parent == null) {
			this.shell = new Shell(display, SWT.ON_TOP);
		} else {
			this.shell = new Shell(parent, SWT.ON_TOP);
		}
		this.shell.setVisible(false);
		this.shell.setLayout(new FillLayout());
		this.composite = new Composite(shell, SWT.NONE);
	}

	/**
	 * <p>CompositeԂ܂B</p>
	 * <p>̃\bhcreateShell(Display, Shell)̌ĂяǒɗLłB</p>
	 * @return Composite
	 */
	public Composite getComposite() {
		return this.composite;
	}

	/**
	 * 1ňړ鋗sNZŐݒ肵܂B
	 * @param resolution
	 */
	public void setResolution(int resolution) {
		this.resolution = resolution;
	}

	/**
	 * ړsԂ̊Ԋu~bŐݒ肵܂B
	 * @param timer ~b
	 */
	public void setTimer(int timer) {
		this.timer = timer;
	}

	/**
	 * EBhEړȂ\܂B
	 */
	public void show() {
		checkShell();

		this.stopHiding = true;
		this.stopShowing = false;

		// EBhETCY̌
		this.shell.pack();
		targetBounds = this.shell.getBounds();

		// fBXvC̈ʒu擾
		Monitor[] monitors = display.getMonitors();
		Monitor monitor = SWTUtils.getPointedMonitor(display, monitors);
		Rectangle clientRect = monitor.getClientArea();

		// ʒu
		int initialX = clientRect.width + clientRect.x - targetBounds.width;
		int initialY = clientRect.y + clientRect.height - 1;
		shell.setBounds(initialX, initialY, targetBounds.width, 1);

		// ړڕWʒu
		targetY = clientRect.height + clientRect.y - targetBounds.height;

		shell.setVisible(true);

		// ړ
		display.timerExec(timer, new ShowingAction());
	}

	/**
	 * ړ
	 */
	private class ShowingAction implements Runnable {
		public void run() {
			if (shell.isDisposed()) {
				return;
			}
			Rectangle rect = shell.getBounds();
			rect.height += resolution;
			rect.y -= resolution;
			shell.setBounds(rect);

			// p
			if (rect.y > targetY && !stopShowing) {
				display.timerExec(timer, this);
			} else {
				if (autoHide) {
					display.timerExec(2000, new AutoHidingAction());
				}
			}
		}
	}

	/**
	 * Phide()Ăяo܂B
	 */
	private class AutoHidingAction implements Runnable {
		public void run() {
			hide();
		}
	}

	/**
	 * EBhEB܂B
	 */
	public void hide() {
		checkShell();

		this.stopShowing = true;
		this.stopHiding = false;

		// fBXvC̈ʒu擾
		Rectangle shellRect = shell.getBounds();
		Rectangle rect = SWTUtils.getMonitorClientByCoordinate(shellRect.x,
				shellRect.y, display);

		// ړڕWʒu
		targetY = rect.height + rect.y;

		// ړ
		display.timerExec(timer, new HidingAction());
	}

	/**
	 * ړ
	 */
	private class HidingAction implements Runnable {
		public void run() {
			if (shell.isDisposed()) {
				return;
			}
			Rectangle rect = shell.getBounds();
			rect.y += resolution;
			rect.height -= resolution;
			shell.setBounds(rect);

			// p
			if (rect.y <= targetY && rect.height >= 1 && !stopHiding) {
				display.timerExec(timer, this);
			} else {
				if (autoClose) {
					close();
				}
			}
		}
	}

	/**
	 * ~tOONɂăEBhËړ~܂B
	 */
	public void stopShowing() {
		this.stopShowing = true;
	}

	/**
	 * ~tOONɂăEBhËړ~܂B
	 */
	public void stopHiding() {
		this.stopHiding = true;
	}

	/**
	 * Shellgp\ǂmFAgpłȂꍇ͗O𓊂܂B
	 */
	private void checkShell() {
		if (this.shell == null || this.shell.isDisposed()) {
			throw new IllegalStateException(
					"createShell(Display,Shell) must be called."); //$NON-NLS-1$
		}
	}

	/**
	 * shellj܂B
	 */
	private void disposeShell() {
		if (this.shell == null) {
			return;
		}
		if (!this.shell.isDisposed()) {
			shell.close();
		}
		shell = null;
	}

	/**
	 * EBhE܂B
	 */
	public void close() {
		disposeShell();
	}

	/**
	 * eShellԂ܂B
	 * @return ShellBnull̏ꍇ܂B
	 */
	public Shell getParent() {
		return (Shell) this.composite.getShell().getParent();
	}

	/**
	 * Iɕ邩ǂ擾܂B
	 * @return Iɕꍇtrue
	 */
	public boolean isAutoClose() {
		return autoClose;
	}

	/**
	 * Iɕ邩ǂݒ肵܂B
	 * @param autoClose Iɕꍇtrue
	 */
	public void setAutoClose(boolean autoClose) {
		this.autoClose = autoClose;
	}

	/**
	 * IɉBǂ擾܂B
	 * @return IɉBꍇtrue
	 */
	public boolean isAutoHide() {
		return autoHide;
	}

	/**
	 * IɉBǂݒ肵܂B
	 * @param autoHide IɉBꍇtrue
	 */
	public void setAutoHide(boolean autoHide) {
		this.autoHide = autoHide;
	}

	/**
	 * Tv
	 * @param args
	 */
	public static void main(String[] args) {
		Display display = Display.getDefault();
		final NotifyWindow w = new NotifyWindow(Display.getDefault(), null);
		w.setAutoHide(true);
		w.getComposite().setSize(100, 100);
		w.getComposite().addMouseListener(new MouseAdapter() {
			public void mouseUp(MouseEvent e) {
				w.hide();
				Shell parent = w.getParent();
				if (parent != null) {
					parent.setFocus();
				}
			}
		});

		w.getComposite().setLayout(new FillLayout());
		Label label = new Label(w.getComposite(), SWT.NONE);
		label.setText("example example example\nexample example example" //$NON-NLS-1$
				+ "\nexample example example\nexample example example" //$NON-NLS-1$
				+ "\nexample example example\nexample example example"); //$NON-NLS-1$

		w.show();

		Shell sh = new Shell(display);
		sh.setSize(150, 150);
		sh.open();
		// 
		while (!sh.isDisposed()) {
			if (!display.readAndDispatch()) {
				display.sleep();
			}
		}
		display.dispose();
	}
}
