package jp.takayuki.tools;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import jp.takayuki.logging.Logger;
import petaclone.AppPreferences;
import extcmd.Restart;

/**
 * ċNƍXV̋ʏ
 * 
 * @author Takayuki Shoji
 */
public abstract class AbstractRestartHelper implements RestartHelper {

	/**
	 * logger
	 */
	private static Logger logger = Logger
			.getLogger(AbstractRestartHelper.class);

	/**
	 * ݒ
	 */
	protected final AppPreferences pref;

	/**
	 * ċNƍXVɎgptH_
	 */
	protected final File updateDir;

	/**
	 * restart.log
	 */
	protected final File logFile;

	/**
	 * RXgN^
	 * @param pref ݒ
	 */
	public AbstractRestartHelper(AppPreferences pref) {
		this.pref = pref;
		this.updateDir = pref.getPolicy().getUpdateDirectoryPath();
		this.logFile = new File(updateDir, "restart.log"); //$NON-NLS-1$
		logger.debug("updateDir=" + updateDir + ", logFile=" + logFile); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * I폜t@C쐬ăpXԂ܂B
	 * @return I폜t@C̃pX
	 * @throws IOException 
	 */
	protected String createLockFile() throws IOException {
		if (!Tool.makeDirectory(updateDir)) {
			throw new IOException("Could not create update directory."); //$NON-NLS-1$
		}
		File f = File.createTempFile("restart", ".lock", updateDir); //$NON-NLS-1$ //$NON-NLS-2$
		f.deleteOnExit();
		return f.getAbsolutePath();
	}

	/**
	 * NXt@Co͂܂B
	 * @throws IOException
	 */
	protected void writeRestartClass() throws IOException {
		File dir = new File(updateDir, "extcmd"); //$NON-NLS-1$
		if (!Tool.makeDirectory(dir)) {
			throw new IOException("Failed to make directory: " //$NON-NLS-1$
					+ dir.getAbsolutePath());
		}
		File restartClass = new File(dir, "Restart.class"); //$NON-NLS-1$
		FileOutputStream out = new FileOutputStream(restartClass);
		try {
			Tool.copyFromResource(out, "/extcmd/Restart.class"); //$NON-NLS-1$
		} finally {
			out.close();
		}
	}

	/**
	 * restart.log 폜܂B
	 * @throws IOException 
	 */
	protected void deleteRestartLog() throws IOException {
		if (logFile.exists()) {
			if (logFile.delete()) {
				logger.info("The old restart.log deleted."); //$NON-NLS-1$
			} else {
				throw new IOException("Failed to delete restart.log."); //$NON-NLS-1$
			}
		}
	}

	/**
	 * R}hzɕϊ܂B
	 * @param args R}hƈ̃Xg
	 * @return R}hƈ̔z
	 */
	protected String[] toCommandArray(List args) {
		String[] command = (String[]) args.toArray(new String[args.size()]);

		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < command.length; i++) {
			sb.append(command[i]).append(' ');
		}
		logger.info(sb);

		return command;
	}

	/**
	 * R}hs܂B
	 * @param argList ̃Xg
	 * @param env ϐ
	 * @return JnvZX
	 * @throws IOException
	 */
	protected Process execute(List argList, Map env) throws IOException {
		String[] envp = null;
		if (env != null) {
			/*
			 * WindowsłSWTDLL̂߂TEMP,TMP̈pKv
			 * ̑ɂ퓮̂߂ɊϐKvB
			 */
			envp = new String[env.size()];
			int i = 0;
			Iterator itr = env.entrySet().iterator();
			while (itr.hasNext()) {
				Entry entry = (Entry) itr.next();
				envp[i] = (String) entry.getKey() + '=' + entry.getValue();
				i++;
			}
			if (logger.isDebugEnabled()) {
				Arrays.sort(envp);
				for (int j = 0; j < envp.length; j++) {
					logger.debug("ENV=" + envp[j]); //$NON-NLS-1$
				}
			}
		}
		String[] args = (String[]) argList.toArray(new String[argList.size()]);
		if (logger.isDebugEnabled()) {
			for (int i = 0; i < args.length; i++) {
				logger.debug("args[" + i + "]=" + args[i]); //$NON-NLS-1$ //$NON-NLS-2$
			}
		}
		return Runtime.getRuntime().exec(args, envp, updateDir);
	}

	public void restart() throws IOException {
		PathResolver pathResolver = pref.getPolicy().getPathResolver();
		deleteRestartLog();
		String lockFile = createLockFile();

		List args = new ArrayList();
		Map envMap = new HashMap(System.getenv());

		// 0: st@C(java)ւ̃pX
		args.add(pathResolver.getJavaPath());

		// 1,2: Restart.class̃NXpX
		args.add("-cp"); //$NON-NLS-1$
		args.add(updateDir.getAbsolutePath());

		// 3: mainNX "Restart"
		args.add(Restart.class.getName());

		// Ot@Cւ̃pX
		envMap.put(Restart.LOG, logFile.getAbsolutePath());

		// Ďt@Cւ̃pX
		envMap.put(Restart.WATCH_FILE, lockFile);

		// Restartpetaclone̋NɓnƃtH_̃pX
		envMap.put(Restart.BASE_DIR, Tool.getCurrentDirectory());

		// 4: R}h "restart"
		args.add("restart"); //$NON-NLS-1$

		// ċNR}h
		String[] cmdSet = pathResolver.getExecutable();
		if (cmdSet == null) {
			throw new IOException("Failed to get restart command."); //$NON-NLS-1$
		}
		for (int i = 0; i < cmdSet.length; i++) {
			envMap.put(Restart.RESTART + i, cmdSet[i]);
		}

		// s
		writeRestartClass();
		execute(args, envMap);
	}

	/**
	 * J̏ꍇ͗O𓊂܂B
	 * @throws IOException
	 */
	protected void checkDevelopmentEnvironment() throws IOException {
		File dir = Tool.getProgramDirectory(getClass());
		if (dir == null || dir.getName().equals("bin")) { //$NON-NLS-1$
			throw new IOException("Development mode:" + dir); //$NON-NLS-1$
		}
	}

	/**
	 * ʏ팠̍XVpR}h擾܂B
	 * @param pathResolver
	 * @return XVpR}h
	 */
	protected List getUpdateCommand(PathResolver pathResolver) {
		List args2 = new ArrayList();
		args2.add(pathResolver.getJavaPath());
		args2.add("-cp"); //$NON-NLS-1$
		args2.add(updateDir.getAbsolutePath());
		args2.add(Restart.class.getName());
		args2.add(Restart.CMD_UPDATE);
		return args2;
	}
}
