package jp.takayuki.tools;

import java.io.UnsupportedEncodingException;
import java.net.Inet4Address;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.InterfaceAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * IPAhXɊւ郆[eBeB\bhW
 * @author Takayuki Shoji
 */
public class InetAddressUtils {
	/**
	 * InetSocketAddressɕϊ܂B
	 * IPv4IPv6ɑΉĂ܂B
	 * @param addrPort AhXƃ|[gԍ
	 * @param defaultPort ftHg|[gԍ
	 * @return InetSocketAddressBϊłȂꍇnullB
	 */
	public static InetSocketAddress toInetSocketAddress(String addrPort,
			int defaultPort) {

		// IPv4
		Pattern ptn = Pattern
				.compile("^(\\d+\\.\\d+\\.\\d+\\.\\d+)(:(\\d+))?$"); //$NON-NLS-1$
		Matcher m = ptn.matcher(addrPort);
		if (m.matches()) {
			return getMatchedAddress(defaultPort, m);
		}

		// IPv6
		ptn = Pattern.compile("^([0-9a-f:%]+)|(\\[([0-9a-f:%]+)\\]:(\\d+))$"); //$NON-NLS-1$
		m = ptn.matcher(addrPort);
		if (m.matches()) {
			return getMatchedAddressV6(defaultPort, m);
		}

		// zXg
		ptn = Pattern.compile("^([a-zA-Z0-9\\.\\-]+)(:(\\d+))?$"); //$NON-NLS-1$
		m = ptn.matcher(addrPort);
		if (m.matches()) {
			return getMatchedAddress(defaultPort, m);
		}
		return null;
	}

	/**
	 * InetSocketAddressɕϊ܂B
	 * IPv4IPv6ɑΉĂ܂B
	 * @param addrPort AhXƃ|[gԍ
	 * @param defaultPort ftHg|[gԍ
	 * @param socketAddrs InetSocketAddressSet
	 */
	public static void toInetSocketAddress(String[] addrPort, int defaultPort,
			Set socketAddrs) {

		// IPv4
		Pattern ptn4 = Pattern
				.compile("^(\\d+\\.\\d+\\.\\d+\\.\\d+)(:(\\d+))?$"); //$NON-NLS-1$
		Matcher m4 = ptn4.matcher(""); //$NON-NLS-1$

		// IPv6
		Pattern ptn6 = Pattern
				.compile("^([0-9a-f:%]+)|(\\[([0-9a-f:%]+)\\]:(\\d+))$"); //$NON-NLS-1$
		Matcher m6 = ptn6.matcher(""); //$NON-NLS-1$

		// zXg
		Pattern ptnh = Pattern.compile("^([a-zA-Z0-9\\.\\-]+)(:(\\d+))?$"); //$NON-NLS-1$
		Matcher mh = ptnh.matcher(""); //$NON-NLS-1$

		for (int i = 0; i < addrPort.length; i++) {
			if (addrPort[i] == null || addrPort[i].length() == 0) {
				continue;
			}
			m4.reset(addrPort[i]);
			if (m4.matches()) {
				socketAddrs.add(getMatchedAddress(defaultPort, m4));
				continue;
			}

			m6.reset(addrPort[i]);
			if (m6.matches()) {
				socketAddrs.add(getMatchedAddressV6(defaultPort, m6));
				continue;
			}

			mh.reset(addrPort[i]);
			if (mh.matches()) {
				socketAddrs.add(getMatchedAddress(defaultPort, mh));
				continue;
			}
		}
	}

	/**
	 * MatcherAhXƃ|[gԍ擾܂B
	 * @param defaultPort w肳ĂȂꍇɎg|[gԍ
	 * @param m Matcher
	 * @return InetSocketAddressBȂꍇnullB
	 */
	private static InetSocketAddress getMatchedAddressV6(int defaultPort,
			Matcher m) {
		int portNum;
		String addr = m.group(1);
		if (addr == null) {
			addr = m.group(3);
			String port = m.group(4);
			portNum = getPort(port);
			if (portNum < 0)
				return null;
		} else {
			portNum = defaultPort;
		}
		InetAddress iaddr;
		try {
			iaddr = InetAddress.getByName(addr);
		} catch (UnknownHostException e) {
			return null;
		}
		return new InetSocketAddress(iaddr, portNum);
	}

	/**
	 * MatcherAhXƃ|[gԍ擾܂B
	 * @param defaultPort w肳ĂȂꍇɎg|[gԍ
	 * @param m Matcher
	 * @return InetSocketAddressBȂꍇnullB
	 */
	private static InetSocketAddress getMatchedAddress(int defaultPort,
			Matcher m) {
		int portNum;
		String addr = m.group(1);
		String port = m.group(3);
		if (port != null) {
			portNum = getPort(port);
			if (portNum < 0)
				return null;
		} else {
			portNum = defaultPort;
		}
		InetAddress iaddr;
		try {
			iaddr = InetAddress.getByName(addr);
		} catch (UnknownHostException e) {
			return null;
		}
		return new InetSocketAddress(iaddr, portNum);
	}

	/**
	 * 񂪃|[gԍ̏𖞂ǂA|[gԍԂ܂B
	 * @param port |[gԍ
	 * @return |[gԍł͖ꍇ͕̒l
	 */
	public static int getPort(String port) {
		try {
			int portNum = Integer.parseInt(port);
			if (portNum > 0 && portNum <= 65535) {
				return portNum;
			}
			return -1;

		} catch (NumberFormatException e) {
			return -1;
		}
	}

	/**
	 * [JzXgԂ܂B
	 * 擾łȂꍇnullԂ܂B
	 * zXg͑啶ŕԂ܂B
	 * @return [JzXg
	 */
	public static String getHostName() {
		try {
			InetAddress addr = InetAddress.getLocalHost();
			return addr.getHostName().toUpperCase();
		} catch (UnknownHostException e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * RZ~RɃGXP[v[JzXgԂ܂B
	 * 擾łȂꍇnullԂ܂B
	 * @return GXP[vꂽ[JzXg
	 */
	public static String getHostNameEscaped() {
		String name = getHostName();
		return Tool.escapeColon(name);
	}

	/**
	 * [JAhX(IPAhX)̔̕zԂ܂B
	 * 擾łȂꍇ͒0̔zԂ܂B
	 * @return [JAhX̔z
	 */
	public static String[] getLocalAddressesString() {
		InetAddress[] addrs = getLocalAddresses();
		String[] strs = new String[addrs.length];
		for (int i = 0; i < strs.length; i++) {
			strs[i] = addrs[i].getHostAddress();
		}
		return strs;
	}

	/**
	 * w肳ꂽAhX[JAhXȂtrue<br>
	 * ̃u[hLXg悯
	 * @param addr `FbNΏۂInetAddress
	 * @return [JAhXȂtrue
	 */
	public static boolean isLocalAddress(InetAddress addr) {
		if (addr == null) {
			throw new IllegalArgumentException("addr == null"); //$NON-NLS-1$
		}

		if (addr.isLoopbackAddress()) {
			return true;
		}

		return isLocalAddress(addr, getLocalAddresses());
	}

	/**
	 * w肳ꂽAhXlocalAddressesŎw肳ꂽ[JAhXȂtrueB
	 * ̃\bh̓[JAhX̎擾1ōς܂邽߂ɗpӂĂ܂B
	 * @see #getLocalAddresses()
	 * @param addr AhX
	 * @param localAddresses [JAhX̃LbV
	 * @return [JAhXȂtrue
	 */
	public static boolean isLocalAddressCached(InetAddress addr,
			InetAddress[] localAddresses) {
		if (addr == null) {
			throw new IllegalArgumentException("addr == null"); //$NON-NLS-1$
		}

		if (addr.isLoopbackAddress()) {
			return true;
		}

		return isLocalAddress(addr, localAddresses);
	}

	/**
	 * w肳ꂽAhXlocalAddressesŎw肳ꂽ[JAhXȂtrueB
	 * @param addr AhX
	 * @param localAddresses [JAhX̃LbV
	 * @return [JAhXȂtrue
	 */
	private static boolean isLocalAddress(InetAddress addr,
			InetAddress[] localAddresses) {
		if (localAddresses == null || localAddresses.length == 0) {
			return false;
		}
		for (int i = 0; i < localAddresses.length; i++) {
			if (addr.equals(localAddresses[i])) {
				return true;
			}
		}
		return false;
	}

	/**
	 * [JAhX擾܂B
	 * @param number [JAhX̔ԍ(ԂFꂽ)
	 * @return -1⑶݂Ȃԍ̏ꍇnullԂ
	 */
	public static InetAddress getLocalAddress(int number) {
		if (number < 0) {
			return null;
		}
		InetAddress[] localAddresses = getLocalAddresses();
		if (number >= localAddresses.length) {
			return null;
		}

		return localAddresses[number];
	}

	/**
	 * IPo[WԂB
	 * @param bindAddress AhX
	 * @return IPo[WB4A6AȊO-1B
	 */
	public static int getIpVersion(InetAddress bindAddress) {
		if (bindAddress instanceof Inet4Address) {
			return 4;
		} else if (bindAddress instanceof Inet6Address) {
			return 6;
		}
		return getIpVersion();
	}

	/**
	 * IPo[WԂ܂B
	 * @return IPv44AIPv66Ȃ-1B
	 */
	public static int getIpVersion() {
		int ipVersion = -1;
		try {
			InetAddress local = InetAddress.getLocalHost();
			if (local != null) {
				if (local instanceof Inet6Address) {
					ipVersion = 6;
				} else if (local instanceof Inet4Address) {
					ipVersion = 4;
				}
			}
		} catch (UnknownHostException ignore) {
			//
		}
		return ipVersion;
	}

	/**
	 * IPo[WɓKu[hLXgAhX擾܂B
	 * <p>
	 * IPv6̏ꍇ̓I[m[h}`LXgAhX(ff02::1)gp܂B
	 * IPo[WsȏꍇIPv4Ƃ݂Ȃ܂B
	 * @param socketAddress oCh\Pbg̃[JAhX
	 * @param addressSet u[hLXgAhX̃Zbg
	 * @param limitedBr IPv4 Limited broadcast addressǉ
	 * @param directedBr IPv4 directed broadcast addressǉ
	 * @throws SocketException NetworkInterface擾łȂꍇ
	 */
	public static void addDefaultBroadcastAddress(
			InetSocketAddress socketAddress, Set addressSet, boolean limitedBr,
			boolean directedBr) throws SocketException {

		InetAddress bindAddress = socketAddress.getAddress();
		int port = socketAddress.getPort();

		if (bindAddress instanceof Inet6Address) {
			addressSet.add(new InetSocketAddress("ff02::1", port)); //$NON-NLS-1$
			return;
		}

		if (Tool.javaVersion1_6_orHigher && directedBr) {
			if (bindAddress != null && !isDefaultRoute(bindAddress)) {
				NetworkInterface ni = NetworkInterface
						.getByInetAddress(bindAddress);
				if (ni != null) {
					collectBroadcastAddressFromNI(ni, port, addressSet);
				}

			} else {
				// ׂẴC^[tF[X𒲂ׂ
				Enumeration enu = NetworkInterface.getNetworkInterfaces();
				while (enu.hasMoreElements()) {
					NetworkInterface ni = (NetworkInterface) enu.nextElement();
					collectBroadcastAddressFromNI(ni, port, addressSet);
				}
			}
		}

		if (limitedBr) {
			// IPv4Ƃ݂Ȃ
			addressSet.add(new InetSocketAddress("255.255.255.255", port)); //$NON-NLS-1$
		}
	}

	/**
	 * ftHg[g(0.0.0.0)ǂԂ܂B
	 * @param addr AhX
	 * @return 0.0.0.0̂Ƃtrue
	 */
	static boolean isDefaultRoute(InetAddress addr) {
		byte[] bytes = addr.getAddress();
		return (addr instanceof Inet4Address) && bytes[0] == 0 && bytes[1] == 0
				&& bytes[2] == 0 && bytes[3] == 0;
	}

	/**
	 * lbg[NC^[tF[Xu[hLXgAhXW܂B
	 * @param ni lbg[NC^[tF[X
	 * @param port |[g
	 * @param addressSet u[hLXgAhX̃Zbg
	 * @throws SocketException lbg[NC^[tF[X̏Ԃ擾łȂꍇ
	 */
	private static void collectBroadcastAddressFromNI(NetworkInterface ni,
			int port, Set addressSet) throws SocketException {
		if (!ni.isUp() || ni.isLoopback()) {
			return;
		}
		List iaddrs = ni.getInterfaceAddresses();
		for (int i = 0; i < iaddrs.size(); i++) {
			InterfaceAddress iaddr = (InterfaceAddress) iaddrs.get(i);
			// getBroadcast()IPv4̂ݒlԂ
			InetAddress br = iaddr.getBroadcast();
			if (br != null) {
				addressSet.add(new InetSocketAddress(br, port));
			} else if (iaddr.getAddress() instanceof Inet6Address) {
				addressSet.add(new InetSocketAddress("ff02::1", port)); //$NON-NLS-1$
			}
		}
	}

	/**
	 * String̃oCgz񉻂ɎgZbg
	 */
	private static final Charset iso8859_1 = Charset.forName("ISO-8859-1"); //$NON-NLS-1$

	/**
	 * NetworkInterface̕\擾܂B
	 * @param ni NetworkInterface
	 * @return \
	 */
	public static String getDisplayName(NetworkInterface ni) {
		if (Tool.javaVersion1_6_orHigher) {
			return new String(ni.getDisplayName().getBytes(iso8859_1));
		}
		try {
			return new String(ni.getDisplayName().getBytes(iso8859_1.name()));
		} catch (UnsupportedEncodingException e) {
			return null;//肦Ȃ
		}
	}

	/**
	 * [JAhX(IPAhX)̔zԂ܂B
	 * 擾łȂꍇ͒0̔zԂ܂B
	 * @return [JAhX̔z
	 */
	public static InetAddress[] getLocalAddresses() {
		ArrayList list = new ArrayList();
		try {
			Enumeration enu = NetworkInterface.getNetworkInterfaces();
			if (enu == null) {
				return new InetAddress[0];
			}
			while (enu.hasMoreElements()) {
				NetworkInterface ni = (NetworkInterface) enu.nextElement();
				Enumeration enu2 = ni.getInetAddresses();
				while (enu2.hasMoreElements()) {
					list.add(enu2.nextElement());
				}
			}
			return (InetAddress[]) list.toArray(new InetAddress[list.size()]);
		} catch (SocketException e) {
			return new InetAddress[0];
		}
	}

	/**
	 * IPAhXr܂B
	 * @param a IPAhX
	 * @param b IPAhX
	 * @return a傫ꍇ͐̒lAbꍇ͕̒lAꍇ0Ԃ܂B
	 */
	public static int compareIpAddress(String a, String b) {
		if (a == null || b == null) {
			if (a == null && b != null) {
				return -1;
			}
			if (a != null && b == null) {
				return 1;
			}
			return 0;
		}
		if (a.equals(b)) {
			return 0;
		}
		InetAddress ipa = null;
		try {
			ipa = InetAddress.getByName(a);
		} catch (UnknownHostException ignore) {
			//
		}
		InetAddress ipb = null;
		try {
			ipb = InetAddress.getByName(b);
		} catch (UnknownHostException ignore) {
			if (ipa != null) {
				return 1;
			}
			return 0;
		}
		if (ipa == null) {
			return -1;
		}

		boolean aIsIpv4 = (ipa instanceof Inet4Address);
		boolean bIsIpv4 = (ipb instanceof Inet4Address);
		if (!aIsIpv4 && bIsIpv4) {
			return 1;
		}
		if (aIsIpv4 && !bIsIpv4) {
			return -1;
		}
		byte[] ipaValue = ipa.getAddress();
		byte[] ipbValue = ipb.getAddress();

		// z̒͏ɓ

		for (int i = 0; i < ipaValue.length && i < ipbValue.length; i++) {
			int diff = Tool.unsignedByteToInt(ipaValue[i])
					- Tool.unsignedByteToInt(ipbValue[i]);
			if (diff != 0) {
				return diff;
			}
		}

		return 0;
	}

	/**
	 * IPAhX64bit̐lɕϊ܂B
	 * 12.34.56.670x4e38220c0000
	 * @param addr IPAhX
	 * @return 64bit̐l
	 */
	public static long getAddressValue(InetAddress addr) {
		long result = 0;
		byte[] bytes = addr.getAddress();
		for (int i = 0; i < bytes.length; i++) {
			result ^= ((long) bytes[i]) << ((i % 8) * 8);
		}
		if (bytes.length <= 4) {
			//IPv4
			result <<= 32;
		}
		return result;
	}
}
