package petaclone;

import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Tool;
import jp.takayuki.tools.Version;

/**
 * \tgEFAXV̊mFs܂B
 * 
 * @author Takayuki Shoji
 */
public class CheckForUpdates implements Runnable {

	/**
	 * logger
	 */
	private static final Logger logger = Logger
			.getLogger(CheckForUpdates.class);

	/**
	 * AvP[Vݒ
	 */
	private final AppPreferences pref;

	/**
	 * T[oXi[(bZ[W\p)
	 */
	private ServerListener serverListener;

	/**
	 * EventBroker
	 */
	private EventBroker broker;

	/**
	 * RXgN^
	 * @param pref AvP[Vݒ
	 * @param serverListener T[oXi[(bZ[W\p)
	 * @param broker EventBroker
	 */
	public CheckForUpdates(AppPreferences pref, ServerListener serverListener,
			EventBroker broker) {
		this.pref = pref;
		this.serverListener = serverListener;
		this.broker = broker;
	}

	/**
	 * VXbhŏJn܂B
	 * @param broker
	 */
	public void startThread() {
		MessengerThreadGroup gr = new MessengerThreadGroup("CheckForUpdates", //$NON-NLS-1$
				broker);
		Thread th = new Thread(gr, this, "CheckForUpdates"); //$NON-NLS-1$
		th.setDaemon(true);
		th.start();
	}

	/**
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		logger.debug("CheckForUpdates start"); //$NON-NLS-1$

		try {
			Version localVersion = Version.loadFromResource();
			if (localVersion.getBuildNumberAsInt() == -1) {
				logger.warn("Failed to load version from local resource."); //$NON-NLS-1$
				return;
			}

			Version latestVersion = Version.loadFromInternet(localVersion);
			if (latestVersion == null) {
				logger.warn("Failed to download updates info."); //$NON-NLS-1$
				return;
			}

			String newVersionText = latestVersion.getVersionText();
			String newBuildNumberStr = latestVersion.getBuildNumberStr();
			int newBuildNumber = latestVersion.getBuildNumberAsInt();
			if (newVersionText == null || newBuildNumber == -1) {
				logger.warn("Version info is invalid."); //$NON-NLS-1$
				return;
			}

			pref.setLastCheckTime(System.currentTimeMillis());

			// _E[hς݂̏ꍇAɒʒm
			Properties cachedProp = pref.getVersionProperties();
			Version cachedVersion = new Version(cachedProp);
			File cachedFile = cachedVersion.getFile();
			if (cachedFile != null) {
				latestVersion.setFile(cachedFile);
				if (latestVersion.validateFile(latestVersion.getFile())) {
					// ʒm
					notifyUpdate(latestVersion);
					return;
				}
			}

			// ݂VĖʒm̏ꍇɃ_E[hEʒm
			if (localVersion.getBuildNumberAsInt() < newBuildNumber) {

				int lastBuildNumber = pref.getLastBuildNumberAsInt();
				if (lastBuildNumber < newBuildNumber) {
					// _E[h
					File updateFile = downloadUpdateFile(latestVersion);
					if (updateFile == null) {
						logger.info("Failed to download update file."); //$NON-NLS-1$
						return;
					}
					latestVersion.setFile(updateFile);

					notifyUpdate(latestVersion);

					// [U[ɒʒmAbvf[g
					//sendToUser(newVersionText, newBuildNumberStr);
				} else {
					logger.info("New version is available but already noticed."); //$NON-NLS-1$
				}
			} else {
				logger.info("Current version is same or newer."); //$NON-NLS-1$
			}
			// ʒmς݂Ƃēo^
			pref.setLastBuildNumber(newBuildNumberStr);
			logger.debug("CheckForUpdates end"); //$NON-NLS-1$

		} catch (IOException e) {
			logger.warn("CheckForUpdates failed", e); //$NON-NLS-1$
		} catch (Throwable th) {
			logger.error("CheckForUpdates failed", th); //$NON-NLS-1$
		}
	}

	/**
	 * XVʒm܂B
	 * @param latest o[W
	 */
	private void notifyUpdate(Version latest) {
		// _E[hʒm
		broker.notifyEvent(Events.UpdateDownloaded, latest);
	}

	/**
	 * _E[hs܂B
	 * @param latest o[W
	 * @return _E[hς݃t@C
	 * @throws IOException _E[hs
	 */
	private File downloadUpdateFile(Version latest) throws IOException {

		URL url = latest.getDownloadUrl();
		int type = latest.getDownloadType();
		if (url == null || type == -1) {
			logger.error("Failed to get URL to download update file."); //$NON-NLS-1$
			return null;
		}

		URL archiveURL;
		switch (type) {
		case 0:
			// ڃ_E[h
			archiveURL = url;
			break;
		case 1:
			// Vector
			archiveURL = getArchiveURL(url);
			if (archiveURL == null) {
				logger.info("Failed to find download URL."); //$NON-NLS-1$
				return null;
			}
			break;
		default:
			// _E[hs
			logger.error("Invalid download type:" + type); //$NON-NLS-1$
			return null;
		}
		// ۑtH_
		File dir = pref.getPolicy().getUpdateDirectoryPath();

		// URLt@C𒊏o
		String filename = archiveURL.getPath();
		int index = filename.lastIndexOf('/');
		if (index != -1) {
			filename = filename.substring(index + 1);
		}
		File file = new File(dir, filename);

		// _E[h
		saveFile(archiveURL, file);

		// 
		if (!latest.validateFile(file)) {
			return null;
		}

		return file;
	}

	/**
	 * eXgp
	 * @param args
	 * @throws IOException
	 */
	public static void main(String[] args) throws IOException {

		//URL url = new URL(
		//		"http://petaclone.sakura.ne.jp/filelist/petaclone_w20130606_1.7.7a.zip");
		CheckForUpdates c = new CheckForUpdates(null, null, null);

		URL dlURL = new URL(
				"http://www.vector.co.jp/download/file/winnt/net/fh605556.html"); //$NON-NLS-1$
		URL url = c.getArchiveURL(dlURL);
		System.out.println(url);
		if (url != null) {
			File f = new File(url.getFile());
			System.out.println(f);
			c.saveFile(url, new File(f.getName()));
		}
	}

	/**
	 * A[JCuփ_CNgy[WA[JCuURL𒊏o܂B
	 * 
	 * @param pageURL A[JCuփ_CNgy[WURL
	 * @return A[JCuURL
	 * @throws IOException
	 */
	private URL getArchiveURL(URL pageURL) throws IOException {
		ByteArrayOutputStream out = new ByteArrayOutputStream(1024 * 4);
		if (Tool.download(pageURL, out) != HttpURLConnection.HTTP_OK) {
			return null;
		}
		String content = new String(out.toByteArray(), Tool.defaultCharset());
		// _E[hURL̒o
		Pattern p = Pattern
				.compile(
						"<META\\s+HTTP-EQUIV=\"Refresh\"\\s+CONTENT=\"\\d+;URL=([^\"]+)\"", //$NON-NLS-1$
						Pattern.CASE_INSENSITIVE);
		Matcher m = p.matcher(content);
		if (m.find()) {
			String urlStr = m.group(1);
			return new URL(urlStr);
		}
		return null;

	}

	/**
	 * w肵URLt@C_E[hAw肵pXɕۑ܂B
	 * 
	 * @param fileUrl _E[ht@CURL
	 * @param savePath ۑpX
	 * @throws IOException
	 */
	private void saveFile(URL fileUrl, File savePath) throws IOException {

		FileOutputStream o = null;
		BufferedOutputStream out = null;
		try {
			o = new FileOutputStream(savePath);
			out = new BufferedOutputStream(o);
			int res = Tool.download(fileUrl, out);
			if (res != HttpURLConnection.HTTP_OK) {
				throw new IOException("HTTP response code=" + res); //$NON-NLS-1$
			}

		} finally {
			if (out != null) {
				out.close();
			} else if (o != null) {
				o.close();
			}
		}
	}

	/**
	 * [U[ɒʒm܂B
	 * @param versionText o[W
	 * @param buildNumber rhԍ
	 */
	private void sendToUser(Version newVersion, Version localVersion) {

		Message msg = new Message();
		msg.initialOpened();
		msg.setForegroundBlue();
		msg.computerName = "localhost"; //$NON-NLS-1$
		msg.ipAddress = "127.0.0.1"; //$NON-NLS-1$
		msg.launchId = ""; //$NON-NLS-1$
		msg.osUserName = ""; //$NON-NLS-1$
		msg.setCurrentTime();
		msg.userName = "Petaclone"; //$NON-NLS-1$

		MessageMap mm = MessageMap.getMessageMap();
		msg.message = mm.get(
				"CheckForUpdates.msgBody", new Object[] { //$NON-NLS-1$
				newVersion.getVersionText(), newVersion.getBuildNumberStr(),
						localVersion.getVersionText(),
						localVersion.getBuildNumberStr() });

		serverListener.onMessage(msg);
	}
}
