package petaclone.denki;

import java.io.IOException;
import java.net.BindException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.MulticastSocket;
import java.net.SocketException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.TimerTask;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.InetAddressUtils;

/**
 * ł񂫗\NCAgUDPT[o
 * @author Takayuki Shoji
 */
public class DenkiUDPServer implements Runnable {
	/**
	 * logger
	 */
	private static final Logger logger = Logger.getLogger(DenkiUDPServer.class);

	/** bZ[W̍ő咷 */
	private static final int MAX_MESSAGE_LENGTH = 16 * 5;

	/** messenger */
	private DenkiYohouClient messenger;

	/** tO */
	private boolean isAlive;

	/** \Pbg */
	private DatagramSocket socket = null;

	/**
	 * RXgN^
	 * @param messenger messenger
	 */
	public DenkiUDPServer(DenkiYohouClient messenger) {
		this.messenger = messenger;
	}

	/**
	 * T[oĂ邩𒲂ׂ܂B
	 * @return Ătrue
	 */
	synchronized boolean getAlive() {
		return isAlive;
	}

	/**
	 * T[oI܂B
	 */
	void quit() {
		this.isAlive = false;
		close();
	}

	/**
	 * \Pbg܂B
	 */
	private void close() {
		if (socket != null && !socket.isClosed()) {
			if (socket instanceof MulticastSocket) {
				MulticastSocket sock = (MulticastSocket) socket;
				try {
					sock.leaveGroup(InetAddress.getByName("ff02::1")); //$NON-NLS-1$
				} catch (IOException e) {
					logger
							.error(
									"Exception occurred while leaving multicast group.", e); //$NON-NLS-1$
				}
			}
			socket.close();
		}
	}

	/**
	 * oCh܂B
	 * @param bindAddress oChAhX(null̂ƂAׂẴ[JAhX)
	 * @throws BindException |[ggp܂̓[JAhX݂Ȃ
	 * @throws SocketException
	 * @throws IOException
	 */
	void bind(InetAddress bindAddress) throws BindException, SocketException,
			IOException {
		int v = InetAddressUtils.getIpVersion(bindAddress);

		switch (v) {
		case 6:
			InetSocketAddress addr = new InetSocketAddress(bindAddress,
					messenger.port);
			MulticastSocket s = new MulticastSocket(addr);
			s.joinGroup(InetAddress.getByName("ff02::1")); //$NON-NLS-1$
			this.socket = s;
			break;
		case 4:
		default:
			this.socket = new DatagramSocket(messenger.port, bindAddress);
			break;
		}

		isAlive = true;
		logger.info("bind: addr=" + bindAddress + ", port=" + messenger.port); //$NON-NLS-1$ //$NON-NLS-2$
	}

	public void run() {
		byte[] buffer = new byte[MAX_MESSAGE_LENGTH];
		DatagramPacket packet = new DatagramPacket(buffer, buffer.length);
		try {
			while (isAlive) {
				try {
					// pPbg邩^CAEg܂ŃubN
					socket.receive(packet);
				} catch (IOException e) {
					if (isAlive) {
						logger.error("IO exception occurred " //$NON-NLS-1$
								+ "while receiving or waiting.", //$NON-NLS-1$
								e);
						continue;
					}
					break;
				}

				// ̃u[hLXgɔȂ悤ɂ
				if (InetAddressUtils.isLocalAddress(packet.getAddress())) {
					logger.debug("Packet from local address is blocked."); //$NON-NLS-1$
					continue;
				}

				// pPbg̑M
				InetSocketAddress source = new InetSocketAddress(packet
						.getAddress(), packet.getPort());

				String packetStr = new String(buffer, 0, packet.getLength());
				if (logger.isDebugEnabled())
					logger.debug("Received: source=" + source + ",data=" //$NON-NLS-1$ //$NON-NLS-2$
							+ packetStr);

				DenkiUsage pack;
				pack = new DenkiUsage();
				if (!pack.parsePacket(packetStr)) {
					logger.warn("Invalid packet data." + packetStr); //$NON-NLS-1$
					continue;
				}

				if (!pack.epco.equals(messenger.epco.getId())) {
					// |[gԍ͌Œ
					logger.warn("Invalid electric power company:" + pack.epco); //$NON-NLS-1$
					return;
				}

				// R}hƂɏ
				dispatchCommand(pack, source);
			}
		} finally {
			isAlive = false;
			close();
		}

	}

	/**
	 * R}hƂɏ܂B
	 * @param pack pPbg
	 * @param source M
	 */
	private void dispatchCommand(DenkiUsage pack, InetSocketAddress source) {
		if (pack.command.equals("denki:get")) { //$NON-NLS-1$
			synchronized (messenger.denki) {
				dispatchGet(pack, source);
			}

		} else if (pack.command.equals("denki:tell")) { //$NON-NLS-1$
			synchronized (messenger.denki) {
				dispatchTell(pack, source);
			}

		} else if (pack.command.equals("denki:wait")) { //$NON-NLS-1$
			synchronized (messenger.denki) {
				dispatchWait(pack, source);
			}
		}
	}

	/**
	 * denki:get(ŐVf[^̎擾v)܂B
	 * @param pack pPbg
	 * @param source M
	 */
	private void dispatchGet(DenkiUsage pack, InetSocketAddress source) {
		ReplyTask task = new ReplyTask(source);
		long delay = (long) (5 * Math.random() * 1000);
		messenger.scheduler.schedule(task, delay);
	}

	/**
	 * replypPbgM^XN
	 */
	class ReplyTask extends TimerTask {

		/**
		 * M(ԐM)
		 */
		InetSocketAddress source;

		/**
		 * @param source M
		 */
		public ReplyTask(InetSocketAddress source) {
			this.source = source;
		}

		public void run() {
			synchronized (messenger.denki) {
				try {
					sendReplyPacket();
				} catch (Exception e) {
					logger.warn("Exception occurred while sending reply", e); //$NON-NLS-1$
				}
			}
		}

		/**
		 * tellpPbg𑗐MɑM܂B
		 * @throws IOException o͗O
		 */
		private void sendReplyPacket() throws IOException {
			if (messenger.denki.demand == 0) {
				logger
						.debug("Could not send reply because the data is not gotten yet."); //$NON-NLS-1$
				return;
			}
			String packetStr = messenger.denki.toPacket("denki:tell"); //$NON-NLS-1$
			byte[] data = packetStr.getBytes();
			DatagramPacket packet = new DatagramPacket(data, data.length,
					source);
			if (logger.isInfoEnabled()) {
				logger.info("Sending reply:" + packetStr); //$NON-NLS-1$
			}
			socket.send(packet);
		}
	}

	/**
	 * denki:tell(ŐVf[^擾ʒm)܂B
	 * @param pack pPbg
	 * @param source M
	 */
	private void dispatchTell(DenkiUsage pack, InetSocketAddress source) {
		try {
			SimpleDateFormat sdf = new SimpleDateFormat("yyyy/M/d HH:mm"); //$NON-NLS-1$

			long currentUpdateTime = DenkiYohouClient.getCurrentUpdateTime(
					messenger.epco, Calendar.getInstance(), messenger.period);

			long denkiUpdateTime;
			if (messenger.denki.updateTime != null
					&& messenger.denki.updateTime.length() != 0) {
				Date d = sdf.parse(messenger.denki.updateTime);
				denkiUpdateTime = d.getTime();
			} else {
				denkiUpdateTime = 0;
			}

			Date d = sdf.parse(pack.updateTime);
			long packetUpdateTime = d.getTime();

			if (denkiUpdateTime == 0
					|| (denkiUpdateTime < currentUpdateTime && currentUpdateTime <= packetUpdateTime)) {
				messenger.denki.updateTime = pack.updateTime;
				messenger.denki.peekSupply = pack.peekSupply;
				messenger.denki.time = pack.time;
				messenger.denki.demand = pack.demand;
				messenger.denki.sourceAddress = source;
				if (logger.isInfoEnabled()) {
					logger.info("Update by shared data=" + messenger.denki //$NON-NLS-1$
							+ ", source=" + source); //$NON-NLS-1$
				}
				messenger.updateTrayItem();
			} else {
				if (logger.isInfoEnabled()) {
					logger.info("Ignored because of old data=" //$NON-NLS-1$
							+ messenger.denki + ", source=" + source); //$NON-NLS-1$
				}
			}

		} catch (ParseException e) {
			logger.warn(
					"UpdateTime received by tell packet=" + pack.updateTime, e); //$NON-NLS-1$
			return;
		}
	}

	/**
	 * denki:wait(XV̒ʒm)܂B
	 * @param pack pPbg
	 * @param source M
	 */
	private void dispatchWait(DenkiUsage pack, InetSocketAddress source) {
		messenger.denki.wait = System.currentTimeMillis();
	}

	/**
	 * denki:get𑗐M܂B
	 */
	void sendGet() {
		messenger.scheduler.schedule(new GetTask(), 5 * 1000);
	}

	/**
	 * getpPbgM^XN
	 */
	class GetTask extends TimerTask {

		public void run() {
			try {
				sendGetPacket();
			} catch (IOException e) {
				logger.warn("Exception occurred while sending get.", e); //$NON-NLS-1$
			}
		}

		/**
		 * getpPbg𑗐M܂B
		 * @throws IOException
		 */
		private void sendGetPacket() throws IOException {
			InetSocketAddress[] addrs = getBroadcastAddresses();
			DenkiUsage denki = new DenkiUsage();
			denki.epco = messenger.epco.getId();
			String packStr = denki.toPacket("denki:get"); //$NON-NLS-1$
			byte[] data = packStr.getBytes();
			for (int i = 0; i < addrs.length; i++) {
				DatagramPacket packet = new DatagramPacket(data, data.length,
						addrs[i]);
				if (logger.isInfoEnabled()) {
					logger.info("Sending get:" + packStr); //$NON-NLS-1$
				}
				socket.send(packet);
			}
		}
	}

	/**
	 * denki:tell𑗐M܂B
	 */
	void sendTell() {
		messenger.scheduler.schedule(new TellTask(), 100);
	}

	/**
	 * tellpPbgM^XN
	 */
	class TellTask extends TimerTask {
		public void run() {
			synchronized (messenger.denki) {
				try {
					sendTellPacket();
				} catch (IOException e) {
					logger.warn("Exception occurred while sending tell.", e); //$NON-NLS-1$
				}

			}
		}

		/**
		 * tellpPbg𑗐M܂B
		 * @throws IOException
		 */
		private void sendTellPacket() throws IOException {
			InetSocketAddress[] addrs = getBroadcastAddresses();
			String packStr = messenger.denki.toPacket("denki:tell"); //$NON-NLS-1$
			byte[] data = packStr.getBytes();
			for (int i = 0; i < addrs.length; i++) {
				DatagramPacket packet = new DatagramPacket(data, data.length,
						addrs[i]);
				if (logger.isInfoEnabled()) {
					logger.info("Sending tell:" + packStr); //$NON-NLS-1$
				}
				socket.send(packet);
			}

		}
	}

	/**
	 * denki:tell𑗐M܂B
	 */
	void sendWait() {
		messenger.scheduler.schedule(new WaitTask(), 100);
	}

	/**
	 * tellpPbgM^XN
	 */
	class WaitTask extends TimerTask {
		public void run() {
			try {
				sendWaitPacket();
			} catch (IOException e) {
				logger.warn("Exception occurred while sending wait.", e); //$NON-NLS-1$
			}
		}

		/**
		 * waitpPbg𑗐M܂B
		 * @throws IOException
		 */
		private void sendWaitPacket() throws IOException {
			InetSocketAddress[] addrs = getBroadcastAddresses();
			DenkiUsage denki = new DenkiUsage();
			denki.epco = messenger.epco.getId();
			String packStr = denki.toPacket("denki:wait"); //$NON-NLS-1$
			byte[] data = packStr.getBytes();
			for (int i = 0; i < addrs.length; i++) {
				DatagramPacket packet = new DatagramPacket(data, data.length,
						addrs[i]);
				if (logger.isInfoEnabled()) {
					logger.info("Sending wait:" + packStr); //$NON-NLS-1$
				}
				socket.send(packet);
			}
		}

	}

	/**
	 * u[hLXgp̃AhX̃Xg擾܂B
	 * @return MIPAhXƃ|[gԍ̔z
	 */
	private InetSocketAddress[] getBroadcastAddresses() {
		return messenger.conf.getBroadcastAddresses(messenger.pref,
				(InetSocketAddress) socket.getLocalSocketAddress());
	}

}
