package petaclone.denki;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.BindException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketException;
import java.net.URLConnection;
import java.net.UnknownHostException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.Timer;
import java.util.TimerTask;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.InetAddressUtils;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Os;
import jp.takayuki.tools.PropertiesUtils;
import jp.takayuki.tools.Tool;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.graphics.FontMetrics;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.ToolTip;
import org.eclipse.swt.widgets.Tray;
import org.eclipse.swt.widgets.TrayItem;

import petaclone.AppPreferences;
import petaclone.Attachment;
import petaclone.ControlCreator;
import petaclone.DownloadListener;
import petaclone.LaunchConfiguration;
import petaclone.LaunchException;
import petaclone.Message;
import petaclone.Messenger;
import petaclone.Protocol;
import petaclone.SendingResultListener;
import petaclone.ServerListener;
import petaclone.UniqueMember;
import petaclone.denki.EpcoProfile.Chubu;
import petaclone.denki.EpcoProfile.Kansai;
import petaclone.denki.EpcoProfile.Kyushu;
import petaclone.denki.EpcoProfile.Tohoku;
import petaclone.denki.EpcoProfile.Tokyo;
import petaclone.swt.SWTShutdownDialog;

/**
 * ł񂫗\NCAg
 * 
 * @author Takayuki Shoji
 */
public class DenkiYohouClient implements Messenger, Runnable {

	/**
	 * logger
	 */
	private static final Logger logger = Logger
			.getLogger(DenkiYohouClient.class);

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap
			.getMessageMap("petaclone.denki.denki"); //$NON-NLS-1$

	/** o[̃^Cg */
	private static final String TITLE = mm.get("balloonTitle"); //$NON-NLS-1$

	/** AvP[Vݒ */
	AppPreferences pref;

	/** Nݒ */
	LaunchConfiguration conf;

	///** T[oXi[ */
	//private ServerListener serverListener;

	/** XPW[ */
	Timer scheduler;

	/** dCgp */
	DenkiUsage denki = new DenkiUsage();

	/** UI:Display */
	private Display display;

	/** UI:VXegC */
	private Tray tray;

	/** UI:VXegC̃ACe */
	private TrayItem trayItem;

	/** d͉ */
	EpcoProfile epco;

	/** 擾Ԋu(ݒ蔽fς) */
	long period;

	/** |[gԍ */
	int port;

	/** oChAhX */
	private InetAddress bindAddress;

	/** UDPT[o */
	private DenkiUDPServer udp;

	/** UDPT[oXbh */
	private Thread udpThread;

	/** 2(1ԃf[^)ɃtH[obNĂꍇtrue */
	public boolean secondData;

	/** O̎gp */
	private double lastRate;

	/** JEg_Eɓꍇtrue */
	private boolean countdownStarted;

	/**
	 * vgR
	 */
	private static final Protocol PROTOCOL = new Protocol();
	static {
		PROTOCOL.maxMessageLength = Integer.MAX_VALUE;
		PROTOCOL.maxUserNameLength = Integer.MAX_VALUE;
		MessageMap mmap = MessageMap.getMessageMap();
		PROTOCOL.protocolName = mmap.get("MessengerFacade.denki"); //$NON-NLS-1$
		PROTOCOL.signature = "denki"; //$NON-NLS-1$
	}

	public Protocol getProtocol() {
		return PROTOCOL;
	}

	public Messenger newInstance() {
		return new DenkiYohouClient();
	}

	public void createConfigControl(ControlCreator creator) {
		{
			String[] values = { mm.get("Tokyo"), //$NON-NLS-1$
					mm.get("Tohoku"), //$NON-NLS-1$
					mm.get("Chubu"), //$NON-NLS-1$
					mm.get("Kansai"), //$NON-NLS-1$
					mm.get("Kyushu") }; //$NON-NLS-1$
			String[] propertyValues = {
					"TOKYO", "TOHOKU", "CHUBU", "KANSAI", "KYUSHU" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
			creator.createComboBox("company", mm.get("Company"), values, //$NON-NLS-1$ //$NON-NLS-2$
					propertyValues, "TOKYO", true); //$NON-NLS-1$
		}
		{
			String[] values = { mm.get("5min"), mm.get("10min"), //$NON-NLS-1$ //$NON-NLS-2$
					mm.get("15min"), mm.get("20min"), mm.get("25min"), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
					mm.get("30min"), mm.get("40min"), mm.get("50min"), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
					mm.get("60min") }; //$NON-NLS-1$ 
			String[] propertyValues = { "5", "10", "15", "20", "25", "30", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$
					"40", "50", "60" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			creator.createComboBox(
					"interval", mm.get("interval"), values, propertyValues, //$NON-NLS-1$ //$NON-NLS-2$
					"5", true); //$NON-NLS-1$
			creator.createLabel(mm.get("intervalNote")); //$NON-NLS-1$
		}
		{
			creator.createSeparator();
			creator.createLabel(mm.get("shutdownDesc")); //$NON-NLS-1$
			if (Os.isWindows()) {

				creator.createSpinner("threshold", mm.get("usage"), 99, true, //$NON-NLS-1$ //$NON-NLS-2$
						90, 100, 1);

				String[] values = { mm.get("action.doNothing"), //$NON-NLS-1$
						mm.get("action.hibernate"), //$NON-NLS-1$
						mm.get("action.shutdown") }; //$NON-NLS-1$
				String[] propertyValues = { "0", "1", "2" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
				creator.createComboBox("action", mm.get("action"), values, //$NON-NLS-1$ //$NON-NLS-2$
						propertyValues, "0", true); //$NON-NLS-1$

				creator.createSpinner("countdown", mm.get("windowTime"), //$NON-NLS-1$ //$NON-NLS-2$
						3 * 60, true, 0, 999, 1);

				creator.createLabel(mm.get("hibernationNote.windows")); //$NON-NLS-1$

			} else if (Os.isMacOSX()) {

				creator.createSpinner("threshold", mm.get("usage"), 99, true, //$NON-NLS-1$ //$NON-NLS-2$
						90, 100, 1);

				String[] values = { mm.get("action.doNothing"), //$NON-NLS-1$
						mm.get("action.sleepHibernate") }; //$NON-NLS-1$
				String[] propertyValues = { "0", "3" }; //$NON-NLS-1$ //$NON-NLS-2$
				creator.createComboBox("action", mm.get("action"), values, //$NON-NLS-1$ //$NON-NLS-2$
						propertyValues, "0", true); //$NON-NLS-1$

				creator.createSpinner("countdown", mm.get("windowTime"), //$NON-NLS-1$ //$NON-NLS-2$
						3 * 60, true, 0, 999, 1);

				creator.createLabel(mm.get("hibernationNote.macosx")); //$NON-NLS-1$
			} else {

				creator.createSpinner("threshold", mm.get("usage"), 99, true, //$NON-NLS-1$ //$NON-NLS-2$
						90, 100, 1);

				String[] values = { mm.get("action.doNothing"), //$NON-NLS-1$
						mm.get("action.shutdown2") }; //$NON-NLS-1$
				String[] propertyValues = { "0", "4" }; //$NON-NLS-1$ //$NON-NLS-2$
				creator.createComboBox("action", mm.get("action"), values, //$NON-NLS-1$ //$NON-NLS-2$
						propertyValues, "0", true); //$NON-NLS-1$

				creator.createSpinner("countdown", mm.get("windowTime"), //$NON-NLS-1$ //$NON-NLS-2$
						3 * 60, true, 0, 999, 1);

				creator.createLabel(mm.get("hibernationNote.other")); //$NON-NLS-1$
			}
		}
		{
			creator.createSeparator();
			creator.createLabel(mm.get("portDesc")); //$NON-NLS-1$
		}
	}

	/** action̕ */
	private static final String[] actionString = { mm.get("action.doNothing"), //$NON-NLS-1$
			mm.get("action.hibernate"), mm.get("action.shutdown"), //$NON-NLS-1$ //$NON-NLS-2$
			mm.get("action.sleepHibernate"), mm.get("action.shutdown2") }; //$NON-NLS-1$ //$NON-NLS-2$

	/**
	 * |[gԍ擾܂B
	 * @param conf Nݒ
	 * @param defaultPort ftHg|[gԍ
	 * @return |[gԍ
	 */
	private int getPort(LaunchConfiguration conf, int defaultPort) {
		String portStr = conf.getProperty("port", null); //$NON-NLS-1$
		if (portStr == null) {
			return defaultPort;
		}
		try {
			int value = Integer.parseInt(portStr.trim());
			if (0 < value && value <= 65535) {
				return value;
			}
		} catch (NumberFormatException e) {
			//
		}
		return defaultPort;
	}

	/**
	 * oChAhX擾܂B
	 * @param conf Nݒ
	 * @return \PbgɎw肷ׂInetAddress
	 */
	private static InetAddress getBindAddress(LaunchConfiguration conf) {
		int addrNumber = conf.getBindAddress();
		return InetAddressUtils.getLocalAddress(addrNumber);
	}

	public void begin(ThreadGroup group) throws LaunchException {
		String company = conf.getProperty("company", "TOKYO"); //$NON-NLS-1$ //$NON-NLS-2$
		epco = getCompany(company);
		denki.epco = epco.getId();
		port = getPort(conf, epco.getPort());
		bindAddress = getBindAddress(conf);

		try {
			String periodStr = conf.getProperty("interval", "5"); //$NON-NLS-1$ //$NON-NLS-2$
			period = Math.max(epco.getMinimumInterval(),
					60 * 1000 * Long.parseLong(periodStr));
		} catch (NumberFormatException e) {
			period = epco.getMinimumInterval();
		}

		// UI
		display = Display.getDefault();
		tray = display.getSystemTray();
		if (tray != null) {
			trayItem = new TrayItem(tray, SWT.NONE);
			trayItem.setToolTipText(mm.get("trayToolTip")); //$NON-NLS-1$
			updateTrayItem();
		}

		if (Tool.javaVersion1_5_orHigher) {
			scheduler = new Timer("DenkiYohouClient-" + company, true); //$NON-NLS-1$
		} else {
			scheduler = new Timer(true);
		}

		// 30bȓ
		Calendar now = Calendar.getInstance();
		long delay = getNextDelay(epco, now, period);
		schedule(Math.min(delay, 30 * 1000));

		// UDPT[oJn
		try {
			udp = new DenkiUDPServer(this);
			udp.bind(this.bindAddress);

		} catch (BindException e) {
			throw new LaunchException(mm.get("BindException"), e, conf); //$NON-NLS-1$

		} catch (SocketException e) {
			throw new LaunchException(mm.get("SocketException"), e, conf); //$NON-NLS-1$

		} catch (IOException e) {
			throw new LaunchException(mm.get("IOException"), e, conf); //$NON-NLS-1$
		}
		// UDPT[oXbh
		udpThread = new Thread(group, udp);
		udpThread.setName("DenkiYohou UDP-" + epco.getId()); //$NON-NLS-1$
		udpThread.start();

		// L
		udp.sendGet();
	}

	/**
	 * 擾\񂵂܂B
	 * @param manualDelay x(~b)BvZꍇ-1B
	 */
	private void schedule(long manualDelay) {
		Calendar now = null;
		long delay;
		if (manualDelay < 0) {
			now = Calendar.getInstance();
			delay = getNextDelay(epco, now, period);
			// 1ȓŃ_
			delay += Math.random() * 1 * 60 * 1000;
		} else {
			delay = manualDelay;
		}

		scheduler.schedule(new Task(), delay);

		if (logger.isInfoEnabled()) {
			if (now == null) {
				now = Calendar.getInstance();
			}
			now.add(Calendar.MILLISECOND, (int) delay);
			SimpleDateFormat f = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss.SSS"); //$NON-NLS-1$
			logger.info("Acquisition was scheduled after " + delay //$NON-NLS-1$
					+ " msec (" + f.format(now.getTime()) + ")."); //$NON-NLS-1$ //$NON-NLS-2$
		}
	}

	/**
	 * [0ɐݒ肵܂B
	 * @param time ύXCalendar
	 */
	static void setMidnight(Calendar time) {
		time.set(Calendar.HOUR_OF_DAY, 0);
		time.set(Calendar.MINUTE, 0);
		time.set(Calendar.SECOND, 0);
		time.set(Calendar.MILLISECOND, 0);
	}

	/**
	 * 0̌oߎԂ擾܂B
	 * @param time 
	 * @return 0̌oߎ(~b)
	 */
	private static long getElapsedFromMid(Calendar time) {
		long fromMid = time.get(Calendar.HOUR_OF_DAY) * 60 * 60 * 1000;
		fromMid += time.get(Calendar.MINUTE) * 60 * 1000;
		fromMid += time.get(Calendar.SECOND) * 1000;
		fromMid += time.get(Calendar.MILLISECOND);
		return fromMid;
	}

	/**
	 * ̃^C~O擾܂B
	 * <p>
	 * delaylA
	 * @param epco d͉
	 * @param now 
	 * @param period (~b)
	 * @return ̃^C~O܂ł̃~b
	 */
	private static long getNextDelay(EpcoProfile epco, Calendar now, long period) {

		// x~ԑ
		long specialDelay = epco.getSpecialDelay(now);
		if (specialDelay > 0) {
			return specialDelay;
		}

		// ʏ
		Calendar baseTime = (Calendar) now.clone();
		long fromMid = getElapsedFromMid(baseTime);

		setMidnight(baseTime);

		long quotient = fromMid / period;
		long diff = fromMid % period;
		if (epco.getDelay() > diff) {
			return (baseTime.getTimeInMillis() + (quotient * period) + epco
					.getDelay()) - now.getTimeInMillis();
		}
		return (baseTime.getTimeInMillis() + ((quotient + 1) * period + epco
				.getDelay())) - now.getTimeInMillis();
	}

	/**
	 * w肳ꂽɎ擾łȃf[^̎߂܂B
	 * @param epco d͉
	 * @param now 
	 * @param period (~b)
	 * @return updateTime(~b)
	 */
	static long getCurrentUpdateTime(EpcoProfile epco, Calendar now, long period) {
		Calendar time = (Calendar) now.clone();
		time.add(Calendar.MILLISECOND,
				(int) -(epco.getPreWait() + epco.getDelay()));
		long fromMid = getElapsedFromMid(time);
		long quotient = fromMid / period;
		setMidnight(time);
		return time.getTimeInMillis() + (quotient * period);
	}

	/**
	 * TrayItemXV܂B
	 */
	void updateTrayItem() {
		if (trayItem == null) {
			return;
		}
		display.asyncExec(this);
	}

	public void run() {
		if (trayItem == null || trayItem.isDisposed()) {
			return;
		}

		boolean denkiError;
		double rate;
		String updateTime;
		String denkiTime;
		int denkiDemand;
		int denkiPeekSupply;
		InetSocketAddress sourceAddress;
		synchronized (denki) {
			denkiError = denki.error;
			rate = denki.getUsageRate();
			updateTime = denki.updateTime;
			denkiTime = denki.time;
			denkiDemand = denki.demand;
			denkiPeekSupply = denki.peekSupply;
			sourceAddress = denki.sourceAddress;
		}

		if (denkiError) {
			showNotification(TITLE, mm.get("couldNotDownloadLatest"), true); //$NON-NLS-1$
			return;
		}

		disposeTrayItemImage();
		String rateStr;
		if (rate == 0d) {
			rateStr = "??"; //$NON-NLS-1$
		} else {
			rateStr = DenkiUsage.formatRateShort(rate);
		}

		Color bgColor;
		Color textColor = display.getSystemColor(SWT.COLOR_BLACK);
		if (rate == 0d) {
			Color wb = display.getSystemColor(SWT.COLOR_WIDGET_BACKGROUND);
			bgColor = new Color(display, wb.getRGB());
		} else if (rate < 90d) {
			// ΐF
			bgColor = new Color(display, 0, 255, 0);
		} else if (rate < 95d) {
			// F
			bgColor = new Color(display, 255, 255, 0);
		} else if (rate < 97d) {
			// F
			bgColor = new Color(display, 255, 0xA5, 0);
		} else {
			// 
			bgColor = new Color(display, 255, 0, 0);
			if (rate >= 99d) {
				textColor = display.getSystemColor(SWT.COLOR_WHITE);
			}
		}

		//
		Image image = new Image(display, 16, 16);
		GC gc = new GC(image);
		gc.setBackground(bgColor);
		gc.fillRectangle(0, 0, 16, 16);

		Color lineColor = new Color(display, epco.getColor() >> 16 & 0xFF,
				epco.getColor() >> 8 & 0xFF, epco.getColor() & 0xFF);
		gc.setBackground(lineColor);
		gc.fillRectangle(0, 14, 16, 2);
		lineColor.dispose();

		gc.setForeground(textColor);

		Font sysFont = display.getSystemFont();
		FontData[] fd = sysFont.getFontData();

		Font font = new Font(display, fd[0].getName(), 8, SWT.NORMAL);
		gc.setFont(font);

		// drawString̕iǂȂ̂1`
		int[] widths = new int[rateStr.length()];
		int totalWidth = rateStr.length() - 1;
		for (int w = 0; w < rateStr.length(); w++) {
			char c = rateStr.charAt(w);
			widths[w] = gc.getCharWidth(c);
			totalWidth += widths[w];
		}

		int x = (16 - totalWidth) / 2;
		FontMetrics fm = gc.getFontMetrics();
		int y = (16 - fm.getAscent() - fm.getLeading()) / 2;

		for (int i = 0; i < rateStr.length(); i++) {
			gc.drawString(rateStr.substring(i, i + 1), x, y, true);
			x += widths[i] + 1;
		}

		bgColor.dispose();
		font.dispose();
		gc.dispose();

		trayItem.setImage(image);

		// c[`bv
		if (rate == 0d) {
			trayItem.setToolTipText(mm.get("pleaseWait")); //$NON-NLS-1$
		} else {
			String dataSrc;
			if (sourceAddress != null) {
				dataSrc = mm.get("sourceAddress", new Object[] { //$NON-NLS-1$
						sourceAddress.getAddress().getHostAddress(),
								new Integer(sourceAddress.getPort()) });
			} else {
				dataSrc = ""; //$NON-NLS-1$
			}
			String toolTipText = mm.get("trayToolTip2", //$NON-NLS-1$
					new Object[] { epco.getName(), updateTime, denkiTime,
							DenkiUsage.formatRate(rate),
							new Integer(denkiDemand),
							new Integer(denkiPeekSupply), dataSrc });
			trayItem.setToolTipText(toolTipText);
		}

		// o[
		if (lastRate < 94d && 94 <= rate) {
			// ͒
			showNotification(TITLE, mm.get("exceed94"), false); //$NON-NLS-1$
		} else if (lastRate < 95d && 95 <= rate) {
			// ͒
			showNotification(TITLE, mm.get("exceed95"), false); //$NON-NLS-1$
		} else if (lastRate < 96d && 96 <= rate) {
			// ͒
			showNotification(TITLE, mm.get("exceed96"), true); //$NON-NLS-1$
		} else if (rate >= 97d) {
			showNotification(TITLE, mm.get("exceed97"), true); //$NON-NLS-1$
		} else if (rate >= 99d) {
			showNotification(TITLE, mm.get("exceed99", //$NON-NLS-1$
					new Object[] { DenkiUsage.formatRate(rate) }), true);
		} else if (rate >= 100d) {
			showNotification(TITLE, mm.get("exceed100"), true); //$NON-NLS-1$
		}

		// Vbg_E
		if (rate < 88d) {
			countdownStarted = false;
		} else if (90d <= rate && !countdownStarted) {
			int action = PropertiesUtils.getInt(conf.getProperties(), "action", //$NON-NLS-1$
					0);
			if (action != 0) {
				int threshold = PropertiesUtils.getInt(conf.getProperties(),
						"threshold", 0); //$NON-NLS-1$
				if (lastRate < threshold && threshold <= rate) {
					countdownStarted = true;
					int countdown = PropertiesUtils.getInt(
							conf.getProperties(), "countdown", 180); //$NON-NLS-1$
					SWTShutdownDialog dialog = new SWTShutdownDialog(action);
					dialog.setText(mm.get("trayToolTip")); //$NON-NLS-1$
					String shutdownMessage = mm.get("shutdownMessage", //$NON-NLS-1$
							new Object[] { new Integer(threshold),
									actionString[action] });
					dialog.setMessage(shutdownMessage);
					dialog.open(System.currentTimeMillis() + 1000 * countdown);
				}
			}
		}
		lastRate = rate;
	}

	/**
	 * [U[ɒʒm܂B
	 * @param title ^Cg
	 * @param msg bZ[W
	 * @param warn xACRgꍇtrue
	 */
	private void showNotification(String title, String msg, boolean warn) {
		Shell shell = display.getShells()[0];
		int style = warn ? SWT.ICON_WARNING : SWT.ICON_INFORMATION;
		ToolTip tip = new ToolTip(shell, SWT.BALLOON | style);
		tip.setText(title);
		tip.setMessage(msg);
		trayItem.setToolTip(tip);
		tip.setVisible(true);
	}

	public synchronized void quit() {
		if (scheduler != null) {
			scheduler.cancel();
			scheduler = null;
		}

		// udpT[o~
		udp.quit();

		// UI
		disposeTrayItemImage();
		if (trayItem != null) {
			trayItem.dispose();
			trayItem = null;
		}
	}

	/**
	 * TrayItemImagej܂B
	 */
	private void disposeTrayItemImage() {
		if (trayItem != null && !trayItem.isDisposed()) {
			Image img = trayItem.getImage();
			if (img != null && !img.isDisposed()) {
				img.dispose();
			}
		}
	}

	public boolean getAlive() {
		if (udp == null)
			return false;
		return udp.getAlive();
	}

	public String getId() {
		if (conf == null) {
			return null;
		}
		return conf.getId();
	}

	public int getMessageLength(String message) {
		return 0;
	}

	public void setAppPreferences(AppPreferences pref) {
		this.pref = pref;
	}

	public void setConfiguration(LaunchConfiguration conf) {
		this.conf = conf;
	}

	public void setServerListener(ServerListener listener) {
		//this.serverListener = listener;
	}

	/**
	 * @param company d͉ID
	 * @return d͉
	 */
	private EpcoProfile getCompany(String company) {
		if (company.equals("TOKYO")) { //$NON-NLS-1$
			return new Tokyo();
		} else if (company.equals("TOHOKU")) { //$NON-NLS-1$
			return new Tohoku();
		} else if (company.equals("CHUBU")) { //$NON-NLS-1$
			return new Chubu();
		} else if (company.equals("KANSAI")) { //$NON-NLS-1$
			return new Kansai();
		} else if (company.equals("KYUSHU")) { //$NON-NLS-1$
			return new Kyushu();
		} else {
			return new Tokyo();
		}
	}

	/**
	 * TEPCÕT[ouł񂫗\v擾ĉ͂܂B
	 */
	class Task extends TimerTask {

		public void run() {
			logger.debug("acquisition: start"); //$NON-NLS-1$

			boolean retry = false;
			try {
				// Mς݃f[^̊mF
				synchronized (denki) {
					// updateTimẽ`FbN
					SimpleDateFormat sdf = new SimpleDateFormat(
							"yyyy/M/d HH:mm"); //$NON-NLS-1$
					if (denki.updateTime != null
							&& denki.updateTime.length() != 0) {

						// 1ԃf[^ւ̑Ή
						if (secondData) {
							// TODO ̂Ƃ1ԂȂ̂Őݒ͖
							DenkiYohouClient.this.period = epco
									.get2ndMinimumInterval();
						}

						long curUpdateTime = getCurrentUpdateTime(epco,
								Calendar.getInstance(),
								DenkiYohouClient.this.period);
						try {
							Date d = sdf.parse(denki.updateTime);
							if (d.getTime() >= curUpdateTime) {
								if (logger.isInfoEnabled()) {
									logger.info("Current data is the latest. " //$NON-NLS-1$
											+ "updateTime=" + denki.updateTime); //$NON-NLS-1$
								}
								return;
							}
						} catch (ParseException e) {
							throw Tool.newIOException("updateTime=" //$NON-NLS-1$
									+ denki.updateTime, e);
						}
					}

					// wait̃`FbN
					if (denki.wait + epco.getRetryDelay() > System
							.currentTimeMillis()) {
						if (logger.isInfoEnabled()) {
							Date d = new Date(denki.wait + epco.getRetryDelay());
							logger.info("Wait by wait packet:" + sdf.format(d)); //$NON-NLS-1$
						}
						retry = true;
						return;
					}
				}

				// 擾
				URLConnection conn = epco.getURL().openConnection();
				InputStreamReader r = null;
				BufferedReader bin = null;
				try {
					r = new InputStreamReader(conn.getInputStream());
					bin = new BufferedReader(r);

					synchronized (denki) {
						retry = parse(bin);
					}

					// ʒm
					updateTrayItem();

				} finally {
					if (bin != null)
						bin.close();
				}

			} catch (Exception e) {
				logger.warn("Could not download the usage report.", e); //$NON-NLS-1$
				try {
					denki.error = true;
					updateTrayItem();
				} catch (Exception e2) {
					logger.error(e2);
				}
			} finally {
				logger.debug("acquisition: end"); //$NON-NLS-1$

				// \
				if (retry) {
					schedule(epco.getRetryDelay());
				} else {
					schedule(-1);
				}
			}
		}

		/**
		 * d͂̂ł񂫗\CSV͂܂B
		 * @param in 
		 * @return gCKvȏꍇtrue
		 * @throws IOException ͗O
		 */
		public boolean parse(BufferedReader in) throws IOException {
			denki.error = true;

			boolean sendTell = false;
			secondData = false;

			ArrayList lines = new ArrayList();
			String line;
			while ((line = in.readLine()) != null) {
				lines.add(line);
			}
			int startLine = epco.getMinDataLine();
			int endLine = epco.getMaxDataLine();
			if (lines.size() < endLine) {
				if (epco.get2ndMinDataLine() < 0) {
					throw new IOException("Needed more lines. required lines=" //$NON-NLS-1$
							+ epco.getMaxDataLine() + ", download lines=" //$NON-NLS-1$
							+ lines.size());
				}
				// 1ԃf[^ւ̑Ή
				startLine = epco.get2ndMinDataLine();
				endLine = epco.get2ndMaxDataLine();
				secondData = true;

				if (lines.size() < endLine) {
					throw new IOException("Needed more lines. required lines=" //$NON-NLS-1$
							+ epco.get2ndMaxDataLine() + ", download lines=" //$NON-NLS-1$
							+ lines.size());
				}
			}
			// XV̎擾 2011/6/27 23:55 UPDATE
			line = (String) lines.get(0);
			int lastsp = line.lastIndexOf(' ');
			String updateTime;
			if (lastsp != -1) {
				updateTime = line.substring(0, lastsp);
			} else {
				updateTime = line;
			}
			if (denki.updateTime != null && denki.updateTime.length() != 0) {
				// NX^m[hւ̃t@C]sĂ邱Ƃ
				// rČÂLZ
				try {
					SimpleDateFormat f = new SimpleDateFormat("yyyy/M/dd HH:mm"); //$NON-NLS-1$
					Date newDate = f.parse(updateTime);
					Date oldDate = f.parse(denki.updateTime);
					if (oldDate.after(newDate)) {
						logger.warn("Mismatch updateTime. previous time=" //$NON-NLS-1$
								+ denki.updateTime + ", this time=" //$NON-NLS-1$
								+ updateTime);
						return true;
					} else if (oldDate.equals(newDate)) {
						if (logger.isInfoEnabled()) {
							logger.info("UpdateTime is same as previous. previous=" //$NON-NLS-1$
									+ denki.updateTime);
						}
						denki.error = false;
						// XVʒm
						udp.sendWait();
						return true;
					} else {
						if (logger.isInfoEnabled()) {
							long elapsed = System.currentTimeMillis()
									- newDate.getTime();
							logger.info("Updated. previous=" + denki.updateTime //$NON-NLS-1$
									+ ",  this=" + updateTime + ", delay=" //$NON-NLS-1$ //$NON-NLS-2$
									+ (elapsed / (60 * 1000)) + ':'
									+ ((elapsed / 1000) % 60));
						}
						// f[^IĂʒm
						sendTell = true;
					}

				} catch (ParseException e) {
					throw Tool.newIOException("updateTime=" + updateTime, e); //$NON-NLS-1$
				}
			}
			denki.updateTime = updateTime;

			// s[ŃiW)
			line = (String) lines.get(2);
			denki.peekSupply = Integer.parseInt(line.substring(0,
					line.indexOf(',')));

			// 1ԃf[^ւ̑Ή
			// 
			denki.demand = 0;
			boolean col4 = false;
			for (int i = startLine; i <= endLine && i < lines.size(); i++) {
				line = (String) lines.get(i);
				String[] cols = line.split(","); //$NON-NLS-1$
				if (cols.length >= 4) {
					col4 |= true;
				}
				if (cols.length >= 3) {
					int demand = Integer.parseInt(cols[2]);
					if (demand == 0) {
						break;
					}
					denki.time = cols[0] + " " + cols[1]; //$NON-NLS-1$
					denki.demand = demand;
				} else {
					break;
				}
			}
			if (denki.demand == 0 && col4) {
				// d͂̐[
				for (int i = startLine; i <= endLine && i < lines.size(); i++) {
					line = (String) lines.get(i);
					String[] cols = line.split(","); //$NON-NLS-1$
					if (cols.length >= 4) {
						int demand = Integer.parseInt(cols[3]);
						if (demand == 0) {
							break;
						}
						denki.time = cols[0] + " " + cols[1]; //$NON-NLS-1$
						denki.demand = demand;
					} else {
						break;
					}
				}
			}
			if (logger.isInfoEnabled()) {
				logger.info(denki);
			}
			denki.sourceAddress = null;
			denki.error = false;
			if (sendTell) {
				udp.sendTell();
			}
			return false;
		}
	}

	public InetSocketAddress[] getTcpBindAddress(LaunchConfiguration conf)
			throws UnknownHostException {
		return new InetSocketAddress[0];
	}

	public InetSocketAddress[] getUdpBindAddress(LaunchConfiguration conf)
			throws UnknownHostException {
		String company = conf.getProperty("company", "TOKYO"); //$NON-NLS-1$ //$NON-NLS-2$
		EpcoProfile epco = getCompany(company);

		int port = getPort(conf, epco.getPort());
		InetAddress address = getBindAddress(conf);
		return new InetSocketAddress[] { new InetSocketAddress(address, port) };
	}

	// 

	public void releaseAttachments(Message message) {
		//
	}

	public void removeDownloadListener(DownloadListener listener) {
		//
	}

	public void saveAttachment(Message message, Attachment attachment, File dest) {
		//
	}

	public void sendAbsenceMode(boolean isAbsence) {
		//
	}

	public void sendFind(boolean addOnly) {
		//
	}

	public void sendMessage(Message message, UniqueMember[] destinations,
			SendingResultListener listener) {
		//
	}

	public void sendOpenNotice(Message message) {
		//
	}

	public void sendRegister() {
		//
	}

	public void sendRename() {
		//
	}

	public void addDownloadListener(DownloadListener listener) {
		//
	}

	public void cancelSavingAttachment(Attachment attachment) {
		//
	}

	public void getStatusMessage(UniqueMember member) {
		//
	}

}
