package petaclone.denki;

import java.net.MalformedURLException;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;

import jp.takayuki.tools.MessageMap;

/**
 * d͉
 * @author Takayuki Shoji
 */
public abstract class EpcoProfile {

	/**
	 * bZ[W\[X
	 */
	static final MessageMap mm = MessageMap
			.getMessageMap("petaclone.denki.denki"); //$NON-NLS-1$

	/** 5 min. */
	static final int MIN5 = 5 * 60 * 1000;

	/** 1 hour */
	static final int HOUR1 = 60 * 60 * 1000;

	/** |[gԍ̃x[X */
	private static final int PORT8496 = 8496;

	/**
	 * @return d͉Ж
	 */
	abstract String getName();

	/**
	 * @return d͉ЂID
	 */
	abstract String getId();

	/**
	 * @return URL
	 */
	abstract URL getURL();

	/**
	 * @return ŏԊu
	 */
	abstract long getMinimumInterval();

	/**
	 * @return 2(1ԃf[^)̍ŏԊu
	 */
	long get2ndMinimumInterval() {
		return HOUR1;
	}

	/**
	 * @return UPDATE̍s
	 */
	int getUpdate() {
		return 0;
	}

	/**
	 * @return s[N͂̍s
	 */
	int getPeek() {
		return 2;
	}

	/**
	 * @return f[^̊Jns
	 */
	abstract int getMinDataLine();

	/**
	 * @return f[^̏Is
	 */
	abstract int getMaxDataLine();

	/**
	 * @return 2(1ԃf[^)̊JnsB̒l̏ꍇ͖B
	 */
	int get2ndMinDataLine() {
		return -1;
	}

	/**
	 * @return 2(1ԃf[^)̏IsB̒l̏ꍇ͖B
	 */
	int get2ndMaxDataLine() {
		return -1;
	}

	/**
	 * @return XV̂
	 */
	abstract long getDelay();

	/**
	 * @return gC̒x
	 */
	long getRetryDelay() {
		return 3 * 60 * 1000; // 3min.
	}

	/**
	 * [Ȃǂ̋x~ԑтɒxԂԂ܂B
	 * @param now 
	 * @return ʒx
	 */
	long getSpecialDelay(Calendar now) {
		return 0;
	}

	/**
	 * @return |[gԍ
	 */
	abstract int getPort();

	/**
	 * Jƃf[^̎̂
	 * @return PreWait=J-updateTime-delay
	 */
	int getPreWait() {
		return 0;
	}

	/**
	 * @return F
	 */
	abstract int getColor();

	/**
	 * @param url URL
	 * @return URL
	 */
	private static URL toURL(String url) {
		try {
			return new URL(url);
		} catch (MalformedURLException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * d͂Ɠkd͂̐[̋x~ԂԂ܂B
	 * @param now 
	 * @param epco d͉
	 * @return x
	 */
	static long getMidnightDelay(Calendar now, EpcoProfile epco) {

		//SimpleDateFormat sdf = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
		// 0:00-1:00
		long baseDelay = epco.getPreWait() + epco.getDelay();
		Calendar mid = (Calendar) now.clone();
		DenkiYohouClient.setMidnight(mid);
		long start = mid.getTimeInMillis() + baseDelay;
		long end = start + (60 * 60 * 1000);
		long nowMillis = now.getTimeInMillis();

		//System.out.println(" start=" + sdf.format(new Date(start)));
		//System.out.println("   end=" + sdf.format(new Date(end)));
		//System.out.println("   now=" + sdf.format(new Date(nowMillis)));

		if (start <= nowMillis && nowMillis < end) {
			Calendar cloned = (Calendar) now.clone();
			cloned.set(Calendar.HOUR_OF_DAY, 1);
			cloned.set(Calendar.MINUTE, 0);
			cloned.set(Calendar.SECOND, 0);
			cloned.set(Calendar.MILLISECOND, 0);
			cloned.add(Calendar.MILLISECOND, (int) baseDelay);
			//System.out.println("target="
			//		+ sdf.format(new Date(cloned.getTimeInMillis())));

			return cloned.getTimeInMillis() - now.getTimeInMillis();
		}
		return 0;
	}

	/**
	 * d
	 */
	static class Tokyo extends EpcoProfile {
		String getName() {
			return mm.get("Tokyo"); //$NON-NLS-1$
		}

		String getId() {
			return "TOKYO"; //$NON-NLS-1$
		}

		URL getURL() {
			return toURL("http://www.tepco.co.jp/forecast/html/images/juyo-j.csv"); //$NON-NLS-1$
		}

		long getMinimumInterval() {
			return MIN5;
		}

		int getMinDataLine() {
			return 44;
		}

		int getMaxDataLine() {
			return 331;
		}

		int get2ndMinDataLine() {
			return 8;
		}

		int get2ndMaxDataLine() {
			return 31;
		}

		long getDelay() {
			return 30 * 1000;
		}

		long getRetryDelay() {
			return 40 * 1000;
		}

		int getPort() {
			return PORT8496 + 0;
		}

		int getPreWait() {
			return MIN5;
		}

		long getSpecialDelay(Calendar now) {
			return getMidnightDelay(now, this);
		}

		int getColor() {
			return 0xD70C19;
		}
	}

	/**
	 * kd
	 */
	static class Tohoku extends EpcoProfile {
		String getName() {
			return mm.get("Tohoku"); //$NON-NLS-1$
		}

		String getId() {
			return "TOHOKU"; //$NON-NLS-1$
		}

		URL getURL() {
			return toURL("http://setsuden.tohoku-epco.co.jp/common/demand/juyo_tohoku.csv"); //$NON-NLS-1$
		}

		long getMinimumInterval() {
			return MIN5;
		}

		int getMinDataLine() {
			return 44;
		}

		int getMaxDataLine() {
			return 331;
		}

		int get2ndMinDataLine() {
			return 8;
		}

		int get2ndMaxDataLine() {
			return 31;
		}

		long getDelay() {
			return 140 * 1000;
		}

		int getPort() {
			return PORT8496 + 1;
		}

		long getSpecialDelay(Calendar now) {
			return getMidnightDelay(now, this);
		}

		int getColor() {
			return 0xE60011;
		}
	}

	/**
	 * d
	 */
	static class Chubu extends EpcoProfile {
		String getName() {
			return mm.get("Chubu"); //$NON-NLS-1$
		}

		String getId() {
			return "CHUBU"; //$NON-NLS-1$
		}

		URL getURL() {
			return toURL("http://denki-yoho.chuden.jp/denki_yoho_content_data/juyo_cepco.csv"); //$NON-NLS-1$
		}

		long getMinimumInterval() {
			return HOUR1;
		}

		int getMinDataLine() {
			return 8;
		}

		int getMaxDataLine() {
			return 31;
		}

		long getDelay() {
			return 30 * 60 * 1000;//"Ԗ̎эXVɂ́A30قǎԂ܂"
		}

		int getPort() {
			return PORT8496 + 2;
		}

		int getColor() {
			return 0xEB8200;
		}

	}

	/**
	 * ֐d
	 */
	static class Kansai extends EpcoProfile {
		String getName() {
			return mm.get("Kansai"); //$NON-NLS-1$
		}

		String getId() {
			return "KANSAI"; //$NON-NLS-1$
		}

		URL getURL() {
			return toURL("http://www.kepco.co.jp/yamasou/juyo1_kansai.csv"); //$NON-NLS-1$
		}

		long getMinimumInterval() {
			return MIN5;
		}

		int getMinDataLine() {
			return 49;
		}

		int getMaxDataLine() {
			return 528;
		}

		int get2ndMinDataLine() {
			return 11;
		}

		int get2ndMaxDataLine() {
			return 34;
		}

		long getDelay() {
			return 10 * 1000;
		}

		int getPreWait() {
			return MIN5;
		}

		int getPort() {
			return PORT8496 + 3;
		}

		int getColor() {
			return 0xEA282A;
		}
	}

	/**
	 * Bd
	 */
	static class Kyushu extends EpcoProfile {
		String getName() {
			return mm.get("Kyushu"); //$NON-NLS-1$
		}

		String getId() {
			return "KYUSHU"; //$NON-NLS-1$
		}

		URL getURL() {
			String yyyyMMdd = new SimpleDateFormat("yyyyMMdd") //$NON-NLS-1$
					.format(new Date());
			return toURL("http://www.kyuden.co.jp/power_usages/csv/electric_power_usage" //$NON-NLS-1$
					+ yyyyMMdd + ".csv"); //$NON-NLS-1$
		}

		long getMinimumInterval() {
			return MIN5; // "т̕\̍XVɂ5قǎԂ܂"
		}

		int getMinDataLine() {
			return 8;
		}

		int getMaxDataLine() {
			return 295;
		}

		long getDelay() {
			return 215 * 1000;
		}

		long getRetryDelay() {
			return 40 * 1000;
		}

		int getPort() {
			return PORT8496 + 4;
		}

		int getColor() {
			return 0x0073BC;
		}
	}
}
