package petaclone.ipmsg;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.NoSuchElementException;
import java.util.Properties;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.ByteList;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.PropertiesUtils;
import jp.takayuki.tools.Tool;
import petaclone.Attachment;
import petaclone.Message;

/**
 * t@C̃_E[h
 * @author Takayuki Shoji
 */
public class IPMTCPClient {

	/**
	 * logger
	 */
	private static final Logger logger = Logger.getLogger(IPMTCPClient.class);

	/** fobOtO */
	private static final boolean DEBUG = logger.isDebugEnabled();

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/** Messenger */
	private IPMessenger messenger;

	/**
	 * ҂̓Ytqueue<br>
	 * NXŋL܂B<br>
	 * l: DownloadKey
	 */
	private final LinkedList dataQueue;

	/**
	 * ̃Xbh<br>
	 * NXŋL܂B<br>
	 * l: Client
	 */
	private final LinkedList activeThreads;

	/**
	 * ҋ@̃_E[hXbh̃X^bN<br>
	 * NXŋL܂B<br>
	 * l: Client
	 */
	private final LinkedList waitingThreads;

	/**
	 * ő哯_E[h(Xbh̍ő吶)
	 */
	private int maxThreadCount = 2;

	/**
	 * ݐĂXbh
	 */
	private int threadCount;

	/**
	 * ItO
	 * <p>
	 * truȅꍇ͏IɓAwaitingThreadsɂ͖߂ȂB
	 */
	private volatile boolean quit = false;

	/**
	 * _E[hXbh̑ҋ@
	 * 1 * 60 * 1000 = 1
	 */
	private static final int LIFETIME = 1 * 60 * 1000;

	/**
	 * RXgN^
	 * @param messenger bZW[
	 */
	IPMTCPClient(IPMessenger messenger) {
		this.messenger = messenger;
		dataQueue = new LinkedList();
		activeThreads = new LinkedList();
		waitingThreads = new LinkedList();
		Properties prop = messenger.configuration.getProperties();
		maxThreadCount = PropertiesUtils.getInt(prop,
				IPMessenger.CONCURRENT_DOWNLOADS, 1);
		if (maxThreadCount <= 0 || maxThreadCount > 4) {
			maxThreadCount = 2;
		}
	}

	/**
	 * ҂Ԃ̃Xbh܂B
	 */
	void releaseThreads() {
		synchronized (dataQueue) {
			synchronized (activeThreads) {
				synchronized (waitingThreads) {
					quit = true;
					Iterator itr = waitingThreads.iterator();
					while (itr.hasNext()) {
						Client client = (Client) itr.next();
						synchronized (client) {
							client.interrupt();
						}
					}
				}
			}
		}
	}

	/**
	 * Ytt@Cۑ܂B
	 * @param message bZ[W
	 * @param attachment Ytt@C
	 * @param dest ۑ
	 * @param utf8 pPbg̃GR[fBOUTF-8gpꍇtrue
	 */
	void saveAttachment(Message message, Attachment attachment, File dest,
			boolean utf8) {
		DownloadKey data = new DownloadKey();
		data.attach = attachment;
		data.dest = dest;
		data.packetId = message.id;
		data.utf8 = utf8;
		try {
			data.addr = message.getAddress();
		} catch (UnknownHostException e) {
			String str = mm.get("IPMTCPClient.convertIPAddress", //$NON-NLS-1$
					new Object[] { message.computerName });
			logger.error(str, e);
			messenger.notifyDownloadListenerToAbort(data.attach,
					data.attach.localPath, str);
			return;
		}
		data.port = message.port;
		if (DEBUG)
			print("IPMTCPClient.saveAttachment(): Attempt to lock(dataQueue)"); //$NON-NLS-1$
		synchronized (dataQueue) {
			if (DEBUG)
				print("IPMTCPClient.saveAttachment(): Attempt to lock(activeThreads)"); //$NON-NLS-1$
			synchronized (activeThreads) {
				if (DEBUG)
					print("IPMTCPClient.saveAttachment(): Attempt to lock(waitingThreads)"); //$NON-NLS-1$
				synchronized (waitingThreads) {
					if (DEBUG)
						print("IPMTCPClient.saveAttachment(): locked."); //$NON-NLS-1$
					dataQueue.addLast(data);
					// _E[hXi[ɒǉʒm
					messenger.notifyDownloadListenerToAdd(data.attach);
					// _E[hXbhɒʒm
					Client client;
					if (waitingThreads.isEmpty()) {
						//if (isConnected(data)) {
						// T[oɃANZX
						// VXbh𐶐Ȃ
						//} else
						if (threadCount < maxThreadCount) {
							// VXbh𐶐
							client = new Client();
							if (DEBUG)
								print("Generating new thread:" + threadCount); //$NON-NLS-1$
							activeThreads.addLast(client);
							client.start();
							// XbhJn1dataQueue`FbN̂
							// IɏJn
						} else {
							// őXbhɒBĂ
							// 쒆̃XbhI1dataQueue
							// `FbN̂ŎIɏJn
						}
					} else {
						// ő҂Xg菜
						// Client͎łXgԂړ悤Ƃ邪A
						// ő҂Xg菜Ȃloop()wakeup()
						// ăfbhbNN
						client = (Client) waitingThreads.removeLast();

						// (ҋ@Xbhpop(client͎remove()))

						if (DEBUG)
							print("Waiting thread " + client.getName() //$NON-NLS-1$
									+ " will be waked up."); //$NON-NLS-1$

						// Jnʒm
						client.wakeup();
						if (DEBUG)
							print("Waiting thread " + client.getName() //$NON-NLS-1$
									+ " has been waked up."); //$NON-NLS-1$
					}
				}
			}
		}
		if (DEBUG)
			print("IPMTCPClient.saveAttachment(): unlock"); //$NON-NLS-1$
	}

	///**
	// * T[oɐڑǂ𒲂ׂ
	// * gpĂȂ
	// */
	//private boolean isConnected(DownloadKey key) {
	//	Iterator itr = activeThreads.iterator();
	//	while (itr.hasNext()) {
	//		Client client = (Client) itr.next();
	//		if (client.data == null) {
	//			continue;
	//		}
	//		if (client.data.addr.equals(key.addr)) {
	//			if (DEBUG)
	//				print("ڑ̃Xbh܂" + client.getName());
	//			return true;
	//		}
	//	}
	//	if (DEBUG)
	//		print("ڑ̃Xbh͂܂");
	//	return false;
	//}

	/**
	 * fobOvg
	 * @param message fobObZ[W
	 */
	static void print(String message) {
		System.out.println(message);
	}

	/**
	 * _E[hLZ܂B
	 * @param attachment Ytt@C
	 */
	void cancelSavingAttachment(Attachment attachment) {
		if (DEBUG)
			print("cancelSavingAttachment"); //$NON-NLS-1$
		// SbN
		synchronized (dataQueue) {
			synchronized (activeThreads) {
				synchronized (waitingThreads) {
					// _E[h҂Ȃ폜
					Iterator itr = dataQueue.iterator();
					while (itr.hasNext()) {
						DownloadKey key = (DownloadKey) itr.next();
						if (key.attach == attachment) {
							itr.remove();

							// ~ʒm
							messenger.notifyDownloadListenerToSuspend(
									attachment, key.dest.getAbsolutePath());
							return;
						}
					}
					// _E[hȂXbhĒ~
					itr = activeThreads.iterator();
					while (itr.hasNext()) {
						Client client = (Client) itr.next();
						if (client.data == null) {
							// notify()Ăf[^擾悤Ƃ܂ł
							// bN̂œB邱Ƃ
							continue;
						}
						if (client.data.attach == attachment) {
							client.cancel();
							return;
						}
					}
					// bN擾܂łɃ_E[hĂ
					// ɓBH
				}
			}
		}
	}

	/**
	 * _E[hXbh<br>
	 * WISHLIST LZ̎dg݂is󋵂̒ʒm܂
	 * is󋵂͊ĎXbhgقVvƎvB
	 */
	private class Client extends Thread {
		/** _E[h */
		private DownloadKey data;

		/** ҋ@ԂłȂȂtrue */
		private boolean active = true;

		/** wait()畜AƂɒʒmĂtrue */
		private boolean notified;

		/** LZtO */
		private boolean canceled;

		/**
		 * \Pbg
		 */
		private Socket socket;

		/**
		 * ̓Xg[
		 */
		private BufferedInputStream inputStream = null;

		/**
		 * o̓obt@
		 */
		private byte[] buffer = new byte[8 * 1024];

		/**
		 * ̃RXgN^
		 */
		Client() {
			super(messenger.getThreadGroup(), "IPMTCPClient$Client"); //$NON-NLS-1$
			this.setDaemon(true);
		}

		/**
		 * method run().
		 */
		public void run() {
			synchronized (IPMTCPClient.this.getClass()) {
				threadCount++;
			}
			loop();
		}

		/**
		 * [v
		 */
		private void loop() {
			while (true) {
				if (DEBUG)
					print("IPMTCPClient.loop(): Attempt to lock(dataQueue)" //$NON-NLS-1$
							+ getName());
				synchronized (dataQueue) {
					if (DEBUG)
						print("IPMTCPClient.loop(): Attempt to lock(activeThreads)" //$NON-NLS-1$
								+ getName());
					synchronized (activeThreads) {
						if (DEBUG)
							print("IPMTCPClient.loop(): Attempt to lock(waitingThreads)" //$NON-NLS-1$
									+ getName());
						synchronized (waitingThreads) {
							if (DEBUG)
								print("IPMTCPClient.loop(): locked" + getName()); //$NON-NLS-1$
							if (quit) {
								// ItO`FbN
								return;
							}
							// f[^`FbN
							try {
								data = (DownloadKey) dataQueue.removeFirst();
								canceled = false;
							} catch (NoSuchElementException e) {
								data = null;
							}
							// 擾łȂ̂őҋ@ɓ
							if (data == null && active) {
								// ړ
								activeThreads.remove(this);
								waitingThreads.addLast(this);
								active = false;
							} else if (data != null && !active) {
								// 擾łԂ̕ύX
								// ړ
								waitingThreads.remove(this);
								activeThreads.addLast(this);
								active = true;
							}
							// bN
						}
					}
				}
				if (DEBUG)
					print("IPMTCPClient.loop(): unlocked" + getName()); //$NON-NLS-1$

				if (data != null) {
					// 
					// _E[h
					if (data.attach.type == Attachment.TYPE_FILE
							|| data.attach.type == Attachment.TYPE_INLINE) {
						downloadFile();
					} else if (data.attach.type == Attachment.TYPE_DIR) {
						downloadDirectory();
					} else {
						// ps
						messenger.notifyDownloadListenerToAbort(data.attach,
								data.dest.getAbsolutePath(),
								mm.get("IPMTCPClient.unsupportedEntryType")); //$NON-NLS-1$

					}
				} else {
					// f[^擾łȂwaiting
					// wait
					if (!waiting()) {
						return;
					}
					if (!notified) {
						expired();
						return;
					}
					notified = false;
					continue;
				}
			}
			// [vI
		}

		/**
		 * wait()܂B
		 * @return true pAfalse I
		 */
		synchronized boolean waiting() {
			if (notified == false) {
				try {
					if (DEBUG)
						print("waiting()" + getName()); //$NON-NLS-1$
					wait(LIFETIME);
				} catch (InterruptedException e) {
					expired();
					return false;
				}
				if (DEBUG)
					print("notify()ed." + getName()); //$NON-NLS-1$
			}
			return true;
		}

		/**
		 * ҋ@Ԗ(X^bNoďł)
		 */
		private void expired() {
			if (DEBUG)
				print("IPMTCPClient.Client.expired(): Attempt to lock(waitingThreads)" //$NON-NLS-1$
						+ getName());
			synchronized (waitingThreads) {
				if (DEBUG)
					print("IPMTCPClient.Client.expired()"); //$NON-NLS-1$
				try {
					waitingThreads.remove(this);
				} catch (NoSuchElementException e) {
					//
				}
				synchronized (IPMTCPClient.this.getClass()) {
					threadCount--;
				}
			}
			if (DEBUG)
				print("IPMTCPClient.Client.expired(): unlocked(waitingThreads)" //$NON-NLS-1$
						+ getName());
			logger.debug("This thread will terminate."); //$NON-NLS-1$
		}

		/**
		 * ʒm
		 */
		synchronized void wakeup() {
			notified = true;
			this.notify();
		}

		/**
		 * _E[hLZ܂B
		 */
		void cancel() {
			if (DEBUG)
				print("cancel(): download will be cancelled."); //$NON-NLS-1$

			synchronized (this) {
				canceled = true;
				closeConnection(null);
			}
		}

		/**
		 * LZƂʒm܂B(fBNg_E[h)
		 */
		private void notifySuspend() {
			if (DEBUG)
				print("notifySuspend(): Broadcast cancelling download."); //$NON-NLS-1$
			messenger.notifyDownloadListenerToSuspend(data.attach, data.dest
					.getAbsolutePath());
			canceled = false;
		}

		/**
		 * w肳ꂽXg[̃f[^t@Cɕۑ܂B
		 * @param is ̓Xg[
		 * @param file ۑ̃t@C(΃pX)
		 * @param length t@C̍ŏITCY
		 * @param append ǉI[v
		 * @return _E[hɎsꍇfalse
		 * @throws IOException o͗O
		 */
		private boolean saveInput(InputStream is, File file, long length,
				boolean append) throws IOException {

			assert !file.isDirectory();

			if (DEBUG)
				print("Saving stream:" + file.getAbsolutePath()); //$NON-NLS-1$

			FileOutputStream fo = null;
			BufferedOutputStream os = null;
			try {
				// _E[hĊĴ߂ɒǋLI[v
				fo = new FileOutputStream(file, append);
				os = new BufferedOutputStream(fo);

				long remain = length;
				remain -= (append ? file.length() : 0);

				int n;
				while (remain > 0) {
					n = is.read(buffer, 0, (int) Math
							.min(remain, buffer.length));
					if (n == -1) {
						messenger.notifyDownloadListenerToAbort(data.attach,
								data.dest.getAbsolutePath(),
								mm.get("IPMTCPClient.unexpectedEndOfStream")); //$NON-NLS-1$

						return false;
					}
					os.write(buffer, 0, n);
					remain -= n;
					data.attach.status += n;
				}

			} finally {
				if (os != null) {
					os.close();
				} else if (fo != null) {
					fo.close();
				}

			}
			return true;
		}

		/**
		 * 1t@C_E[h܂B
		 */
		private void downloadFile() {
			// Mf[^
			byte[] requestData = data.getPacketData(data.dest.length());

			OutputStream os = null;
			InputStream is = null;
			InputStream localStream = null;
			try {
				if (DEBUG) {
					print("FileID=" + data.attach.id); //$NON-NLS-1$
					print("destination=" + data.addr.getHostAddress()); //$NON-NLS-1$
					print("port=" + data.port); //$NON-NLS-1$
				}

				synchronized (this) {
					// ڑ
					socket = new Socket(data.addr, data.port);
					socket.setSoTimeout(30 * 1000);

					// v
					os = socket.getOutputStream();
					os.write(requestData);

					is = socket.getInputStream();
					inputStream = new BufferedInputStream(is);
					localStream = inputStream;
				}
				// ڑ̒ʒm
				messenger.notifyDownloadListenerToConnect(data.attach);

				// v
				long size = Math.max(localStream.available(), data.attach.size);
				long start = System.currentTimeMillis();

				// ۑ
				boolean result = saveInput(localStream, data.dest, size, true);

				// v
				long end = System.currentTimeMillis();
				data.attach.bytePerSec = 1000 * size / Math.max(1, end - start);
				if (DEBUG)
					print("size:" + size + ", time(msec):" + (end - start)); //$NON-NLS-1$ //$NON-NLS-2$

				if (result) {
					// t@CȂǂ̔f
					setFileAttribute(data.dest, data.attach);
				} else {
					// _E[hɎsăt@CTCY[Ȃ폜
					if (!canceled && data.dest.length() == 0) {
						data.dest.delete();
					}
					if (!canceled)
						return;
				}
			} catch (IOException e) {
				if (DEBUG)
					print("IPMTCPClient#canceled=" + canceled); //$NON-NLS-1$

				if (!canceled) {
					logger.error("Failed to download file.", e); //$NON-NLS-1$
					// Xi[ɈُIʒm
					messenger.notifyDownloadListenerToAbort(data.attach,
							data.dest.getAbsolutePath(), e
									.getLocalizedMessage());
					return;
				}
			} finally {
				try {
					if (os != null) {
						os.close();
					}
				} catch (IOException e) {
					logger.error("Failed to close socket output stream " //$NON-NLS-1$
							+ "during file download.", e); //$NON-NLS-1$
				}

				// ؒf
				closeConnection(is);
			}

			//
			if (canceled) {
				// LZ̒ʒm
				notifySuspend();
			} else {
				// I̒ʒm
				messenger.notifyDownloadListenerToFinish(data.attach, data.dest
						.getAbsolutePath());
				if (DEBUG)
					print("Download successful:thread: " + getName() //$NON-NLS-1$
							+ " Attachment: " + data.attach); //$NON-NLS-1$
			}
		}

		/**
		 * T[oؒf܂B
		 * @param is inputStream̐eXg[
		 */
		private synchronized void closeConnection(InputStream is) {
			try {
				if (inputStream != null) {
					inputStream.close();
					inputStream = null;
				} else if (is != null) {
					is.close();
				}
			} catch (IOException e) {
				logger.error("Failed to close socket input stream.", e); //$NON-NLS-1$
			}
			if (socket != null && !socket.isClosed()) {
				try {
					socket.close();
				} catch (IOException e) {
					logger.error("Failed to close socket.", e); //$NON-NLS-1$
				}
			}
			socket = null;
		}

		/**
		 * t@Cݒ肵܂B
		 * @param file t@C
		 * @param att Ytt@C
		 */
		private void setFileAttribute(File file, Attachment att) {
			file.setLastModified(att.lastModified * 1000);
			if (att.isReadOnly) {
				file.setReadOnly();
			}
		}

		/**
		 * Kw_E[h܂B
		 */
		private void downloadDirectory() {
			// Mf[^
			byte[] requestData = data.getPacketData(0);

			OutputStream os = null;
			InputStream is = null;
			InputStream localStream = null;
			try {
				if (DEBUG) {
					print("FileID=" + data.attach.id); //$NON-NLS-1$
					print("destination=" + data.addr.getHostAddress()); //$NON-NLS-1$
					print("port=" + data.port); //$NON-NLS-1$
				}

				synchronized (this) {
					// ڑ
					socket = new Socket(data.addr, data.port);
					socket.setSoTimeout(30 * 1000);

					// v
					os = socket.getOutputStream();
					os.write(requestData);

					is = socket.getInputStream();
					inputStream = new BufferedInputStream(is);
					localStream = inputStream;
				}
				// ڑ̒ʒm
				messenger.notifyDownloadListenerToConnect(data.attach);

				// ݂̃fBNg(data.dest̓[gɂȂ)
				File current = data.dest;
				// ŏ̃fBNg͖
				String encoding = IPMessenger.getEncoding(data.utf8);
				Attachment first = parseHierarchy(localStream, encoding);
				//System.out.println("ŏ̊Kw: " + first);
				if (first == null || first.type != Attachment.TYPE_DIR) {
					messenger.notifyDownloadListenerToAbort(data.attach,
							data.dest.getAbsolutePath(),
							mm.get("IPMTCPClient.invalidHierarchyInformation")); //$NON-NLS-1$
					return;
				}
				int dirCount = 1;
				while (true) {
					// Kw̉
					Attachment attach = parseHierarchy(localStream, encoding);
					//System.out.println("Kw " + attach);
					if (attach == null) {
						messenger
								.notifyDownloadListenerToAbort(
										data.attach,
										data.dest.getAbsolutePath(),
										mm
												.get("IPMTCPClient.invalidHierarchyInformation")); //$NON-NLS-1$
						return;
					}
					switch (attach.type) {
					case Attachment.TYPE_DIR:
						// fBNg
						dirCount++;
						attach.validateName();
						current = new File(current, attach.name);
						current.mkdirs();
						//current.setLastModified(attach.lastModified * 1000);
						if (attach.isReadOnly) {
							current.setReadOnly();
						}
						break;
					case Attachment.TYPE_PARENT:
						// 1Kwɖ߂
						dirCount--;
						current = current.getParentFile();
						current.setLastModified(attach.lastModified * 1000);
						//System.out.println(
						//	"ŏIXV" + current.lastModified());
						break;
					case Attachment.TYPE_FILE:
						// t@C
						attach.validateName();
						File saveTo = new File(current, attach.name);
						saveInput(localStream, saveTo, attach.size, false);
						setFileAttribute(saveTo, attach);
						break;
					default:
						if (DEBUG)
							print("Other " + attach); //$NON-NLS-1$
						// ̑(XLbv)
						long skip = attach.size;
						while (skip > 0) {
							long skipped = localStream.skip(skip);
							if (skipped <= 0) {
								break;
							}
							skip -= skipped;
						}
						break;
					}
					// ݂̃fBNg[gfBNĝƂI
					if (dirCount == 0) {
						break;
					}
				}

			} catch (IOException e) {
				if (DEBUG)
					print("IPMTCPClient#canceled=" + canceled); //$NON-NLS-1$

				if (!canceled) {
					logger.error("Failed to download directory.", e); //$NON-NLS-1$
					// Xi[ɈُIʒm
					messenger.notifyDownloadListenerToAbort(data.attach,
							data.dest.getAbsolutePath(), e.toString());
					return;
				}
			} finally {
				if (os != null) {
					try {
						os.close();
					} catch (IOException e) {
						logger.error("Failed to close socket output stream.", //$NON-NLS-1$
								e);
					}
				}

				closeConnection(is);
			}

			if (canceled) {
				// LZ̒ʒm
				notifySuspend();
			} else {
				// I
				messenger.notifyDownloadListenerToFinish(data.attach, data.dest
						.getAbsolutePath());
				if (DEBUG)
					print("Directory download successful:thread: " + getName()); //$NON-NLS-1$
			}
		}

		/**
		 * Kwp[X܂B
		 * @param is ̓Xg[
		 * @param encoding wb_̃GR[fBO
		 * @return null - f[^̌`sK؂ŏpłȂ
		 * @throws IOException o͗O
		 */
		private Attachment parseHierarchy(InputStream is, String encoding)
				throws IOException {
			// ŏ̃R܂œǂ
			int b;
			// {ƂIP Messenger̃wb_TCY4oCg
			StringBuffer sb = new StringBuffer(4);
			while (true) {
				b = is.read();
				if (b == ':') {
					break;
				} else if (b == -1) {
					return null;
				}
				sb.append((char) b);
			}
			int headerSize;
			try {
				headerSize = Integer.parseInt(sb.toString(), 16);
			} catch (NumberFormatException e) {
				logger.warn("Invalid header size of hierarchy information.", e); //$NON-NLS-1$
				return null;
			}
			// łɓǂݍ񂾕(R̕)
			headerSize -= sb.length() + 1;
			byte[] buffer = new byte[headerSize];
			// c̃f[^擾
			int remain = headerSize;
			int n;
			while (remain > 0) {
				n = is.read(buffer, headerSize - remain, remain);
				if (n == -1) {
					return null;
				}
				remain -= n;
			}

			// 
			ArrayList elements = new ArrayList();
			ByteList buf = new ByteList();
			for (int i = 0; i < buffer.length; i++) {
				if (buffer[i] == ':') {
					if (elements.isEmpty() && i + 1 < headerSize
							&& buffer[i + 1] == ':') {
						// GXP[vꂽ
						buf.add(buffer[i]);
						i++;
					} else {
						// ʏ̋؂
						elements.add(buf.toArray());
						buf = new ByteList();
					}
				} else {
					buf.add(buffer[i]);
				}
			}

			if (elements.size() < 3) {
				return null;
			}
			// ẻ
			Attachment att = new Attachment();
			// t@C
			att.name = new String((byte[]) elements.get(0), encoding);
			try {
				// t@CTCY
				att.size = Long.parseLong(new String((byte[]) elements.get(1),
						encoding), 16);
				// 
				int type = Integer.parseInt(new String(
						(byte[]) elements.get(2), encoding), 16);
				// 8rbg͎
				switch (type & 0xff) {
				case (int) IPMSG.FILE_REGULAR:
					att.type = Attachment.TYPE_FILE;
					break;
				case (int) IPMSG.FILE_DIR:
					att.type = Attachment.TYPE_DIR;
					break;
				case (int) IPMSG.FILE_RETPARENT:
					att.type = Attachment.TYPE_PARENT;
					break;
				default:
					att.type = Attachment.TYPE_OTHER;
					break;
				}
				// 24rbg͑
				int attr = type & 0xffffff00;
				if ((attr & IPMSG.FILE_RONLYOPT) != 0) {
					att.isReadOnly = true;
				}
				if ((attr & IPMSG.FILE_HIDDENOPT) != 0) {
					att.isHidden = true;
				}
			} catch (NumberFormatException e) {
				logger
						.warn(
								"Invalid file size or entry type in hierarchy information.", e); //$NON-NLS-1$
				return null;
			}
			// t@C
			for (int i = 3; i < elements.size(); i++) {
				byte[] bytes = (byte[]) elements.get(i);
				int pos = Tool.indexOf(bytes, 0, '=');
				if (pos == bytes.length) {
					// =ꍇ͓ɑΉĂȂ̂ŃpX
					continue;
				}
				String key = new String(bytes, 0, pos, encoding);
				try {
					long attr = Long.parseLong(key, 16);
					pos++;
					if (attr == IPMSG.FILE_MTIME) {
						String val = new String(bytes, pos, bytes.length - pos,
								encoding);
						att.lastModified = Long.parseLong(val, 16);
					} else if (attr == IPMSG.FILE_UNICODEFNAME) {
						att.name = new String(bytes, pos, bytes.length - pos,
								IPMessenger.ENCODING_UTF8);
					}
				} catch (NumberFormatException e) {
					logger.warn("Invalid extended attributes.", e); //$NON-NLS-1$
					continue;
				}
			}
			return att;
		}
	}

	/**
	 * _E[hɕKvȃf[^
	 */
	private class DownloadKey {
		/** v */
		InetAddress addr;

		/** |[gԍ */
		int port;

		/** pPbgID */
		long packetId;

		/** Yt */
		Attachment attach;

		/** ۑ̃fBNg܂̓t@C */
		File dest;

		/** pPbg̃GR[fBOUTF-8gpꍇtrue */
		boolean utf8;

		/**
		 * pPbgf[^擾܂B
		 * @param offset _E[h̊Jnʒu
		 * @return oCi
		 */
		byte[] getPacketData(long offset) {
			IPMPacket packet = new IPMPacket(messenger);
			// v
			StringBuffer sb = new StringBuffer();
			sb.append(Long.toHexString(packetId)).append(':');
			sb.append(Integer.toHexString(attach.id)).append(':');
			if (attach.type != Attachment.TYPE_DIR) {
				sb.append(Long.toHexString(offset)).append(':');
			}
			packet.data = sb.toString();
			// R}h
			long utf8opt = utf8 ? IPMSG.UTF8OPT : 0;
			switch (attach.type) {
			case Attachment.TYPE_FILE:
			case Attachment.TYPE_INLINE:
				packet.command = IPMSG.GETFILEDATA | utf8opt;
				break;
			case Attachment.TYPE_DIR:
				packet.command = IPMSG.GETDIRFILES | utf8opt;
				break;
			default:
				return null;
			}
			// pPbgID
			packet.setPacketNumber();
			return packet.getBytes();
		}
	}
}
