package petaclone.ipmsg;

import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.BindException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.security.KeyFactory;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Security;
import java.security.interfaces.RSAPrivateKey;
import java.security.interfaces.RSAPublicKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.RSAPrivateKeySpec;
import java.security.spec.RSAPublicKeySpec;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Timer;
import java.util.Vector;

import javax.crypto.Cipher;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.InetAddressUtils;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.PropertiesUtils;
import jp.takayuki.tools.Task;
import jp.takayuki.tools.Tool;

import org.bouncycastle.jce.provider.BouncyCastleProvider;

import petaclone.AppPreferences;
import petaclone.Attachment;
import petaclone.ControlCreator;
import petaclone.DownloadListener;
import petaclone.LaunchConfiguration;
import petaclone.LaunchException;
import petaclone.Member;
import petaclone.Message;
import petaclone.Messenger;
import petaclone.Protocol;
import petaclone.SendingResultListener;
import petaclone.ServerListener;
import petaclone.UniqueMember;

/**
 * IP Messenger
 * @author Takayuki Shoji
 */
public class IPMessenger implements Messenger {

	/**
	 * logger
	 */
	private static final Logger logger = Logger.getLogger(IPMessenger.class);

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/**
	 * _CAbvIvV
	 */
	private static final String DIALUP = "dialup"; //$NON-NLS-1$

	/**
	 * ÍIvV
	 */
	private static final String USE_PLAIN = "usePlain"; //$NON-NLS-1$

	/**
	 * ÍNƂɐ
	 */
	private static final String GENERATE_KEY_EACH_TIME = "generateKeyEachTime"; //$NON-NLS-1$

	/**
	 * Unicodet@C𑗐M
	 */
	private static final String SEND_UNICODE_FILENAME = "sendUnicodeFilename"; //$NON-NLS-1$

	/**
	 * _E[h
	 */
	static final String CONCURRENT_DOWNLOADS = "concurrentDownloads"; //$NON-NLS-1$

	/** fobOtO */
	private static final boolean DEBUG = logger.isDebugEnabled();

	/** GR[fBO */
	static final String ENCODING;
	static {
		ENCODING = mm.get("IPMessenger.encoding"); //$NON-NLS-1$
	}

	/**
	 * UTF8OPT̃GR[fBO
	 */
	static final String ENCODING_UTF8 = "UTF-8"; //$NON-NLS-1$

	/**
	 * vgRB
	 */
	private static Protocol protocol;
	static {
		// vgR
		protocol = new Protocol();
		//IP Messenger̃eLXg{bNXɓől
		protocol.maxMessageLength = 30000;
		protocol.maxUserNameLength = 16;
		protocol.protocolName = "IP Messenger"; //$NON-NLS-1$
		protocol.signature = "ipmsg"; //$NON-NLS-1$
	}

	/** pPbgԍp̃JE^ */
	static int count = 1;

	/**
	 * AvP[Vݒ
	 */
	private AppPreferences pref;

	/**
	 * CX^XƂ̐ݒ
	 */
	LaunchConfiguration configuration;

	/**
	 * T[oXi[
	 */
	private ServerListener serverListener;

	/** |[gԍ(configurationǂݎAۑ) */
	int port = 2425;

	/** |[gԍ̃vpeB[ */
	final static String PORT = "port"; //$NON-NLS-1$

	/** oChAhX */
	private InetAddress bindAddress = null;

	/** UDPT[o */
	IPMUDPServer udp;

	/** UDPT[oXbh */
	Thread udpThread;

	/** TCPT[o */
	IPMTCPServer tcp;

	/** TCPT[oXbh */
	Thread tcpThread;

	/** _E[hXi[ */
	Vector downloadListeners;

	/** TCPNCAg */
	IPMTCPClient tcpClient;

	/** ^C}[ */
	Timer timer;

	/** J */
	PublicKey publicKey;

	/** 閧 */
	PrivateKey privateKey;

	/**
	 * ÍIvV
	 */
	private long encryptOption;

	static {
		if (!Tool.javaVersion1_5_orHigher)
			Security.addProvider(new BouncyCastleProvider());
	}

	/**
	 * XbhO[v
	 */
	private ThreadGroup threadGroup;

	/**
	 * ̃RXgN^
	 */
	public IPMessenger() {
		//
	}

	/**
	 * VCX^X𐶐܂B
	 */
	public Messenger newInstance() {
		return new IPMessenger();
	}

	/**
	 * ÍΉ\͂Ԃ܂B
	 * @return ÍΉ\
	 */
	long getCipherCapability() {

		if (pref.useSmallKey()) {
			// RC2 40bit[h
			return IPMSG.RSA_512 | IPMSG.RC2_40;
		}

		long cap = IPMSG.RSA_1024 | IPMSG.BLOWFISH_128;
		//cap |= IPMSG.SIGN_MD5;

		// x̊Ǌ|V[t@CCXg[Ă邩
		if (pref.getCipherIsUnlimited()) {
			cap |= IPMSG.RC2_40 | IPMSG.RC2_128;
		}
		return cap;
	}

	/**
	 * JA閧̐
	 * @return IPMSG.ENCRYPTOPT
	 */
	private long generateKeys() {

		if (isOptionUsePlain()) {
			return 0;
		}

		try {
			if (!isOptionGenerateKeyEachTime()) {
				long result = restoreRSAKeys();
				if (result != 0) {
					return result;
				}
			}
		} catch (Exception e) {
			logger.warn("Exception occurred while restoring key. " //$NON-NLS-1$
					+ "Keys will be generated.", e); //$NON-NLS-1$
		}
		try {
			// V
			KeyPairGenerator kpg = KeyPairGenerator.getInstance("RSA"); //$NON-NLS-1$

			if (pref.useSmallKey()) {
				kpg.initialize(512);
			} else {
				kpg.initialize(1024);
			}

			KeyPair kp = kpg.generateKeyPair();

			// ̐
			this.publicKey = kp.getPublic();
			this.privateKey = kp.getPrivate();

			// 
			String pub = IPMCipher
					.makeRSAPublicKeyString((RSAPublicKey) this.publicKey);
			String pri = IPMCipher
					.makeRSAPrivateKeyString((RSAPrivateKey) this.privateKey);

			// N\ɕۑ
			this.configuration.setProperty("publicKey", pub); //$NON-NLS-1$
			this.configuration.setProperty("privateKey", pri); //$NON-NLS-1$
			this.configuration.save();
			logger.info("Keys were generated."); //$NON-NLS-1$

		} catch (Exception e) {
			logger.warn("Exception occurred while generating keys.", e); //$NON-NLS-1$
			return 0;
		}
		return IPMSG.ENCRYPTOPT;
	}

	/**
	 * RSǍJƔ閧vpeBtB[hɕ܂B
	 * 
	 * @return ÍΉIvV(IPMSG.ENCRYPTOPT)܂0B
	 * @throws NoSuchAlgorithmException
	 * @throws InvalidKeySpecException
	 */
	private long restoreRSAKeys() throws NoSuchAlgorithmException,
			InvalidKeySpecException {
		String pub = this.configuration.getProperty("publicKey", null); //$NON-NLS-1$
		String pri = this.configuration.getProperty("privateKey", null); //$NON-NLS-1$
		if (pub == null || pri == null) {
			return 0;
		}
		if (pref.useSmallKey() && pri.length() > 128 * 2 + 1) {
			// 128 / 2 * 8 = 512(bit)
			return 0;
		}

		KeyFactory kf = KeyFactory.getInstance("RSA"); //$NON-NLS-1$

		// J
		RSAPublicKeySpec pubspec = IPMCipher.makePublicKeySpec(pub);
		this.publicKey = kf.generatePublic(pubspec);

		// 閧
		RSAPrivateKeySpec privspec = IPMCipher.makePrivateKeySpec(pri);
		this.privateKey = kf.generatePrivate(privspec);
		logger.info("Keys were restored."); //$NON-NLS-1$

		return IPMSG.ENCRYPTOPT;
	}

	/**
	 * @throws LaunchException 
	 * @see petaclone.Messenger#begin(ThreadGroup)
	 */
	public void begin(ThreadGroup gr) throws LaunchException {
		this.threadGroup = gr;

		// JA閧̐
		this.encryptOption = generateKeys();

		// |[gݒ
		this.port = getPort(this.configuration);
		// oChAhX擾
		this.bindAddress = getBindAddress(this.configuration);

		if (Tool.javaVersion1_5_orHigher) {
			this.timer = new Timer("IP Messenger Timer " + port, true); //$NON-NLS-1$
		} else {
			this.timer = new Timer(true);
		}

		try {
			tcp = new IPMTCPServer(this);
			tcp.bind(this.bindAddress);
		} catch (BindException e) {
			throw new LaunchException(mm
					.get("IPMessenger.serverAddressIsUnavailable"), //$NON-NLS-1$
					e, this.configuration);

		} catch (SocketException e) {
			throw new LaunchException(mm
					.get("IPMessenger.underlyingProtocolError"), //$NON-NLS-1$
					e, this.configuration);

		} catch (IOException e) {
			throw new LaunchException(mm.get("IPMessenger.IOException"), //$NON-NLS-1$
					e, this.configuration);
		}

		try {
			udp = new IPMUDPServer(this, this.encryptOption);
			udp.bind(this.bindAddress);

		} catch (BindException e) {
			tcp.quit();
			throw new LaunchException(mm
					.get("IPMessenger.serverAddressIsUnavailable"), //$NON-NLS-1$
					e, this.configuration);

		} catch (SocketException e) {
			tcp.quit();
			throw new LaunchException(mm
					.get("IPMessenger.underlyingProtocolError"), //$NON-NLS-1$
					e, this.configuration);

		} catch (IOException e) {
			tcp.quit();
			throw new LaunchException(mm.get("IPMessenger.IOException"), //$NON-NLS-1$
					e, this.configuration);
		}

		// TCPT[oXbh
		tcpThread = new Thread(gr, tcp);
		tcpThread.setName("IP Messenger TCP " + port); //$NON-NLS-1$
		tcpThread.start();

		// UDPT[oXbh
		udpThread = new Thread(gr, udp);
		udpThread.setName("IP Messenger UDP " + port); //$NON-NLS-1$
		udpThread.start();

		// TCPNCAg(Xbh͓Ǘ)
		tcpClient = new IPMTCPClient(this);
	}

	/**
	 * |[gԍ擾܂B<br>
	 * Nݒ肩|[gԍ擾܂B
	 * |[gԍȂtB[hɐݒ肵܂B
	 * lɕϊłȂftHg|[gԍԂ܂B
	 * @param conf Nݒ
	 * @return |[gԍ
	 */
	private int getPort(LaunchConfiguration conf) {
		String portStr = conf.getProperty(PORT, "2425"); //$NON-NLS-1$
		try {
			int value = Integer.parseInt(portStr.trim());
			if (0 < value && value <= 65535) {
				return value;
			}
		} catch (NumberFormatException e) {
			//
		}
		return 2425;
	}

	/**
	 * oChAhX擾܂B
	 * @param conf Nݒ
	 * @return \PbgɎw肷ׂInetAddress
	 */
	private InetAddress getBindAddress(LaunchConfiguration conf) {
		int addrNumber = conf.getBindAddress();
		return InetAddressUtils.getLocalAddress(addrNumber);
	}

	/**
	 * @see petaclone.Messenger#quit()
	 */
	public void quit() {
		udp.sendDelete();
		Task.schedule(timer, new QuitTask(), 800);
		try {
			udpThread.join();
		} catch (InterruptedException e) {
			logger.error(e);
		}
	}

	/**
	 * I^XNB
	 */
	private class QuitTask implements Runnable {

		public void run() {
			udp.quit();
			tcp.quit();
			tcpClient.releaseThreads();
			try {
				udpThread.join();
			} catch (InterruptedException e) {
				logger.error("Interrupted while waiting for UDP thread", e); //$NON-NLS-1$
			}
			timer.cancel();
		}
	}

	/**
	 * AppPreferencesݒ肵܂B
	 * @param pref AvP[Vݒ
	 */
	public void setAppPreferences(AppPreferences pref) {
		this.pref = pref;
	}

	/**
	 * AvP[Vݒ擾܂B
	 * @return AppPreferences
	 */
	AppPreferences getAppPreferences() {
		return this.pref;
	}

	/**
	 * bZ[W𕜍܂B
	 * @param src ÍbZ[W
	 * @param encoding GR[fBO
	 * @return bZ[W
	 */
	String decryptMessage(String src, String encoding) {
		if (this.privateKey == null) {
			return src;
		}
		long start = System.currentTimeMillis();
		try {
			StringTokenizer token = new StringTokenizer(src, ":"); //$NON-NLS-1$
			long cap = Long.parseLong(token.nextToken(), 16);
			byte[] keydata = Tool.hexStringToBytes(token.nextToken());
			byte[] msgdata = Tool.hexStringToBytes(token.nextToken());

			// ÍASY擾
			String cipherName, algorithm;
			if ((cap & (IPMSG.BLOWFISH_128 | IPMSG.BLOWFISH_256)) != 0) {
				cipherName = IPMSG.CIPHER_BLOWFISH;
				algorithm = "Blowfish"; //$NON-NLS-1$
			} else if ((cap & (IPMSG.RC2_40 | IPMSG.RC2_128 | IPMSG.RC2_256)) != 0) {
				cipherName = IPMSG.CIPHER_RC2;
				algorithm = "RC2"; //$NON-NLS-1$
			} else {
				throw new Exception(mm.get("IPMessenger.unsupportedEncryption", //$NON-NLS-1$
						new Object[] { Long.toHexString(cap) }));
			}

			// 閧gČJňÍꂽʌ𕜍
			Cipher c = Cipher.getInstance(IPMSG.CIPHER_RSA);
			c.init(Cipher.DECRYPT_MODE, privateKey);
			byte[] cKey = c.doFinal(keydata);

			// ʌňÍꂽf[^𕜍
			c = Cipher.getInstance(cipherName);
			IvParameterSpec iv = new IvParameterSpec(new byte[8]);
			SecretKeySpec sks = new SecretKeySpec(cKey, algorithm);

			c.init(Cipher.DECRYPT_MODE, sks, iv);
			byte[] msg = c.doFinal(msgdata);

			return new String(msg, 0, Tool.lengthToZero(msg, 0, msg.length),
					encoding);
		} catch (Exception ex) {
			logger.error("Failed to decrypt. src=" + src, ex); //$NON-NLS-1$
			StringBuffer sb = new StringBuffer();
			sb.append(ex.getMessage());
			sb.append('\n');
			sb.append(mm.get("IPMessenger.failedToDecrypt")).append('\n'); //$NON-NLS-1$
			sb.append("---\n"); //$NON-NLS-1$
			sb.append(src);
			return sb.toString();
		} finally {
			if (DEBUG) {
				long time = System.currentTimeMillis() - start;
				logger.debug("decryptMessage(String, String) end. time=" //$NON-NLS-1$
						+ time);
			}
		}
	}

	/**
	 * CX^XƂ̐ݒݒ肵܂B
	 */
	public void setConfiguration(LaunchConfiguration conf) {
		this.configuration = conf;
	}

	/**
	 * NIDԂ܂B
	 */
	public String getId() {
		if (configuration == null) {
			return null;
		}
		return configuration.getId();
	}

	/**
	 * @see petaclone.Messenger#getAlive()
	 */
	public boolean getAlive() {
		if (udp == null || tcp == null) {
			return false;
		}
		return udp.getAlive() && tcp.getAlive();
	}

	/**
	 * T[oXi[ݒ肵܂B
	 */
	public void setServerListener(ServerListener listener) {
		serverListener = listener;
	}

	/**
	 * T[oXi[ɃG[ʒm܂B
	 * @param message G[bZ[W
	 * @param t ThrowableIuWFNg
	 */
	void notifyError(String message, Throwable t) {
		serverListener.onError(message, t);
	}

	/**
	 * T[oXi[ɃbZ[W̓ʒm܂B
	 * @param message bZ[W
	 */
	void notifyMessage(Message message) {
		serverListener.onMessage(message);
	}

	/**
	 * T[oXi[ɊJʒm̓ʒm܂B
	 * @param message Jʒm
	 */
	void notifyOpenNotice(Message message) {
		serverListener.onOpenNotice(message);
	}

	/**
	 * T[oXi[Ƀo[̍폜ʒm܂B
	 * @param member o[
	 */
	void notifyDeleteMember(Member member) {
		serverListener.deleteMember(member);
	}

	/**
	 * T[oXi[Ƀo[̒ǉʒm܂B
	 * @param member VKo[
	 */
	void notifyAddMember(Member member) {
		serverListener.addMember(member);
	}

	/**
	 * T[oXi[Ƀo[̍XVʒm܂B
	 * @param member XVꂽo[
	 */
	void notifyUpdateMember(Member member) {
		serverListener.addMember(member);
	}

	/**
	 * @see petaclone.Messenger#sendRegister()
	 */
	public void sendRegister() {
		udp.sendRegister();
	}

	/**
	 * @see petaclone.Messenger#sendFind(boolean)
	 */
	public void sendFind(boolean addOnly) {
		udp.sendFind(addOnly);
	}

	/**
	 * @see petaclone.Messenger#sendRename()
	 */
	public void sendRename() {
		udp.sendUpdate();
	}

	/**
	 * @see petaclone.Messenger#sendAbsenceMode(boolean)
	 */
	public void sendAbsenceMode(boolean isAbsence) {
		udp.sendUpdate();
	}

	/**
	 * bZ[W𑗐M܂B
	 * @see petaclone.Messenger#sendMessage(Message, UniqueMember[], SendingResultListener)
	 */
	public void sendMessage(Message message, UniqueMember[] destinations,
			SendingResultListener listener) {

		// FIXME Xbhg
		SendMessage send = new SendMessage(message, destinations, listener);
		send.start();
	}

	/**
	 * MXbh
	 */
	private class SendMessage extends Thread {
		/** bZ[W */
		Message message;

		/** Đ */
		UniqueMember[] dest;

		/** MʃXi[ */
		SendingResultListener listener;

		/**
		 * RXgN^
		 * @param message bZ[W
		 * @param dest Đ
		 * @param listener MʃXi[
		 */
		SendMessage(Message message, UniqueMember[] dest,
				SendingResultListener listener) {

			super(IPMessenger.this.getThreadGroup(), "IPMessenger$SendMessage"); //$NON-NLS-1$
			this.message = message;
			this.dest = dest;
			this.listener = listener;
		}

		public void run() {
			udp.sendMessage(message, dest, listener);
		}
	}

	/**
	 * @see petaclone.Messenger#sendOpenNotice(Message)
	 */
	public void sendOpenNotice(Message message) {
		udp.sendOpenNotice(message);
	}

	/**
	 * @see petaclone.Messenger#getProtocol()
	 */
	public Protocol getProtocol() {
		return protocol;
	}

	/**
	 * @see petaclone.Messenger#createConfigControl(ControlCreator)
	 */
	public void createConfigControl(ControlCreator creator) {
		// _CAbv

		creator.createCheckBox(DIALUP, //
				mm.get("IPMessenger.dialupConnection.label"), false); //$NON-NLS-1$
		creator.createLabel(mm.get("IPMessenger.dialupConnectionDesc.label")); //$NON-NLS-1$
		creator.createSeparator();
		// |[gԍ
		creator.createSpinner(PORT, mm.get("IPMessenger.portNumber.label"), //$NON-NLS-1$
				2425, true, 1, 65535, 1);
		creator.createLabel(mm.get("IPMessenger.portNumberDesc.label")); //$NON-NLS-1$
		creator.createSeparator();

		// _E[h
		creator.createSpinner(CONCURRENT_DOWNLOADS,
				mm.get("IPMessenger.concurrentDownloads"), 2, true, 1, 4, 1); //$NON-NLS-1$

		// ̐
		creator.createCheckBox(GENERATE_KEY_EACH_TIME, //
				mm.get("IPMessenger.generateNewKeyAtEachLaunch"), //$NON-NLS-1$
				false);
		creator.createLabel(//
				mm.get("IPMessenger.generateNewKeyAtEachLaunchDesc")); //$NON-NLS-1$

		// Í
		creator.createCheckBox(USE_PLAIN, //
				mm.get("IPMessenger.doNotEncrypt.label"), false); //$NON-NLS-1$
		creator.createSeparator();

		// Unicodet@C̑M
		creator.createCheckBox(SEND_UNICODE_FILENAME, //
				mm.get("IPMessenger.sendUnicodeFileName"), false); //$NON-NLS-1$
		creator.createLabel(mm.get("IPMessenger.sendUnicodeFileNameDesc")); //$NON-NLS-1$
		creator.createSeparator();

		//
		creator.createLabel(mm.get("IPMessenger.ipmsgURL")); //$NON-NLS-1$

	}

	/**
	 * _E[hXi[ǉ܂B
	 */
	public void addDownloadListener(DownloadListener listener) {
		if (downloadListeners == null) {
			downloadListeners = new Vector();
		}
		downloadListeners.add(listener);
	}

	/**
	 * _E[hXi[폜܂B
	 */
	public void removeDownloadListener(DownloadListener listener) {
		assert downloadListeners != null;
		downloadListeners.remove(listener);
	}

	/**
	 * TCPT[oɓYtt@C̔jʒm܂B
	 * (UDPT[oĂяo)
	 * @param packetId pPbgID
	 */
	void releaseAttachment(long packetId) {
		tcp.removePermission(packetId);
	}

	/**
	 * UDPT[oTCPT[oɓYtt@C̃_E[h^܂B
	 * @param packetId pPbgID
	 * @param attachments Ytt@CXg(l: Attachment)B
	 * List͕ύX邽߁AMessage.attachmentsƂ͕ʂ̃CX^Xw肷KvB
	 */
	void addDownloadPermission(long packetId, List attachments) {
		tcp.addPermission(packetId, attachments);
	}

	/**
	 * Ytt@Cۑ܂B
	 */
	public void saveAttachment(Message message, Attachment attachment, File dest) {
		// ΃pX
		assert dest.isAbsolute();
		boolean utf8 = udp.isMessageFromUTF8Member(message);
		tcpClient.saveAttachment(message, attachment, dest, utf8);
		// _E[hXi[ւ̒ʒmtcpClients
	}

	/**
	 * Ytt@C̃_E[hLZ܂B
	 */
	public void cancelSavingAttachment(Attachment attachment) {
		tcpClient.cancelSavingAttachment(attachment);
		// _E[hXi[ւ̒ʒmtcpClients
	}

	/**
	 * Ytt@Cj܂B
	 * @param message bZ[W
	 */
	public void releaseAttachments(Message message) {
		udp.releaseAttachments(message);
	}

	/**
	 * _E[hXi[ɐV_E[hn܂邱Ƃʒm܂B
	 * @param attachment Ώۂ̓Ytt@C
	 */
	void notifyDownloadListenerToAdd(Attachment attachment) {
		attachment.status = Attachment.STATUS_WAIT;
		if (attachment.listener != null) {
			attachment.listener.onAdd(this, attachment);
		}

		if (downloadListeners == null) {
			return;
		}
		synchronized (downloadListeners) {
			Iterator itr = downloadListeners.iterator();
			while (itr.hasNext()) {
				DownloadListener listener = (DownloadListener) itr.next();
				listener.onAdd(this, attachment);
			}
		}
	}

	/**
	 * _E[hXi[ɃT[oɐڑƂʒm܂B
	 * @param attachment Ώۂ̓Ytt@C
	 */
	void notifyDownloadListenerToConnect(Attachment attachment) {
		attachment.status = Attachment.STATUS_CONNECTED;
		if (attachment.listener != null) {
			attachment.listener.onConnect(attachment);
		}

		if (downloadListeners == null) {
			return;
		}
		synchronized (downloadListeners) {
			Iterator itr = downloadListeners.iterator();
			while (itr.hasNext()) {
				DownloadListener listener = (DownloadListener) itr.next();
				listener.onConnect(attachment);
			}
		}
	}

	/**
	 * _E[hXi[Ƀ_E[hIƂʒm܂B
	 * @param attachment Ytt@C
	 * @param localPath ۑ
	 */
	void notifyDownloadListenerToFinish(Attachment attachment, String localPath) {

		attachment.status = Attachment.STATUS_FINISHED;
		attachment.localPath = localPath;
		// c[`bvLbVXV
		attachment.setToolTipCache(null);

		if (attachment.listener != null) {
			attachment.listener.onFinished(attachment);
		}

		if (downloadListeners == null) {
			return;
		}
		synchronized (downloadListeners) {
			Iterator itr = downloadListeners.iterator();
			while (itr.hasNext()) {
				DownloadListener listener = (DownloadListener) itr.next();
				listener.onFinished(attachment);
			}
		}
	}

	/**
	 * _E[hXi[Ƀ_E[hُIƂʒm܂B
	 * @param attachment ُIΏ
	 * @param localPath ۑpX
	 * @param reason ُǏBÕbZ[WȂǁB
	 */
	void notifyDownloadListenerToAbort(Attachment attachment, String localPath,
			String reason) {

		attachment.status = Attachment.STATUS_ABORTED;
		attachment.localPath = localPath;

		if (attachment.listener != null) {
			attachment.listener.onAbort(attachment, reason);
		}

		if (downloadListeners == null) {
			return;
		}
		synchronized (downloadListeners) {
			Iterator itr = downloadListeners.iterator();
			while (itr.hasNext()) {
				DownloadListener listener = (DownloadListener) itr.next();
				listener.onAbort(attachment, reason);
			}
		}
	}

	/**
	 * _E[hXi[Ƀ_E[hꎞ~Ƃʒm܂B
	 * ([U[ɂLZ)
	 * @param attachment ꎞ~Ώ
	 * @param localPath ۑpX
	 */
	void notifyDownloadListenerToSuspend(Attachment attachment, String localPath) {

		attachment.status = Attachment.STATUS_SUSPENDED;
		attachment.localPath = localPath;

		if (DEBUG)
			logger.debug("notifyDownloadListenerToSuspend"); //$NON-NLS-1$

		if (attachment.listener != null) {
			attachment.listener.onSuspend(attachment);
		}
		if (downloadListeners == null) {
			return;
		}
		synchronized (downloadListeners) {
			Iterator itr = downloadListeners.iterator();
			while (itr.hasNext()) {
				DownloadListener listener = (DownloadListener) itr.next();
				listener.onSuspend(attachment);
			}
		}
	}

	/**
	 * <p>bZ[W̃oCgԂ܂B</p>
	 * <p>CX^XŐݒ肳ꂽGR[fBOgp܂B</p>
	 * @param message bZ[W
	 * @return oCg
	 */
	public int getMessageLength(String message) {
		if (message == null || message.length() == 0) {
			return 0;
		}
		try {
			return message.getBytes(ENCODING).length;
		} catch (UnsupportedEncodingException e) {
			logger.error("Cannot get byte length of message. encoding=" //$NON-NLS-1$
					+ ENCODING, e);
			return 0;
		}
	}

	/**
	 * XbhO[v擾܂B
	 * @return XbhO[v
	 */
	ThreadGroup getThreadGroup() {
		return this.threadGroup;
	}

	/**
	 * @see Messenger#getStatusMessage(UniqueMember)
	 */
	public void getStatusMessage(UniqueMember member) {
		//FIXME s݃bZ[W̎擾
	}

	/**
	 * Nݒ肩MessengeroChTCP̃AhXƃ|[gԍ擾܂B
	 * oChȂꍇ͗vfȂ̔z񂪕Ԃ܂B
	 * @param conf Nݒ
	 * @return oChAhXƃ|[gԍ̔zB
	 * @throws UnknownHostException 
	 */
	public InetSocketAddress[] getTcpBindAddress(LaunchConfiguration conf)
			throws UnknownHostException {
		return getAddressAndPort(conf);
	}

	/**
	 * Nݒ肩MessengeroChUDP̃AhXƃ|[gԍ擾܂B
	 * oChȂꍇ͗vfȂ̔z񂪕Ԃ܂B
	 * @param conf Nݒ
	 * @return oChAhXƃ|[gԍ̔zB
	 * @throws UnknownHostException 
	 */
	public InetSocketAddress[] getUdpBindAddress(LaunchConfiguration conf)
			throws UnknownHostException {
		return getAddressAndPort(conf);
	}

	/**
	 * Nݒ肩MessengeroChAhXƃ|[gԍ擾܂B
	 * oChȂꍇ͗vfȂ̔z񂪕Ԃ܂B
	 * @param conf Nݒ
	 * @return oChAhXƃ|[gԍ̔zB
	 */
	private InetSocketAddress[] getAddressAndPort(LaunchConfiguration conf) {
		int port = getPort(conf);
		InetAddress addr = getBindAddress(conf);
		InetSocketAddress[] addrs = new InetSocketAddress[1];
		addrs[0] = new InetSocketAddress(addr, port);
		return addrs;
	}

	/**
	 * @param utf8 UTF-8
	 * @return truêƂUTF-8BfalsêƂWindows-31JB
	 */
	static String getEncoding(boolean utf8) {
		return utf8 ? IPMessenger.ENCODING_UTF8 : IPMessenger.ENCODING;
	}

	/**
	 * _CAAbvIvV
	 * @return _CAAbvIvVgpƂtrueB
	 */
	boolean isOptionDialup() {
		Properties prop = this.configuration.getProperties();
		return PropertiesUtils.getBoolean(prop, DIALUP, false);
	}

	/**
	 * ÍgpȂ
	 * @return ÍgpȂꍇtrueB
	 */
	boolean isOptionUsePlain() {
		Properties prop = this.configuration.getProperties();
		return PropertiesUtils.getBoolean(prop, USE_PLAIN, false);
	}

	/**
	 * ÍNƂɐB
	 * @return ÍNƂɐꍇtrueB
	 */
	private boolean isOptionGenerateKeyEachTime() {
		Properties prop = this.configuration.getProperties();
		return PropertiesUtils.getBoolean(prop, GENERATE_KEY_EACH_TIME, true);
	}

	/**
	 * Unicodet@C𑗐MB
	 * @return Unicodet@C𑗐MꍇtrueB
	 */
	boolean isOptionSendUnicodeFilename() {
		Properties prop = this.configuration.getProperties();
		return PropertiesUtils.getBoolean(prop, SEND_UNICODE_FILENAME, true);
	}
}
