package petaclone.petarou;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import jp.takayuki.tools.PrefixExtensionFilter;
import jp.takayuki.tools.Tool;

/**
 * EPMt@Cǂ݂܂B
 * @author Takayuki Shoji
 */
public class PetaEpmReader {

	/**
	 * y^L̕R[hZbg
	 */
	private static final String CHARSET = "Windows-31J"; //$NON-NLS-1$

	/**
	 * EPMǂݍ݂܂B
	 * @param dir EPMۑfBNg
	 * @return Map(L[: t@C, l: y^L)BG[nullB
	 */
	public Map readFiles(File dir) {
		if (dir == null || !dir.isDirectory()) {
			return null;
		}
		PrefixExtensionFilter filter = new PrefixExtensionFilter("", ".EPM"); //$NON-NLS-1$ //$NON-NLS-2$
		File[] epms = dir.listFiles(filter);
		if (epms == null) {
			return null;
		}

		Map charactors = new HashMap();
		// t@CJăy^L擾
		for (int i = 0; i < epms.length; i++) {
			File f = epms[i];
			String cName = null;
			try {
				cName = getCharacterName(f);
			} catch (IOException e) {
				e.printStackTrace();
			}
			if (cName == null || cName.length() == 0) {
				continue;
			}
			charactors.put(f.getName(), cName);
		}
		return charactors;
	}

	/**
	 * 0x12#**EPETA_MOD1.1**#
	 */
	private static final byte[] SIGNATURE = new byte[] { 0x12, 0x23, 0x2a,
			0x2a, 0x45, 0x50, 0x45, 0x54, 0x41, 0x5f, 0x4d, 0x4f, 0x44, 0x31,
			0x2e, 0x31, 0x2a, 0x2a, 0x23 };

	/**
	 * y^L擾܂B
	 * @param f EPMt@C
	 * @return LN^
	 * @throws IOException 
	 */
	private String getCharacterName(File f) throws IOException {

		FileInputStream r = null;

		try {
			r = new FileInputStream(f);

			byte[] buf = new byte[SIGNATURE.length];

			// VOj`̓ǂݍ
			if (Tool.readComplete(r, buf) != SIGNATURE.length) {
				return null;
			}

			// VOj`̌
			for (int i = 0; i < buf.length; i++) {
				if (buf[i] != SIGNATURE[i]) {
					return null;
				}
			}

			// ǂݔ΂
			long skip = 14;
			long skipSize = 0;
			while (true) {
				long added = r.skip(skip - skipSize);
				if (added == -1) {
					return null;
				}
				skipSize += added;
				if (skipSize >= skip) {
					break;
				}
			}

			// pXJ̎擾
			buf = new byte[1];
			if (Tool.readComplete(r, buf) != 1) {
				return null;
			}
			int length = Tool.unsignedByteToInt(buf[0]);

			// y^L̎擾
			buf = new byte[length];
			if (Tool.readComplete(r, buf) != length) {
				return null;
			}

			return new String(buf, CHARSET);

		} finally {
			if (r != null)
				r.close();
		}
	}
}
