package petaclone.petarou;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.BindException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.InetAddressUtils;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Os;
import jp.takayuki.tools.Tool;
import jp.takayuki.tools.WindowsPathResolver;
import petaclone.AppPreferences;
import petaclone.Attachment;
import petaclone.ControlCreator;
import petaclone.Destination;
import petaclone.DownloadListener;
import petaclone.LaunchConfiguration;
import petaclone.LaunchException;
import petaclone.Member;
import petaclone.Message;
import petaclone.Messenger;
import petaclone.Protocol;
import petaclone.SendingResultListener;
import petaclone.ServerListener;
import petaclone.UniqueMember;

/**
 * @author Takayuki Shoji
 */
public class PetaMessenger implements Messenger {

	/** GR[fBO */
	final static String ENCODING = "Windows-31J"; //$NON-NLS-1$

	/** y^L̃vpeB */
	private static final String CHARACTER = "character"; //$NON-NLS-1$

	/** ftHg̃y^L */
	private static final String DEFAULT_CHARACTER = "DEFAULT.EPM"; //$NON-NLS-1$

	/** \̃vpeB */
	private static final String EXPRESSION = "expression"; //$NON-NLS-1$

	/**
	 * logger
	 */
	private static final Logger logger = Logger.getLogger(PetaMessenger.class);

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/**
	 * AvP[Vݒ
	 */
	private AppPreferences pref;

	/**
	 * CX^X̐ݒ
	 */
	LaunchConfiguration configuration;

	/**
	 * vgRB
	 */
	private static Protocol protocol;
	static {
		protocol = new Protocol();
		protocol.maxMessageLength = 1024;
		protocol.maxUserNameLength = 16;
		protocol.protocolName = mm.get("PetaMessenger.protocolName"); //$NON-NLS-1$
		protocol.signature = "peta"; //$NON-NLS-1$
	}

	/** oChAhX */
	InetAddress bindAddress;

	/**
	 * TCPT[o
	 */
	private PetaTCPServer tcp;

	/** TCPT[õXbh */
	private Thread tcpThread;

	/** UDPT[o */
	private PetaUDPServer udp;

	/** UDPT[õXbh */
	private Thread udpThread;

	/**
	 * T[oXi[
	 */
	private ServerListener serverListener;

	/** messageCount̃L[ */
	private final static String PETA_MESSAGE_COUNT = "peta.messageCount"; //$NON-NLS-1$

	/**
	 * ̃RXgN^
	 */
	public PetaMessenger() {
		//
	}

	/**
	 * VCX^X𐶐܂B
	 */
	public Messenger newInstance() {
		return new PetaMessenger();
	}

	/**
	 * AppPreferencesݒ肵܂B
	 * @param pref AvP[Vݒ
	 */
	public void setAppPreferences(AppPreferences pref) {
		this.pref = pref;
	}

	/**
	 * AvP[Vݒ擾܂B
	 * @return AvP[Vݒ
	 */
	AppPreferences getAppPreferences() {
		return pref;
	}

	/**
	 * CX^X̐ݒݒ肵܂B
	 */
	public void setConfiguration(LaunchConfiguration conf) {
		this.configuration = conf;
	}

	/**
	 * NIDԂ܂B
	 */
	public String getId() {
		if (configuration == null) {
			return null;
		}
		return configuration.getId();
	}

	/**
	 * TCPUDPgp\Ȃtrue
	 * @see petaclone.Messenger#getAlive()
	 */
	public boolean getAlive() {
		if (udp == null || tcp == null) {
			return false;
		}
		return tcp.getAlive() && udp.getAlive();
	}

	/**
	 * T[oXi[ݒ肵܂B
	 */
	public void setServerListener(ServerListener listener) {
		serverListener = listener;
	}

	/**
	 * T[oXi[ɃG[ʒm܂B
	 * @param message G[bZ[W
	 * @param t ThrowableIuWFNg
	 */
	void notifyError(String message, Throwable t) {
		serverListener.onError(message, t);
	}

	/**
	 * T[oXi[ɃbZ[W̓ʒm܂B
	 * @param message bZ[W
	 */
	void notifyMessage(Message message) {
		serverListener.onMessage(message);
	}

	/**
	 * T[oXi[ɊJʒm̓ʒm܂B
	 * @param message Jʒm
	 */
	void notifyOpenNotice(Message message) {
		serverListener.onOpenNotice(message);
	}

	/**
	 * T[oXi[Ƀo[̍폜ʒm܂B
	 * @param member 폜郁o[
	 */
	void notifyDeleteMember(Member member) {
		serverListener.deleteMember(member);
	}

	/**
	 * T[oXi[Ƀo[̒ǉʒm܂B
	 * @param member ǉ郁o[
	 */
	void notifyAddMember(Member member) {
		serverListener.addMember(member);
	}

	/**
	 * oChAhX擾܂B
	 * \PbgɎw肷ׂInetAddressƃoChAhXԍtB[hɐݒ肳܂B
	 * @param conf Nݒ
	 * @return oChAhX
	 */
	private InetAddress getBindAddress(LaunchConfiguration conf) {
		int bindAddressNumber = conf.getBindAddress();
		return InetAddressUtils.getLocalAddress(bindAddressNumber);
	}

	/**
	 * T[oJn܂B
	 * @throws LaunchException NO
	 * @see petaclone.Messenger#begin()
	 */
	public void begin(ThreadGroup gr) throws LaunchException {
		this.bindAddress = getBindAddress(this.configuration);

		boolean result = false;
		try {
			tcp = new PetaTCPServer(this);
			tcp.bind();

			udp = new PetaUDPServer(this);
			udp.bind();

			result = true;
		} catch (BindException e) {
			throw new LaunchException(mm
					.get("PetaMessenger.serverAddressIsUnavailable"), e, //$NON-NLS-1$
					this.configuration);

		} catch (SocketException e) {
			throw new LaunchException(mm
					.get("PetaMessenger.underlyingProtocolError"), e, //$NON-NLS-1$
					this.configuration);

		} catch (IOException e) {
			throw new LaunchException(mm.get("PetaMessenger.IOException"), e, //$NON-NLS-1$
					this.configuration);

		} finally {
			if (!result) {
				if (tcp != null) {
					tcp.quit();
				}
				if (udp != null) {
					udp.quit();
				}
			}
		}

		// TCPT[o̊Jn
		tcpThread = new Thread(gr, tcp);
		tcpThread.setName("Petarou TCP"); //$NON-NLS-1$
		tcpThread.start();
		// UDPT[o̊Jn
		udpThread = new Thread(gr, udp);
		udpThread.setName("Petarou UDP"); //$NON-NLS-1$
		udpThread.start();
	}

	/**
	 * T[oI܂B
	 * @see petaclone.Messenger#quit()
	 */
	public void quit() {
		udp.sendDelete();
		tcp.quit();
		udp.quit();
	}

	/**
	 * @see petaclone.Messenger#getProtocol()
	 */
	public Protocol getProtocol() {
		return protocol;
	}

	//
	// UDP
	//
	/**
	 * @see petaclone.Messenger#sendFind(boolean)
	 */
	public void sendFind(boolean addOnly) {
		udp.sendFind(addOnly);
	}

	/**
	 * @see petaclone.Messenger#sendRegister()
	 */
	public void sendRegister() {
		udp.sendRegister();
	}

	/**
	 * @see petaclone.Messenger#sendRename()
	 */
	public void sendRename() {
		udp.sendRename();
	}

	//
	// TCP
	//
	/**
	 * s݃[hʒmBʒm̓vgRT|[gĂȂ̂
	 * ȂB[h̓C擾̂łł͉ȂB
	 * @see petaclone.Messenger#sendAbsenceMode(boolean)
	 */
	public void sendAbsenceMode(boolean isAbsence) {
		// Ȃ
	}

	/**
	 * bZ[W𑗐M܂B
	 * @see petaclone.Messenger#sendMessage(Message, UniqueMember[], SendingResultListener)
	 */
	public void sendMessage(Message message, UniqueMember[] destinations,
			SendingResultListener listener) {

		Send send = new Send(message, destinations, listener);
		Thread th = new Thread(send);
		th.start();

	}

	/**
	 * bZ[W𑗐M܂B
	 * sݒʒmpB
	 * @param message sݒʒmbZ[W
	 * @param to Đ
	 * @throws IOException Đ悪Ȃꍇ
	 */
	void sendMessage(Message message, InetSocketAddress to) throws IOException {
		// bZ[WID𓾂
		message.id = getNextMessageCount();
		// f[^
		byte[] buffer = makeData(message);
		// M
		sendRaw(to, buffer);
	}

	/**
	 * MXbh<br>
	 * 1ʂM܂B
	 */
	private class Send implements Runnable {

		/** bZ[W */
		Message message;

		/** Đ */
		UniqueMember[] dest;

		/** MʃXi[ */
		SendingResultListener listener;

		/**
		 * RXgN^
		 * @param message M郁bZ[W
		 * @param destinations Đ
		 * @param listener MʃXi[
		 */
		Send(Message message, UniqueMember[] destinations,
				SendingResultListener listener) {
			this.message = message;
			this.dest = destinations;
			this.listener = listener;
		}

		public void run() {
			for (int i = 0; i < dest.length; i++) {
				InetSocketAddress addr = null;
				try {
					// f[^(idς̂Ŗ)
					message.id = getNextMessageCount();
					byte[] buffer = makeData(message);
					// M
					/*
					 * InetAddress̎擾
					 * 1. UDPT[oInetAddress擾
					 * 2. Rs[^InetAddress擾
					 * 
					 * 2̏ꍇ̓TulbgقȂƒʏ͎擾łȂB
					 */
					addr = udp.getIPAddress(dest[i].getIdentifier());
					if (addr == null) {
						// Rs[^
						InetAddress iaddr = InetAddress.getByName(dest[i]
								.getComputerName());
						// ło^Ă
						dest[i].setIpAddress(iaddr.getHostAddress());
						addr = new InetSocketAddress(iaddr, PetaTCPServer.port);

					} else {
						addr = new InetSocketAddress(addr.getAddress(),
								PetaTCPServer.port);
					}
					sendRaw(addr, buffer);
					Destination d = new Destination(configuration.getId(),
							dest[i].getOsUserName(), dest[i].getComputerName(),
							dest[i].getUserName(), dest[i].getIpAddress(), 0,
							true);
					listener.onSendResult(d, null);
				} catch (Exception e) {
					logger.warn("Failed to send." + addr, e); //$NON-NLS-1$
					// IOException, UnknownHostException, SocketTimeoutException
					Destination d = new Destination(configuration.getId(),
							dest[i].getOsUserName(), dest[i].getComputerName(),
							dest[i].getUserName(), dest[i].getIpAddress(), 0,
							false);
					listener.onSendResult(d, e);
				}
			}
			// MI
		}
	}

	/**
	 * vpeB̃bZ[WJEg+1ĕԂ܂B
	 * @return bZ[WJEg
	 */
	private synchronized int getNextMessageCount() {
		try {
			int result = Integer.parseInt(configuration.getProperty(
					PETA_MESSAGE_COUNT, "0")); //$NON-NLS-1$
			result++;
			configuration.setProperty(PETA_MESSAGE_COUNT, Integer
					.toString(result));
			return result;
		} catch (NumberFormatException e) {
			configuration.setProperty(PETA_MESSAGE_COUNT, "1"); //$NON-NLS-1$
			return 1;
		}
	}

	/**
	 * f[^𑗐M܂B
	 * @param addr MAhX
	 * @param rawData Mf[^
	 * @throws IOException o͗O
	 */
	private void sendRaw(InetSocketAddress addr, byte[] rawData)
			throws IOException {
		if (addr == null || rawData == null) {
			return;
		}
		Socket sock = null;
		OutputStream os = null;
		try {
			logger.info("Sending packet: addr=" + addr); //$NON-NLS-1$
			sock = new Socket();
			// 5bȓɃzXgȂƂ͗O
			sock.connect(addr, 5 * 1000);
			try {
				os = sock.getOutputStream();
				os.write(rawData);
			} finally {
				if (os != null)
					os.close();
			}
		} finally {
			if (sock != null)
				sock.close();
		}
	}

	/**
	 * @see petaclone.Messenger#sendOpenNotice(Message)
	 */
	public void sendOpenNotice(Message message) {
		// M悪NĂ邩mF
		if (!this.udp.isAliveMember(message.getIdentifier())) {
			return;
		}

		// Jʒmf[^
		Message msg = new Message();
		msg.messageType = Message.OPEN_NOTICE;
		msg.id = message.id;

		// Mf[^
		byte[] buffer = makeData(msg);
		// M
		try {
			InetSocketAddress addr = new InetSocketAddress(
					message.getAddress(), message.port);
			sendRaw(addr, buffer);
			logger.info("Sent open notice." + msg.id); //$NON-NLS-1$
		} catch (IOException e) {
			logger.error("Failed to send open notice.", e); //$NON-NLS-1$
		}
	}

	/**
	 * bZ[W̃f[^܂B
	 * @param msg bZ[W
	 * @return MoCif[^
	 */
	private byte[] makeData(Message msg) {
		byte[] result = null;
		if (msg.messageType == Message.MESSAGE) {
			result = makeMessageData(msg);
		} else {
			result = makeOpenNoticeData(msg);
		}
		return result;
	}

	/**
	 * Mf[^쐬܂B(Jʒmp)
	 * @param msg Jʒm
	 * @return MoCif[^
	 */
	private byte[] makeOpenNoticeData(Message msg) {
		byte[] result = new byte[315];
		byte[] buffer;

		// 02 00 00 00
		result[0] = 2;
		// ĂVA
		System.arraycopy(intToLittleEndianByte(msg.id), 0, result, 4, 4);

		try {
			// y^[U[(ő咷16)
			buffer = pref.getUserName(protocol.maxUserNameLength, ENCODING)
					.getBytes(ENCODING);
			System.arraycopy(buffer, 0, result, 8, buffer.length);
			// IPAhX
			if (bindAddress != null) {
				buffer = bindAddress.getHostAddress().getBytes(ENCODING);
			} else {
				/*
				 * IPAhXƂɖ]܂IPAhX擾łȂꍇB
				 * ]܂ȂAhXƂĂ127.0.0.1ȂǁB
				 */
				buffer = InetAddress.getLocalHost().getHostAddress().getBytes(
						ENCODING);
			}

			System.arraycopy(buffer, 0, result, 41, buffer.length);
			// ̃Rs[^
			if (bindAddress != null) {
				buffer = bindAddress.getHostName().toUpperCase().getBytes(
						ENCODING);
			} else {
				buffer = InetAddress.getLocalHost().getHostName().toUpperCase()
						.getBytes(ENCODING);
			}

			System.arraycopy(buffer, 0, result, 58, buffer.length);
		} catch (UnsupportedEncodingException e) {
			logger.error("Charactor encoding \"" + ENCODING //$NON-NLS-1$
					+ "\" is not supported on this computer."); //$NON-NLS-1$
		} catch (UnknownHostException e) {
			e.printStackTrace();
		}
		return result;
	}

	/**
	 * Mf[^쐬܂B(bZ[Wp)B
	 * : IPAhXƃRs[^͓IɎ擾B
	 * @param msg bZ[W
	 * @return MoCif[^
	 */
	private byte[] makeMessageData(Message msg) {
		byte[] result = new byte[1639];
		byte[] buffer = null;

		// 01 00 00 00
		result[0] = 1;

		// VA
		System.arraycopy(intToLittleEndianByte(msg.id), 0, result, 4, 4);

		try {
			// ݒ肵L̃t@C
			buffer = configuration.getProperty(CHARACTER, DEFAULT_CHARACTER)
					.getBytes(ENCODING);
			System.arraycopy(buffer, 0, result, 8, buffer.length);

			// ̃[U[
			buffer = pref.getUserName(protocol.maxUserNameLength, ENCODING)
					.getBytes(ENCODING);
			System.arraycopy(buffer, 0, result, 265, buffer.length);
			// IPAhX
			if (bindAddress != null) {
				buffer = bindAddress.getHostAddress().getBytes(ENCODING);
			} else {
				InetAddress addr = InetAddressUtils.getLocalAddress(0);
				buffer = addr.getHostAddress().getBytes(ENCODING);
			}

			System.arraycopy(buffer, 0, result, 298, buffer.length);
			// ̃Rs[^
			if (bindAddress != null) {
				buffer = bindAddress.getHostName().toUpperCase().getBytes(
						ENCODING);
			} else {
				String name = InetAddressUtils.getHostName();
				if (name != null) {
					buffer = name.getBytes(ENCODING);
				}
				//buffer =
				//	InetAddress
				//		.getLocalHost()
				//		.getHostName()
				//		.toUpperCase()
				//		.getBytes(
				//		ENCODING);
			}

			System.arraycopy(buffer, 0, result, 315, buffer.length);

			// t
			Calendar calendar = new GregorianCalendar();
			//msg.setTime(calendar); // ۑp

			buffer = dateToString(calendar).getBytes(ENCODING);
			System.arraycopy(buffer, 0, result, 572, buffer.length);

			// bZ[W
			StringBuffer sb = new StringBuffer(msg.message);
			Tool.replaceAll(sb, "\r\n", "\n"); //$NON-NLS-1$ //$NON-NLS-2$
			Tool.replaceAll(sb, "\r", "\n"); //$NON-NLS-1$ //$NON-NLS-2$
			Tool.replaceAll(sb, "\n", "\r\n"); //$NON-NLS-1$ //$NON-NLS-2$
			buffer = sb.toString().getBytes(ENCODING);
			if (buffer.length > 1024) {
				System.arraycopy(buffer, 0, result, 605, 1024);
			} else {
				System.arraycopy(buffer, 0, result, 605, buffer.length);
			}

			// F[1631],[1632],[1633]
			if (msg.color != null) {
				result[1631] = (byte) msg.color[0];
				result[1632] = (byte) msg.color[1];
				result[1633] = (byte) msg.color[2];
			}

		} catch (UnsupportedEncodingException e) {
			logger.error("Charactor encoding \"" + ENCODING //$NON-NLS-1$
					+ "\" is not supported on this computer."); //$NON-NLS-1$
		}

		// \
		result[1635] = (byte) (configuration.getProperty(EXPRESSION, "0") //$NON-NLS-1$
				.charAt(0) - '0');

		return result;
	}

	/**
	 * t𕶎ɕϊBy^낤vgRpB
	 * @param c ϊtݒ肵J_[IuWFNg
	 * @return t("yyyy/MM/dd HH:mm:ss")
	 */
	public static String dateToString(Calendar c) {
		SimpleDateFormat f = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss"); //$NON-NLS-1$
		return f.format(c.getTime());
	}

	/**
	 * intgGfBAbyte[]ɕϊ܂B
	 * @param integer int^̐l
	 * @return 4oCg̔z
	 */
	public static byte[] intToLittleEndianByte(int integer) {
		byte[] result = new byte[4];

		// ʂႢ >>>ȂEVtg
		result[0] = new Integer(integer).byteValue();
		result[1] = new Integer(integer >>> 8).byteValue();
		result[2] = new Integer(integer >>> 16).byteValue();
		result[3] = new Integer(integer >>> 24).byteValue();

		return result;
	}

	/**
	 * ݒ_CAÕRg[zu܂B
	 * @see petaclone.Messenger#createConfigControl(ControlCreator)
	 */
	public void createConfigControl(ControlCreator creator) {
		// \
		final String expressions[] = {
				mm.get("PetaMessenger.expression.normal"), //$NON-NLS-1$
				mm.get("PetaMessenger.expression.smiling"), //$NON-NLS-1$
				mm.get("PetaMessenger.expression.angry"), //$NON-NLS-1$
				mm.get("PetaMessenger.expression.crying"), //$NON-NLS-1$
				mm.get("PetaMessenger.expression.puzzled"), //$NON-NLS-1$
				mm.get("PetaMessenger.expression.surprised"), //$NON-NLS-1$
				mm.get("PetaMessenger.expression.sleeping"), //$NON-NLS-1$
				"" }; //$NON-NLS-1$
		final String exValues[] = { "0", "1", "2", "3", "4", "5", "6", "7" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$
		creator.createComboBox(EXPRESSION, //
				mm.get("PetaMessenger.expression.label"), //$NON-NLS-1$
				expressions, exValues, "0", //$NON-NLS-1$
				true);

		// y^LfBNg
		String defaultDir;
		if (Os.isWindows()) {
			WindowsPathResolver r = new WindowsPathResolver();
			File app = r.getApplicationData();
			File dir = new File(app, "PETA"); //$NON-NLS-1$
			defaultDir = dir.getAbsolutePath();
		} else {
			defaultDir = "PETA"; //$NON-NLS-1$
		}
		creator.createDirectorySelector("epm.dir", //$NON-NLS-1$
				mm.get("PetaMessenger.petacharDir.label"), //$NON-NLS-1$
				defaultDir, null, true, true);
		creator.createLabel(mm.get("PetaMessenger.toUpdateList.label")); //$NON-NLS-1$

		String epmDir;
		if (configuration != null) {
			epmDir = configuration.getProperty("epm.dir", defaultDir); //$NON-NLS-1$
		} else {
			epmDir = defaultDir;
		}
		PetaEpmReader reader = new PetaEpmReader();
		Map map = reader.readFiles(new File(epmDir));
		int size = 1;
		if (map != null) {
			size += map.size();
		}
		String[] characters = new String[size];
		String[] charNames = new String[size];
		characters[0] = "DEFAULT.EPM"; //$NON-NLS-1$
		charNames[0] = mm.get("PetaMessenger.petachar.default"); //$NON-NLS-1$

		if (map != null) {
			Iterator itr = map.entrySet().iterator();
			int i = 1;
			while (itr.hasNext()) {
				Entry entry = (Entry) itr.next();
				characters[i] = (String) entry.getKey();
				charNames[i] = (String) entry.getValue();
				i++;
			}
		}

		creator.createComboBox(CHARACTER, //
				mm.get("PetaMessenger.petacharList.label"), //$NON-NLS-1$
				charNames, characters, DEFAULT_CHARACTER, true);
		creator.createSeparator();
		creator.createLabel(mm.get("PetaMessenger.portInfo.label")); //$NON-NLS-1$
		creator.createSeparator();
		creator.createLabel(mm.get("PetaMessenger.original.label")); //$NON-NLS-1$
	}

	/**
	 * _E[hXi[ǉ܂B
	 */
	public void addDownloadListener(DownloadListener listener) {
		// vgRT|[gĂȂ
	}

	/**
	 * _E[hXi[폜܂B
	 */
	public void removeDownloadListener(DownloadListener listener) {
		// vgRT|[gĂȂ
	}

	/**
	 * Ytt@Cۑ܂B
	 * vgRYtt@CT|[gĂȂ߁A
	 * s܂B
	 */
	public void saveAttachment(Message message, Attachment attachment, File dest) {
		// vgRT|[gĂȂ
	}

	/**
	 * Ytt@C̃_E[hLZ܂B
	 * vgRYtt@CT|[gĂȂ߁A
	 * s܂B
	 */
	public void cancelSavingAttachment(Attachment attachment) {
		// vgRT|[gĂȂ
	}

	/**
	 * <p>bZ[W̃oCgԂ܂B</p>
	 * <p>CX^XŐݒ肳ꂽGR[fBOgp܂B</p>
	 * @param message bZ[W
	 * @return oCg
	 */
	public int getMessageLength(String message) {
		if (message == null || message.length() == 0) {
			return 0;
		}
		try {
			return message.getBytes(ENCODING).length;
		} catch (UnsupportedEncodingException e) {
			logger.error("Charactor encoding \"" + ENCODING //$NON-NLS-1$
					+ "\" is not supported on this computer."); //$NON-NLS-1$
			return 0;
		}
	}

	/**
	 * @param message j郁bZ[W
	 */
	public void releaseAttachments(Message message) {
		// vgRT|[gĂȂ
	}

	/**
	 * @see Messenger#getStatusMessage(UniqueMember)
	 */
	public void getStatusMessage(UniqueMember member) {
		// vgRT|[gĂȂ
	}

	/**
	 * Nݒ肩MessengeroChTCP̃AhXƃ|[gԍ擾܂B
	 * oChȂꍇ͗vfȂ̔z񂪕Ԃ܂B
	 * @param conf Nݒ
	 * @return oChAhXƃ|[gԍ̔zB
	 * @throws UnknownHostException 
	 */
	public InetSocketAddress[] getTcpBindAddress(LaunchConfiguration conf)
			throws UnknownHostException {
		return getAddressAndPort(conf, PetaTCPServer.port);
	}

	/**
	 * Nݒ肩MessengeroChUDP̃AhXƃ|[gԍ擾܂B
	 * oChȂꍇ͗vfȂ̔z񂪕Ԃ܂B
	 * @param conf Nݒ
	 * @return oChAhXƃ|[gԍ̔zB
	 * @throws UnknownHostException 
	 */
	public InetSocketAddress[] getUdpBindAddress(LaunchConfiguration conf)
			throws UnknownHostException {
		return getAddressAndPort(conf, PetaUDPServer.port);
	}

	/**
	 * Nݒ肩MessengeroChUDP̃AhXƃ|[gԍ擾܂B
	 * oChȂꍇ͗vfȂ̔z񂪕Ԃ܂B
	 * @param conf Nݒ
	 * @param port |[gԍ
	 * @return oChAhXƃ|[gԍ̔zB
	 */
	private InetSocketAddress[] getAddressAndPort(LaunchConfiguration conf,
			int port) {
		InetAddress addr = getBindAddress(conf);
		InetSocketAddress[] addrs = new InetSocketAddress[1];
		addrs[0] = new InetSocketAddress(addr, port);
		return addrs;
	}
}
