package petaclone.petarou;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.BindException;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Tool;
import petaclone.AppPreferences;
import petaclone.Message;
import petaclone.RuntimeProperties;

/**
 * @author Takayuki Shoji
 */
public class PetaTCPServer implements Runnable {

	/**
	 * logger
	 */
	private static final Logger logger = Logger.getLogger(PetaTCPServer.class);

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();
	
	/** |[gԍ */
	static final int port = 39377;

	/** NȂtrue */
	private boolean isAlive;

	/** T[o\Pbg */
	private ServerSocket server;

	/** bZW[ */
	private PetaMessenger messenger;

	/**
	 * sR[h
	 */
	private static final String EOL = System.getProperty("line.separator"); //$NON-NLS-1$

	/**
	 * RXgN^
	 * @param messenger bZW[
	 */
	PetaTCPServer(PetaMessenger messenger) {
		this.messenger = messenger;
	}

	/**
	 * T[oĂ邩Ԃ܂B
	 * @return Ăꍇtrue
	 */
	synchronized boolean getAlive() {
		return isAlive;
	}

	/**
	 * T[oɏIʒm܂B
	 */
	void quit() {
		isAlive = false;
		close();
	}

	/**
	 * @see java.lang.Runnable#run()
	 */
	public void run() {

		try {
			while (isAlive) {
				Socket socket = null;
				try {
					socket = server.accept();
				} catch (IOException e) {
					if (isAlive) {
						logger.error("IOException occurred during accept().", e); //$NON-NLS-1$
						messenger.notifyError(//
								mm.get("PetaTCPServer.IOException.accept"), e); //$NON-NLS-1$
					}
					if (server.isBound() && !server.isClosed()) {
						continue;
					}
					return;
				}

				byte[] buffer = readStream(socket);
				if (buffer == null) {
					continue;
				}
				// Mf[^̏
				handleReceiveData(buffer,
						(InetSocketAddress) socket.getRemoteSocketAddress());
			}
			// T[o\Pbg
		} finally {
			isAlive = false;
			close();
		}
	}

	/**
	 * oCh܂B
	 * @throws BindException 
	 * @throws SocketException 
	 * @throws IOException 
	 */
	void bind() throws BindException, SocketException, IOException {
		server = new ServerSocket(port, -1, messenger.bindAddress);
		isAlive = true;
		logger.info("Bound: addr=" + messenger.bindAddress + ", port=" + port); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * T[o\Pbg܂B
	 */
	private void close() {
		try {
			if (server != null && !server.isClosed()) {
				server.close();
			}
		} catch (IOException e) {
			logger.error("Exception occurred during server.close().", e); //$NON-NLS-1$
		}
	}

	/**
	 * ҋ@ƎMB
	 * @param socket \Pbg
	 * @return Mf[^
	 */
	private byte[] readStream(Socket socket) {

		// M
		byte[] buffer = null;
		InputStream is = null;
		try {
			is = socket.getInputStream();
			buffer = new byte[1639]; //1639
			int n;
			int read = 0;
			while ((n = is.read(buffer, read, buffer.length - read)) != -1) {
				read += n;
				if (read >= buffer.length) {
					break;
				}
			}
			return checkLength(buffer, read);

		} catch (IOException e) {
			messenger.notifyError(//
					mm.get("PetaTCPServer.IOException.receive"), e); //$NON-NLS-1$
			//e.printStackTrace();
			return null;
		} finally {
			try {
				if (is != null) {
					is.close();
				}
			} catch (IOException e) {
				logger.error("Exception occurred " //$NON-NLS-1$
						+ "while closing socket input stream.", e); //$NON-NLS-1$
			}
			try {
				if (socket != null) {
					socket.close();
				}
			} catch (IOException e) {
				logger.error("Exception occurred while closing socket.", e); //$NON-NLS-1$
			}
		}
	}

	/**
	 * obt@ɓǂݍ񂾃f[^ƒmF܂B
	 * @param buffer obt@
	 * @param read obt@ɓǂݍ񂾒
	 * @return ȏꍇbufferAُȏꍇnullB
	 */
	private byte[] checkLength(byte[] buffer, int read) {
		if (read < 315) {
			// f[^Z
			logger.warn("Received data is too short."); //$NON-NLS-1$
			return null;
		}
		int type = Tool.byteToInt(buffer, 0);
		switch (type) {
		case Message.OPEN_NOTICE:
			if (read < 315) {
				return null;
			}
			break;
		case Message.MESSAGE:
			if (read < 1639) {
				logger.warn("Received message is too short."); //$NON-NLS-1$
				return null;
			}
			break;
		default:
			logger.warn("Received message type is incorrect: " + type); //$NON-NLS-1$
			return null;
		}
		return buffer;
	}

	/**
	 * Mf[^܂B
	 * ʏ̃bZ[WAsݒʒmAJʒmB
	 * @param buffer Mf[^B
	 * @param remoteAddress MAhX
	 */
	private void handleReceiveData(byte[] buffer, InetSocketAddress remoteAddress) {
		// Mf[^̏
		Message msg = analyzeRawData(buffer);
		if (msg == null) {
			return;
		}
		// pPbgɋLڂĂIPAhXۂ̃[gIPAhXŏ㏑
		msg.ipAddress = remoteAddress.getAddress().getHostAddress();
		// 
		if (msg.messageType == Message.MESSAGE) {
			messenger.notifyMessage(msg);
			// s݂Ȃsݒʒm𑗂

			AppPreferences pref = messenger.getAppPreferences();
			RuntimeProperties rp = pref.getRuntime();
			if (!rp.isAbsence()) {
				return;
			}

			Message absence = createAbsenceMessage(pref);

			try {
				// M̃T[o|[gɑM
				InetSocketAddress addr = new InetSocketAddress(
						msg.getAddress(), msg.port);
				messenger.sendMessage(absence, addr);
			} catch (IOException e) {
				logger.error("Failed to send absence notice.", e); //$NON-NLS-1$
			}
		} else if (msg.messageType == Message.OPEN_NOTICE) {
			messenger.notifyOpenNotice(msg);
		}
	}

	/**
	 * s݃bZ[W쐬܂B
	 * @param pref AvP[Vݒ
	 * @return Message
	 */
	private Message createAbsenceMessage(AppPreferences pref) {
		Message absence = new Message();
		// VA̓[
		absence.id = 0;
		// LN^[Ȃ
		absence.setExt("peta.character", ""); //$NON-NLS-1$ //$NON-NLS-2$
		// 
		absence.setForegroundBlue();
		absence.message = pref.getRuntime().getAbsenceMessage();
		return absence;
	}

	/**
	 * ͂f[^͂܂B
	 * @param rawData oCif[^
	 * @return bZ[W
	 */
	private Message analyzeRawData(byte[] rawData) {
		Message msg = new Message();
		msg.initialSealed();
		msg.launchId = messenger.getId();
		msg.port = port; // Œ

		if (rawData == null || rawData.length < 315) {
			return null;
		}

		msg.messageType = Tool.byteToInt(rawData, 0);
		try {
			if (msg.messageType == Message.OPEN_NOTICE) {
				// JmF

				// VAԍ
				msg.id = Tool.byteToInt(rawData, 4);

				byte[] buffer = new byte[20];
				byte[] clear = new byte[20]; // obt@p

				// y^[U[(16)
				System.arraycopy(rawData, 8, buffer, 0, 16);
				msg.userName = new String(buffer, 0, Tool.lengthToZero(buffer,
						0, 16), PetaMessenger.ENCODING);

				// IPAhX(15)
				System.arraycopy(clear, 0, buffer, 0, buffer.length);
				System.arraycopy(rawData, 41, buffer, 0, 15);
				msg.ipAddress = new String(buffer, 0, Tool.lengthToZero(buffer,
						0, 15), PetaMessenger.ENCODING);

				// Rs[^(20)
				System.arraycopy(clear, 0, buffer, 0, buffer.length);
				System.arraycopy(rawData, 58, buffer, 0, 20);
				msg.computerName = new String(buffer, 0, Tool.lengthToZero(
						buffer, 0, 20), PetaMessenger.ENCODING);
				msg.computerName = msg.computerName.toUpperCase();
				// M: 
				msg.setTime(Calendar.getInstance());

			} else if (msg.messageType == Message.MESSAGE) {
				// ʏ̃bZ[W
				if (rawData.length < 1639) {
					return null;
				}
				// VAԍ
				msg.id = Tool.byteToInt(rawData, 4);

				byte[] buffer = new byte[1024];
				byte[] clear = new byte[1024]; // obt@p

				// L̃t@C(257AԂBC邯ǁB)
				System.arraycopy(rawData, 8, buffer, 0, 257);
				msg.setExt("peta.character", new String(buffer, 0, Tool //$NON-NLS-1$
						.lengthToZero(buffer, 0, 257), PetaMessenger.ENCODING));

				// y^[U[(16)
				System.arraycopy(clear, 0, buffer, 0, buffer.length);
				System.arraycopy(rawData, 265, buffer, 0, 16);
				msg.userName = new String(buffer, 0, Tool.lengthToZero(buffer,
						0, 16), PetaMessenger.ENCODING);

				// IPAhX(15)
				System.arraycopy(clear, 0, buffer, 0, buffer.length);
				System.arraycopy(rawData, 298, buffer, 0, 15);
				msg.ipAddress = new String(buffer, 0, Tool.lengthToZero(buffer,
						0, 15), PetaMessenger.ENCODING);

				// Rs[^(20)
				System.arraycopy(clear, 0, buffer, 0, buffer.length);
				System.arraycopy(rawData, 315, buffer, 0, 20);
				msg.computerName = new String(buffer, 0, Tool.lengthToZero(
						buffer, 0, 20), PetaMessenger.ENCODING);
				msg.computerName = msg.computerName.toUpperCase();

				// (19)
				System.arraycopy(clear, 0, buffer, 0, buffer.length);
				System.arraycopy(rawData, 572, buffer, 0, 19);
				String tmpTime = new String(buffer, 0, Tool.lengthToZero(
						buffer, 0, 19), PetaMessenger.ENCODING);

				// y^낤`Calendarɕϊ
				// Message#setTime(String time)łp\ł͂邪A
				// `IɂŕϊB
				// ͎Ɠ^C][Ɖ肷
				try {
					SimpleDateFormat format = new SimpleDateFormat(
							"yyyy/MM/dd HH:mm:ss"); //$NON-NLS-1$
					Calendar calendar = Calendar.getInstance();
					calendar.setTime(format.parse(tmpTime));
					msg.setTime(calendar);
				} catch (ParseException e) {
					// Mɂ
					msg.setTime(Calendar.getInstance());
				}

				// {(1024)
				System.arraycopy(clear, 0, buffer, 0, buffer.length);
				System.arraycopy(rawData, 605, buffer, 0, 1024);
				String message = new String(buffer, 0, Tool.lengthToZero(
						buffer, 0, 1024), PetaMessenger.ENCODING);
				if (!EOL.equals("\r\n")) { //$NON-NLS-1$
					msg.message = Tool.replaceAll(message, "\r\n", EOL); //$NON-NLS-1$
				} else {
					msg.message = message;
				}

				// F((ftHg)Ȃݒ肷)
				int[] tmp = new int[3];
				tmp[0] = Tool.unsignedByteToInt(rawData[1631]);
				tmp[1] = Tool.unsignedByteToInt(rawData[1632]);
				tmp[2] = Tool.unsignedByteToInt(rawData[1633]);
				if (tmp[0] != 0 || tmp[1] != 0 || tmp[2] != 0) {
					msg.color = tmp;
				}

				// \
				msg.setExt("peta.expression", Integer.toString(Tool.byteToInt( //$NON-NLS-1$
						rawData, 1635)));

			} else {
				// sȃf[^
				return msg;
			}
		} catch (UnsupportedEncodingException e) {
			logger.warn(e);
			return null;
		}
		return msg;
	}
}
