package petaclone.petarou;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.BindException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetSocketAddress;
import java.net.NoRouteToHostException;
import java.net.SocketException;
import java.util.HashMap;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.InetAddressUtils;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Tool;
import petaclone.AppPreferences;
import petaclone.Member;
import petaclone.RuntimeProperties;

/**
 * @author Takayuki Shoji
 */
public class PetaUDPServer implements Runnable {

	/**
	 * logger
	 */
	private static final Logger logger = Logger.getLogger(PetaUDPServer.class);

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/** fobOtO */
	private static final boolean DEBUG = false;

	/** NȂtrue */
	private boolean isAlive;

	/** JE^ */
	private static int count = 1;

	/** f[^O\Pbg */
	private DatagramSocket socket = null;

	/** |[gԍ */
	static final int port = 39376;

	/** bZW[ */
	private PetaMessenger messenger;

	/**
	 * UDPbZ[W̎Mpobt@TCY
	 */
	private final static int MAX_MESSAGE_LENGTH = 80;

	/**
	 * o[ǗpB<br>
	 * L[: Member#getIdentifier()<br>
	 * l: PetaUDPMessage
	 */
	private HashMap members;

	/**
	 * RXgN^
	 * @param messenger bZW[
	 */
	PetaUDPServer(PetaMessenger messenger) {
		this.messenger = messenger;
		this.members = new HashMap();
	}

	/**
	 * T[oĂ邩𒲂ׂ܂B
	 * @return Ăꍇtrue
	 */
	synchronized boolean getAlive() {
		return isAlive;
	}

	/**
	 * o[NĂ邩ׂ܂B
	 * @param key o[ʎq
	 * @return NĂꍇtrue
	 */
	boolean isAliveMember(String key) {
		return members.containsKey(key);
	}

	/**
	 * T[oJn
	 * @see java.lang.Runnable#run()
	 */
	public void run() {

		byte[] buffer = new byte[MAX_MESSAGE_LENGTH];
		DatagramPacket packet = new DatagramPacket(buffer, buffer.length);
		try {
			while (isAlive) {
				try {
					socket.receive(packet);
				} catch (IOException e) {
					if (isAlive) {
						logger.error("I/O exception occurred " //$NON-NLS-1$
								+ "while waiting or receiving.", e); //$NON-NLS-1$
						messenger.notifyError(//
								mm.get("PetaUDPServer.IOException.receive"), e); //$NON-NLS-1$
						continue;
					}
					break;
				}

				// ̃u[hLXgɔȂ悤ɂ
				if (!messenger.configuration.acceptSelfBroadcast()
						&& InetAddressUtils.isLocalAddress(packet.getAddress())) {
					if (DEBUG)
						logger.debug("Broadcast from local address."); //$NON-NLS-1$
					continue;
				}

				// obt@̒𒲐߂AUDPMessage쐬
				byte[] buffer2 = new byte[packet.getLength()];
				System.arraycopy(buffer, 0, buffer2, 0, buffer2.length);
				PetaUDPMessage udpMsg = new PetaUDPMessage(buffer2,
						new InetSocketAddress(packet.getAddress(), packet
								.getPort()));

				// 
				switch (udpMsg.messageType) {
				case PetaUDPMessage.REGISTER:
					// o^ԓēo^
					InetSocketAddress[] addrs = new InetSocketAddress[] { udpMsg.address };
					send(PetaUDPMessage.REPLY, addrs);
					// breakȂ

				case PetaUDPMessage.REPLY:
					// ԓo^
					synchronized (members) {
						if (members.containsKey(udpMsg.member.getIdentifier())) {

							// 폜Ēʒm
							members.remove(udpMsg.member.getIdentifier());
							messenger.notifyDeleteMember(udpMsg.member);
							// o^Ȃ
						}
						// o^Ēʒm
						members.put(udpMsg.member.getIdentifier(), udpMsg);
					}

					messenger.notifyAddMember(udpMsg.member);
					break;

				case PetaUDPMessage.DELETE:
					// 폜
					synchronized (members) {
						if (!members.containsKey(udpMsg.member.getIdentifier())) {
							break;
						}
						members.remove(udpMsg.member.getIdentifier());
					}
					messenger.notifyDeleteMember(udpMsg.member);
					break;

				default:
					break;
				}
			}
		} finally {
			isAlive = false;
			close();
		}
	}

	/**
	 * oCh܂B
	 * @throws BindException
	 * @throws SocketException
	 * @throws IOException
	 */
	void bind() throws BindException, SocketException, IOException {
		this.socket = new DatagramSocket(port, messenger.bindAddress);
		isAlive = true;
		logger.info("Bound: addr=" + messenger.bindAddress + ", port=" + port); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * \PbgN[Y܂B
	 */
	private void close() {
		if (socket != null && !socket.isClosed())
			socket.close();
	}

	/**
	 * Method send.
	 * @param messageType bZ[W^CvB
	 * 	PetaUDPMessage.DELETE, 
	 * 	PetaUDPMessage.REGISTER, 
	 * 	PetaUDPMessage.REPLY
	 * @param addrs M
	 */
	private void send(int messageType, InetSocketAddress[] addrs) {
		// M镶p
		byte[] msg = getString(messageType);
		if (msg == null) {
			return;
		}
		for (int i = 0; i < addrs.length; i++) {
			// UDPőM
			try {
				DatagramPacket packet = new DatagramPacket(msg, msg.length,
						addrs[i]);
				socket.send(packet);
			} catch (NoRouteToHostException e) {
				e.printStackTrace();
				continue;
			} catch (IOException e) {
				e.printStackTrace();
				continue;
			}
			count++;
		}
	}

	/**
	 * Method getString.Mp̕擾܂B
	 * @param mType bZ[W^Cv 
	 * 	PetaUDPMessage.DELETE, 
	 * 	PetaUDPMessage.REGISTER, 
	 * 	PetaUDPMessage.REPLY
	 * @return byte[] Mp̃f[^
	 */
	private byte[] getString(int mType) {
		AppPreferences pref = messenger.getAppPreferences();
		RuntimeProperties rp = pref.getRuntime();

		StringBuffer sb = new StringBuffer("1:").append(count); //$NON-NLS-1$

		sb.append(':').append(rp.getLoginNameEscaped());
		sb.append(':').append(InetAddressUtils.getHostNameEscaped());

		sb.append(':').append(mType).append(':');
		try {
			// 폜ȊÔƂɃ[U[
			if (mType != PetaUDPMessage.DELETE) {
				sb.append(pref.getUserName(
						messenger.getProtocol().maxUserNameLength,
						PetaMessenger.ENCODING));
			}
			return sb.toString().getBytes(PetaMessenger.ENCODING);
		} catch (UnsupportedEncodingException e) {
			return null;
		}
	}

	/**
	 * o[ʎqɑ΂AhXԂ܂B
	 * ݂ȂꍇnullԂ܂B
	 * @param id o[ʎq
	 * @return AhX
	 */
	InetSocketAddress getIPAddress(String id) {
		PetaUDPMessage udpMsg = null;
		synchronized (members) {
			if (!members.containsKey(id)) {
				return null;
			}
			udpMsg = (PetaUDPMessage) members.get(id);
		}

		return udpMsg.address;
	}

	/**
	 * o^
	 */
	void sendRegister() {
		AppPreferences pref = messenger.getAppPreferences();
		InetSocketAddress[] addrs = messenger.configuration
				.getBroadcastAddresses(pref, (InetSocketAddress) socket
						.getLocalSocketAddress());

		send(PetaUDPMessage.REGISTER, addrs);
	}

	/**
	 * 폜
	 */
	void sendDelete() {
		AppPreferences pref = messenger.getAppPreferences();
		InetSocketAddress[] addrs = messenger.configuration
				.getBroadcastAddresses(pref, (InetSocketAddress) socket
						.getLocalSocketAddress());
		send(PetaUDPMessage.DELETE, addrs);
	}

	/**
	 * Č
	 * @param addOnly ̃XgNAȂꍇtrue
	 */
	void sendFind(boolean addOnly) {
		AppPreferences pref = messenger.getAppPreferences();
		InetSocketAddress[] addrs = messenger.configuration
				.getBroadcastAddresses(pref, (InetSocketAddress) socket
						.getLocalSocketAddress());
		if (!addOnly) {
			synchronized (members) {
				members.clear();
			}
		}
		send(PetaUDPMessage.DELETE, addrs);
		send(PetaUDPMessage.REGISTER, addrs);
	}

	/**
	 * [U[̕ύXʒmB
	 * 폜Ƃɂ͌ÂO͕svȂ̂łłɐݒ肳ꂽOœo^sB
	 * ČƓBAǉ݂̂ŌB
	 */
	void sendRename() {
		sendFind(true);
	}

	/**
	 * T[oɏIʒm܂B
	 */
	void quit() {
		this.isAlive = false;
		// Iɕ
		close();
	}

	/**
	 * UDPbZ[W
	 */
	private class PetaUDPMessage {
		/** bZ[W^Cv */
		int messageType;

		/** messageType萔 */
		static final int REGISTER = 1; // (NAČ)o^ĕԎ傤

		/** messageType萔 */
		static final int DELETE = 2; // (IAČ)폜Ă

		/** messageType萔 */
		static final int REPLY = 3; // (ԓ)o^Ă

		/**
		 * [U[
		 */
		Member member;

		/**
		 * IPAhXƃ|[gԍ
		 */
		InetSocketAddress address;

		/**
		 * RXgN^
		 * @param rawData oCif[^
		 * @param address MƂIPAhX
		 */
		PetaUDPMessage(byte[] rawData, InetSocketAddress address) {
			this.address = address;
			member = new Member();
			// y^낤vgR̓O[vT|[gĂȂ
			member.group = null;
			// NID
			member.launchId = messenger.getId();
			// t@CMT|[gĂȂ
			member.canReceiveFile = false;

			member.setLocation(address);

			// JavalCeBuȃR[hɕϊ
			String str = null;
			try {
				str = new String(rawData, PetaMessenger.ENCODING);
			} catch (UnsupportedEncodingException e) {
				e.printStackTrace();
				return;
			}

			// ĊetB[hɑ
			String[] fields = Tool.split(str, ':', 6);
			switch (fields.length) {
			case 6:
				member.userName = fields[5];

				member.osUserName = fields[2];
				member.computerName = fields[3];
				messageType = Integer.parseInt(fields[4]);
				break;
			case 5:
				member.osUserName = fields[2];
				member.computerName = fields[3];
				messageType = Integer.parseInt(fields[4]);
				break;
			default:
				logger.warn("Invalid data. " + fields.length); //$NON-NLS-1$
				break;
			}
			if (member.computerName != null) {
				member.computerName = member.computerName.toUpperCase();
			}
		}
	}
}
