package petaclone.swt;

import java.io.IOException;
import java.nio.charset.Charset;
import java.util.Iterator;
import java.util.List;
import java.util.SortedMap;

import jp.takayuki.logging.Logger;
import jp.takayuki.swt.util.SWTUtils;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Os;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.TabFolder;
import org.eclipse.swt.widgets.TabItem;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableItem;

import petaclone.AppPreferences;
import petaclone.CollectableControl;
import petaclone.LaunchConfiguration;
import petaclone.LaunchConfigurations;
import petaclone.MessengerFacade;
import petaclone.StringControlChecker;
import petaclone.swt.SWTLaunchConfigEditDialog.LaunchConfigCallback;

/**
 * ݒ_CAOB
 * 
 * @author Takayuki Shoji
 */
public class SWTConfigurationDialog implements LaunchConfigCallback {

	/**
	 * logger
	 */
	private static final Logger logger = Logger
			.getLogger(SWTConfigurationDialog.class);

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/**
	 * ̒̃`FbNɎgGR[fBO
	 */
	private static final String ENCODING = "Windows-31J"; //$NON-NLS-1$

	/** CEBhE */
	private SWTMainWindow main;

	/** C[W擾 */
	private ImageProvider imgProvider;

	/** MessengerFacade */
	private MessengerFacade facade;

	/** ݒ */
	private AppPreferences pref;

	/** N\ */
	private LaunchConfigurations launchConfigData;

	/** Rg[NG[^[ */
	SWTControlCreator creator;

	/** _CAO */
	private Shell shell;

	/**
	 * t@CI_CAÕtB^
	 */
	private final static String[] filterExtentions = { "*.wav", "*.au", //$NON-NLS-1$ //$NON-NLS-2$
			"*.aiff", "*.aifc", "*.snd", "*.*" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$

	/**
	 * t@CI_CAÕtB^
	 */
	private final static String[] filterNames = { "Wave(*.wav)", "AU(*.au)", //$NON-NLS-1$ //$NON-NLS-2$
			"AIFF(*.aiff)", "AIFF-C(*.aifc)", "SND(*.snd)", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			mm.get("SWTConfigurationDialog.allTypes") }; //$NON-NLS-1$ 

	/** N\^u */
	LaunchConfigContent launchConfig;

	/** N\ύXꂽꍇtrue */
	private boolean launchConfigurationUpdated;

	/** ^utH_ */
	private TabFolder tabFolder;

	/**
	 * RXgN^
	 * @param main CEBhE
	 * @param imgProvider ImageProvider
	 * @param facade MessengerFacade
	 * @param pref ݒ
	 * @param launchConfig N\ l:LaunchConfiguration
	 */
	public SWTConfigurationDialog(SWTMainWindow main,
			ImageProvider imgProvider, MessengerFacade facade,
			AppPreferences pref, LaunchConfigurations launchConfig) {

		this.imgProvider = imgProvider;
		this.facade = facade;
		this.pref = pref;
		this.launchConfigData = launchConfig;

		this.main = main;
		this.shell = new Shell(main.getShell(), SWT.DIALOG_TRIM
				| SWT.APPLICATION_MODAL | SWT.RESIZE);
		createWidgets();
	}

	/**
	 * EBWFbg̏
	 */
	private void createWidgets() {
		// EBhE^Cg
		this.shell.setText(mm.get("SWTConfigurationDialog.title")); //$NON-NLS-1$
		if (!Os.isMacOSX())
			this.shell.setImage(this.imgProvider.getImage("preference.gif")); //$NON-NLS-1$

		// CAEg
		GridLayout shellLayout = new GridLayout();
		shellLayout.numColumns = 1;
		shell.setLayout(shellLayout);

		// ^utH_[̐
		tabFolder = new TabFolder(shell, SWT.NONE);
		GridData tabFolderData = new GridData();
		tabFolderData.verticalAlignment = SWT.FILL;
		tabFolderData.horizontalAlignment = SWT.FILL;
		tabFolderData.grabExcessHorizontalSpace = true;
		tabFolderData.grabExcessVerticalSpace = true;

		if (Os.isWindows()) {
			tabFolderData.widthHint = 400;
			tabFolderData.minimumWidth = 400;
		} else {
			tabFolderData.widthHint = 550;
			tabFolderData.minimumWidth = 550;
		}
		tabFolder.setLayoutData(tabFolderData);

		// *̓AvP[V̍ċNɓKp܂B
		Label reboot = new Label(shell, SWT.NONE);
		reboot.setText(mm.get("SWTConfigurationDialog.needToRestart")); //$NON-NLS-1$

		// {^̐
		// {^̈
		Composite buttonsArea = new Composite(shell, SWT.NONE);
		// E
		buttonsArea.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_END));

		// {^̈̃CAEg
		RowLayout buttonsLayout = new RowLayout(SWT.HORIZONTAL);
		buttonsLayout.spacing = 5;
		buttonsLayout.pack = false;
		buttonsArea.setLayout(buttonsLayout);
		createButtons(buttonsArea);
		//buttonsArea.pack();

		// Rg[NG[^[̐
		creator = new SWTControlCreator(pref.getProperties());

		// Sʃ^u
		TabItem generalItem = new TabItem(tabFolder, SWT.NONE);
		generalItem.setText(mm.get("SWTConfigurationDialog.tab.general.label")); //$NON-NLS-1$
		Composite generalItemComposite = prepareTabItem(generalItem);
		creator.setParent(generalItemComposite);
		createGeneralConfigContent(creator);

		// M^u
		TabItem receiveItem = new TabItem(tabFolder, SWT.NONE);
		receiveItem.setText(mm.get("SWTConfigurationDialog.tab.receive.label")); //$NON-NLS-1$
		Composite receiveItemComposite = prepareTabItem(receiveItem);
		creator.setParent(receiveItemComposite);
		createReceiveConfigContent(creator);

		// ۊǔEO^u
		TabItem logItem = new TabItem(tabFolder, SWT.NONE);
		logItem.setText(//
		mm.get("SWTConfigurationDialog.tab.storageBoxAndLog.label")); //$NON-NLS-1$
		Composite logItemComposite = prepareTabItem(logItem);
		creator.setParent(logItemComposite);
		createLogConfigContent(creator);

		// \1^u
		TabItem viewItem = new TabItem(tabFolder, SWT.NONE);
		viewItem.setText(mm.get("SWTConfigurationDialog.tab.view1.label")); //$NON-NLS-1$
		Composite viewItemComposite = prepareTabItem(viewItem);
		creator.setParent(viewItemComposite);
		createViewConfigContent(creator);

		// \2^u
		TabItem view2Item = new TabItem(tabFolder, SWT.NONE);
		view2Item.setText(mm.get("SWTConfigurationDialog.tab.view2.label")); //$NON-NLS-1$
		Composite view2ItemComposite = prepareTabItem(view2Item);
		creator.setParent(view2ItemComposite);
		createView2ConfigContent(creator);

		// lbg[N^u(u[hLXgݒ)
		TabItem networkItem = new TabItem(tabFolder, SWT.NONE);
		networkItem.setText(mm.get("SWTConfigurationDialog.tab.network.label")); //$NON-NLS-1$
		Composite networkItemComposite = prepareTabItem(networkItem);
		creator.setParent(networkItemComposite);
		createNetworkConfigContent(creator);

		// N\^u
		TabItem launchItem = new TabItem(tabFolder, SWT.NONE);
		launchItem.setText(mm
				.get("SWTConfigurationDialog.tab.launchConfiguration.label")); //$NON-NLS-1$
		createLaunchConfigContent(launchItem);

		// zbgL[^u
		if (Os.isWindows()) {
			TabItem hotkeyItem = new TabItem(tabFolder, SWT.NONE);
			hotkeyItem.setText(mm
					.get("SWTConfigurationDialog.tab.hotkey.label")); //$NON-NLS-1$
			Composite hotkeyComposite = prepareTabItem(hotkeyItem);
			creator.setParent(hotkeyComposite);
			createHotkeyConfigContent(creator);
		}

		// ȇ傫ɍ킹
		//generalItemComposite.pack();
		//receiveItemComposite.pack();
		//logItemComposite.pack();
		//viewItemComposite.pack();
		//networkItemComposite.pack();
		//tabFolder.pack();
		this.shell.pack();
		//System.out.println("M^u" + receiveItemComposite.getSize());
		//System.out.println("^utH_"+tabFolder.getSize());

		int index = pref.getLastPreferenceTabIndex();
		if (index >= 0 && index < tabFolder.getItemCount()) {
			tabFolder.setSelection(index);
		}
	}

	/**
	 * zbgL[^u
	 * 
	 * @param creator Rg[NG[^
	 */
	private void createHotkeyConfigContent(SWTControlCreator creator) {

		creator.createLabel(//
		mm.get("SWTConfigurationDialog.hotkey.explanation")); //$NON-NLS-1$

		creator.createKeyCombinationSelector(
				AppPreferences.HOTKEY_DISPLAYWINDOW, "", //$NON-NLS-1$
				mm.get("SWTConfigurationDialog.hotkey.showPetaclone")); //$NON-NLS-1$

		creator.createKeyCombinationSelector(
				AppPreferences.HOTKEY_SHOW_IN_INDEPENDENT_WINDOW, "", //$NON-NLS-1$
				mm.get("SWTConfigurationDialog.hotkey.showInIndependentWindow")); //$NON-NLS-1$

		creator.createKeyCombinationSelector(
				AppPreferences.HOTKEY_OPENBLANKMESSAGEWINDOW, "", //$NON-NLS-1$
				mm.get("SWTConfigurationDialog.hotkey.newMessage")); //$NON-NLS-1$

		creator.createKeyCombinationSelector(
				AppPreferences.HOTKEY_OPENBLANKINDEPENDENTMESSAGEWINDOW,
				"", //$NON-NLS-1$
				mm.get("SWTConfigurationDialog.hotkey.newMessageInIndependentWindow")); //$NON-NLS-1$

		creator.createKeyCombinationSelector(
				AppPreferences.HOTKEY_OPENLOGVIEWER, "", //$NON-NLS-1$
				mm.get("SWTConfigurationDialog.hotkey.openLogViewer")); //$NON-NLS-1$
	}

	/**
	 * M^u
	 * @param creator Rg[NG[^
	 */
	private void createReceiveConfigContent(SWTControlCreator creator) {

		final String[] openNoticeDisplayDisplayValues = {
				mm.get("SWTConfigurationDialog.openNoticeMethod.baloon"), //$NON-NLS-1$
				mm.get("SWTConfigurationDialog.openNoticeMethod.alert"), //$NON-NLS-1$
				mm.get("SWTConfigurationDialog.openNoticeMethod.doNothing") //$NON-NLS-1$
		};
		final String[] openNoticeDisplayPropertyValues = { "0", "1", "2" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$

		creator.createRadioButton(
				AppPreferences.OPEN_NOTICE_DISPLAY, //
				mm.get("SWTConfigurationDialog.openNoticeMethod.label"), //$NON-NLS-1$
				openNoticeDisplayDisplayValues,
				openNoticeDisplayPropertyValues,
				Integer.toString(AppPreferences.OPEN_NOTICE_DISPLAY_BALLOON));
		//
		final String[] displayValues = {
				mm.get("SWTConfigurationDialog.receiveTone.doNothing"), //$NON-NLS-1$
				mm.get("SWTConfigurationDialog.receiveTone.hardwareBeep"), //$NON-NLS-1$
				mm.get("SWTConfigurationDialog.receiveTone.playSound") //$NON-NLS-1$
		};
		final String[] propertyValues = { "0", "beep", "sound" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		creator.createRadioButton(AppPreferences.PLAY_SOUND,
				mm.get("SWTConfigurationDialog.receiveTone.label"), //$NON-NLS-1$
				displayValues, propertyValues, "0"); //$NON-NLS-1$

		//
		// tH_
		String path;
		if (Os.isWindows()) {
			path = "C:\\WINDOWS\\Media\\"; //$NON-NLS-1$
		} else {
			path = ""; //$NON-NLS-1$
		}
		creator.createFileSelector(AppPreferences.SOUND_URL, //
				mm.get("SWTConfigurationDialog.receiveTone.path.label"), //$NON-NLS-1$
				"", path, filterNames, filterExtentions, false); //$NON-NLS-1$
		//
		creator.createDirectorySelector(
				AppPreferences.ATTACHMENT_STORAGE, //
				mm.get("SWTConfigurationDialog.attachmentTempDir.label"), //$NON-NLS-1$
				pref.getAttachmentStorage(), pref.getAttachmentStorage(),
				false, true);
		//
		creator.createSpinner(AppPreferences.MAX_AUTO_DOWNLOAD_SIZE,//
				mm.get("SWTConfigurationDialog.maxAutoDownloadSize.label"), //$NON-NLS-1$
				20 * 1024 * 1024, false, 0, Integer.MAX_VALUE, 1024 * 1024);
		//
		creator.createCheckBox(AppPreferences.INLINE_IMAGE_AUTO_DOWNLOAD,//
				mm.get("SWTConfigurationDialog.autoInlineImageDownload.label"), //$NON-NLS-1$
				true);
		//
		creator.createCheckBox(AppPreferences.FOLDER_AUTO_DOWNLOAD,//
				mm.get("SWTConfigurationDialog.autoFolderDownload.label"), //$NON-NLS-1$
				false);
		//
		creator.createCheckBox(
				AppPreferences.WARN_CLOSING_UNSAVED_RECEIVED_MESSAGE,
				mm.get("SWTConfigurationDialog.warnWhenClosingUnsavedMessage.label"), //$NON-NLS-1$
				false);
		//
		creator.createPasswordBox(AppPreferences.OPEN_PASSWORD, mm
				.get("SWTConfigurationDialog.changeOpenPassword.button.label"), //$NON-NLS-1$
				mm.get("SWTConfigurationDialog.changeOpenPassword.title")); //$NON-NLS-1$
		//
		creator.createCheckBox(
				AppPreferences.ATTACH_CLIPBOARD_CONTENT_WITH_DIALOG,
				mm.get("SWTConfigurationDialog.showDialogWhenAttachingClipboard.label"), //$NON-NLS-1$
				false);
		//
		creator.createFileSelector(AppPreferences.GRAPHIC_EDITOR, //
				mm.get("SWTConfigurationDialog.graphicsEditor.label"), //$NON-NLS-1$
				pref.getGraphicEditor(), "", null, null, true); //$NON-NLS-1$
		//
		//creator.createCheckBox(AppPreferences.FIT_CONTENT_OF_NEW_MESSAGE,
		//		"VbZ[W̃EBhETCY͓eɍ킹(I)", false);

	}

	/**
	 * lbg[N^u
	 * @param creator Rg[NG[^
	 */
	void createNetworkConfigContent(SWTControlCreator creator) {
		creator.createBroadcastGroup(AppPreferences.BROADCAST,
				AppPreferences.LIMITED_BROADCAST,
				AppPreferences.DIRECTED_BROADCAST);

		creator.createCheckBox(AppPreferences.USE_SYSTEM_PROXIES, //
				mm.get("SWTConfigurationDialog.useSystemProxiesSetting.label"), //$NON-NLS-1$
				true);

		creator.createCheckBox(AppPreferences.PREFER_IPV6_ADDRESSES, //
				mm.get("SWTConfigurationDialog.preferIPv6.label"), //$NON-NLS-1$
				false);

		creator.createSeparator();

		creator.createCheckBox(AppPreferences.ENABLE_IPMESSENGER_COMPONENT,//
				mm.get("SWTConfigurationDialog.enableIPMessengerComponent"), //$NON-NLS-1$
				true);

		creator.createCheckBox(AppPreferences.ENABLE_PETAROUX_COMPONENT,//
				mm.get("SWTConfigurationDialog.enablePetarouXComponent"), //$NON-NLS-1$
				true);

		creator.createCheckBox(AppPreferences.ENABLE_OLD_PETAROU_COMPONENT, //
				mm.get("SWTConfigurationDialog.enableOldPetarouComponent"), //$NON-NLS-1$
				false);

		creator.createCheckBox(AppPreferences.ENABLE_DENKI_YOHOU_CLIENT, //
				mm.get("SWTConfigurationDialog.enableDenkiYohouClient"), //$NON-NLS-1$
				false);
	}

	/**
	 * N\^u
	 */
	class LaunchConfigContent {
		/** N\̕\ */
		private Table table;

		/** \錳f[^ */
		private List confs;

		/** ǉ...{^ */
		private Button buttonAdd;

		/** ҏW...{^ */
		private Button buttonEdit;

		/** 폜{^ */
		private Button buttonDelete;

		/** փ{^ */
		private Button buttonUp;

		/** փ{^ */
		private Button buttonDown;

		/** ^u */
		private TabItem parent;

		/**
		 * RXgN^ 
		 * @param parent ^uACe
		 */
		LaunchConfigContent(TabItem parent) {
			this.parent = parent;

			// GUI
			Composite c = new Composite(parent.getParent(), SWT.NONE);
			GridLayout cLayout = new GridLayout();
			cLayout.numColumns = 2;
			c.setLayout(cLayout);
			//c.setLayoutData(new GridData(GridData.FILL_BOTH));

			table = new Table(c, SWT.CHECK | SWT.BORDER | SWT.SINGLE
					| SWT.FULL_SELECTION);
			table.setHeaderVisible(true);

			SWTUtils.createTableColumn(table, null, 100,
					"LaunchConfigContent.col.name"); //$NON-NLS-1$
			SWTUtils.createTableColumn(table, null, 70,
					"LaunchConfigContent.col.protocol"); //$NON-NLS-1$
			SWTUtils.createTableColumn(table, null, 80,
					"LaunchConfigContent.col.detail"); //$NON-NLS-1$

			table.setLayoutData(new GridData(GridData.FILL_BOTH));
			table.addSelectionListener(new SelectionAdapter() {
				public void widgetDefaultSelected(SelectionEvent e) {
					openEditDialog();
				}
			});

			// {^
			Composite buttonsArea = new Composite(c, SWT.NONE);
			buttonsArea.setLayoutData(new GridData(
					GridData.VERTICAL_ALIGN_BEGINNING));
			RowLayout buttonsAreaLayout = new RowLayout(SWT.VERTICAL);
			buttonsAreaLayout.pack = false;
			buttonsAreaLayout.marginTop = 20;
			buttonsAreaLayout.marginBottom = 0;
			buttonsAreaLayout.marginLeft = 0;
			buttonsAreaLayout.marginRight = 0;
			buttonsAreaLayout.spacing = 5;
			buttonsArea.setLayout(buttonsAreaLayout);

			// {^̃Xi[

			buttonAdd = new Button(buttonsArea, SWT.PUSH);
			buttonAdd.setText(mm.get("LaunchConfigContent.add.label")); //$NON-NLS-1$
			buttonAdd.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent event) {

					new SWTLaunchConfigEditDialog(pref, facade,
							SWTConfigurationDialog.this, shell, null);
				}
			});

			buttonEdit = new Button(buttonsArea, SWT.PUSH);
			buttonEdit.setText(mm.get("LaunchConfigContent.edit.label")); //$NON-NLS-1$
			buttonEdit.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent event) {
					openEditDialog();
				}
			});

			buttonDelete = new Button(buttonsArea, SWT.PUSH);
			buttonDelete.setText(mm.get("LaunchConfigContent.delete.label")); //$NON-NLS-1$
			buttonDelete.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent event) {

					if (!SWTUtils.showConfirm(shell, //
							mm.get("LaunchConfigContent.delete.title"), //$NON-NLS-1$
							mm.get("LaunchConfigContent.delete.message"))) { //$NON-NLS-1$
						return;
					}
					// IACe̎擾
					int index = table.getSelectionIndex();
					if (index == -1) {
						return;
					}
					//TableItem item = table.getItem(index);
					// Nݒ̎擾
					LaunchConfiguration conf = (LaunchConfiguration) confs
							.get(index);
					// 폜
					conf.delete();
					confs.remove(index);
					table.remove(index);
					launchConfigurationUpdated = true;
				}
			});

			buttonUp = new Button(buttonsArea, SWT.PUSH);
			buttonUp.setText(mm.get("LaunchConfigContent.up.label")); //$NON-NLS-1$
			buttonUp.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent event) {
					int index = table.getSelectionIndex();
					swapItem(index - 1);
				}
			});

			buttonDown = new Button(buttonsArea, SWT.PUSH);
			buttonDown.setText(mm.get("LaunchConfigContent.down.label")); //$NON-NLS-1$
			buttonDown.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent event) {
					int index = table.getSelectionIndex();
					swapItem(index);
				}
			});

			// e[uɃf[^ݒ
			confs = SWTConfigurationDialog.this.launchConfigData
					.getConfigurations();
			for (int i = 0; i < confs.size(); i++) {
				new TableItem(table, SWT.LEFT);
				updateTable(i);
			}

			// |V[̓Kp
			if (!pref.getPolicy().canEditLaunchConfiguration()) {
				table.setEnabled(false);
				buttonEdit.setEnabled(false);
				buttonAdd.setEnabled(false);
				buttonDelete.setEnabled(false);
				buttonUp.setEnabled(false);
				buttonDown.setEnabled(false);
			}

			c.pack();
			parent.setControl(c);
		}

		/**
		 * ҏW_CAOJ܂B
		 */
		private void openEditDialog() {
			// IACe̎擾
			int index = table.getSelectionIndex();
			if (index == -1) {
				return;
			}
			TableItem item = table.getItem(index);
			// Nݒ̎擾
			LaunchConfiguration conf = (LaunchConfiguration) confs.get(index);
			// `FbNԂ𔽉f
			conf.setIsValid(item.getChecked());
			// _CAOJ
			new SWTLaunchConfigEditDialog(pref, facade,
					SWTConfigurationDialog.this, shell, conf);
		}

		/**
		 * ACȅ㉺܂B
		 * @param row sԍ row+1オ
		 */
		private void swapItem(int row) {
			if (row < 0 || row >= confs.size() - 1) {
				return;
			}
			// `FbNԂ𔽉f
			for (int i = 0; i < 2; i++) {
				LaunchConfiguration conf = (LaunchConfiguration) confs.get(row
						+ i);
				TableItem item = table.getItem(row + i);
				conf.setIsValid(item.getChecked());
			}

			// f[^
			LaunchConfiguration conf = (LaunchConfiguration) confs.remove(row);
			confs.add(row + 1, conf);
			// IԂ擾
			int select = table.getSelectionIndex();
			if (select == row) {
				select++;
			} else {
				select--;
			}
			// e[u̍XV
			updateTable(row);
			updateTable(row + 1);
			// IԂ̍XV
			table.setSelection(select);
		}

		/**
		 * e[u̎w肳ꂽs̕\XV܂B
		 * @param row sԍ
		 */
		void updateTable(int row) {
			assert row < confs.size();
			assert row < table.getItemCount();

			LaunchConfiguration conf = (LaunchConfiguration) confs.get(row);
			TableItem item = table.getItem(row);
			item.setText(conf.getRowStrings());
			item.setChecked(conf.getIsValid());
		}

		/**
		 * e[u̎w肳ꂽACe̕\XV܂B
		 * @param conf Nݒ
		 */
		void updateTable(LaunchConfiguration conf) {
			int index = confs.indexOf(conf);
			updateTable(index);
		}

		/**
		 * ݒǉ܂B
		 * @param conf Nݒ
		 */
		void addConf(LaunchConfiguration conf) {
			confs.add(conf);
			TableItem item = new TableItem(table, SWT.LEFT);
			item.setText(conf.getRowStrings());
			item.setChecked(conf.getIsValid());
		}

		/**
		 * `FbN{bNXf[^ɔf܂B
		 * @return XVꂽꍇtrue
		 */
		boolean synchCheckbox() {
			boolean updated = false;
			for (int i = 0; i < confs.size(); i++) {
				LaunchConfiguration conf = (LaunchConfiguration) confs.get(i);
				TableItem item = table.getItem(i);
				// `FbN{bNXf[^ɔf
				boolean enabled = item.getChecked();
				if (!updated) {
					updated = conf.getIsValid() != enabled;
				}
				conf.setIsValid(enabled);
			}
			return updated;
		}

		/**
		 * vpeBۑ܂B
		 */
		void saveProperties() {
			for (int i = 0; i < confs.size(); i++) {
				LaunchConfiguration conf = (LaunchConfiguration) confs.get(i);
				// t@Cɕۑ
				try {
					conf.save();
				} catch (IOException e) {

					String msg = mm.get("LaunchConfigContent.error", //$NON-NLS-1$
							new Object[] { conf.getName() });
					logger.error(msg, e);
					SWTUtils.showError(shell, msg + e);
				}
			}
		}

		/**
		 * ^utH[JX܂B
		 */
		void setFocus() {
			parent.getParent().setSelection(parent);
		}
	}

	/**
	 * N\^u̓e𐶐܂B
	 * @param parent ^uACe
	 */
	private void createLaunchConfigContent(TabItem parent) {
		launchConfig = new LaunchConfigContent(parent);
	}

	/**
	 * Sʃ^u̓e𐶐܂B
	 * @param creator Rg[NG[^
	 */
	private void createGeneralConfigContent(SWTControlCreator creator) {

		// O
		creator.createText(AppPreferences.USER_NAME, //
				mm.get("GeneralConfigContent.screenName"), //$NON-NLS-1$
				pref.getUserName(), true, new StringControlChecker(ENCODING,
						200));

		// O[v
		creator.createText(AppPreferences.GROUP_NAME, //
				mm.get("GeneralConfigContent.groupName"), //$NON-NLS-1$
				"", //$NON-NLS-1$
				true, new StringControlChecker(ENCODING, 200));
		//
		creator.createCheckBox(AppPreferences.IS_ABSENCE_AT_STARTUP, //
				mm.get("GeneralConfigContent.absentAtStartup"), //$NON-NLS-1$
				false);
		//
		creator.createCheckBox(AppPreferences.AUTO_ABSENCE, //
				mm.get("GeneralConfigContent.autoAbsence"), //$NON-NLS-1$
				false);
		//
		creator.createSpinner(AppPreferences.AUTO_ABSENCE_LIMIT_SEC, //
				mm.get("GeneralConfigContent.autoAbsenceTimeSec"), //$NON-NLS-1$
				600, false, 60, 24 * 60 * 60, 60);
		//
		creator.createCheckBox(AppPreferences.NEVER_USE_SEND_ALL,//
				mm.get("GeneralConfigContent.donotUseSendToAll"), //$NON-NLS-1$
				true);
		//
		creator.createCheckBox(AppPreferences.NEVER_SEND_EMPTY_MESSAGE,//
				mm.get("GeneralConfigContent.donotSendBlankMessage"), //$NON-NLS-1$
				false);
		//
		creator.createCheckBox(AppPreferences.USE_IM_MSG,//
				mm.get("GeneralConfigContent.useIMAtMessageWindow"), //$NON-NLS-1$
				true);
		//
		creator.createCheckBox(AppPreferences.USE_IM_SEND,//
				mm.get("GeneralConfigContent.useIMAtSendDialog"), //$NON-NLS-1$
				false);
		//
		creator.createCheckBox(AppPreferences.IM_MODE_IS_ROMAN,//
				mm.get("GeneralConfigContent.romanInputMode"), //$NON-NLS-1$
				true);
		//
		String[] propertyValues = new String[] { "0", /*"1",*/ "2" }; //$NON-NLS-1$ //$NON-NLS-2$
		String[] values = new String[] {
				mm.get("GeneralConfigContent.softwareUpdate.doNotCheck"), //$NON-NLS-1$
				//mm.get("GeneralConfigContent.softwareUpdate.notifyBeforeDownload"), //$NON-NLS-1$
				mm.get("GeneralConfigContent.softwareUpdate.notifyAfterDownload") //$NON-NLS-1$
		};
		creator.createRadioButton(AppPreferences.CHECK_FOR_UPDATES,
				mm.get("GeneralConfigContent.softwareUpdate"), values, //$NON-NLS-1$
				propertyValues, "2"); //$NON-NLS-1$
		//
		creator.createText(AppPreferences.QUOTE,//
				mm.get("GeneralConfigContent.quoteCharacter"), //$NON-NLS-1$
				AppPreferences.QUOTE_DEFAULT, true, null);
		//
		String[] newMessageValue = {
				mm.get("GeneralConfigContent.privacyNormal"), //$NON-NLS-1$
				mm.get("GeneralConfigContent.privacySealed"), //$NON-NLS-1$
				mm.get("GeneralConfigContent.privacyPassword") //$NON-NLS-1$
		};

		String[] newMessageProperties = {
				AppPreferences.PRIVACY_PROTECTION_NONE,
				AppPreferences.PRIVACY_PROTECTION_SEALED,
				AppPreferences.PRIVACY_PROTECTION_PASSWORD };
		creator.createRadioButton(
				AppPreferences.PRIVACY_PROTECTION, //
				mm.get("GeneralConfigContent.defaultPrivacyLevel"), //$NON-NLS-1$
				newMessageValue, newMessageProperties,
				AppPreferences.PRIVACY_PROTECTION_NONE);
		//
		creator.createCheckBox(AppPreferences.CIPHER_IS_UNLIMITED,//
				mm.get("GeneralConfigContent.JCEUnlimitedStrength"), //$NON-NLS-1$
				false);
		//
		creator.createCheckBox(AppPreferences.USE_SMALL_KEY,//
				mm.get("GeneralConfigContent.forceRSA512withRC2_40"), //$NON-NLS-1$
				false);
	}

	/**
	 * \1^u
	 * @param creator Rg[NG[^
	 */
	private void createViewConfigContent(SWTControlCreator creator) {
		//
		creator.createCheckBox(AppPreferences.USE_REGISTERED_USER_NAME,//
				mm.get("ViewConfigContent.preferRegisteredScreenName"), //$NON-NLS-1$
				false);
		//
		creator.createCheckBox(AppPreferences.USE_REGISTERED_GROUP_NAME,//
				mm.get("ViewConfigContent.preferRegisteredGroupName"), //$NON-NLS-1$
				false);

		//
		creator.createCheckBox(AppPreferences.ICONIFY_INSTEDOF_QUITTING,//
				mm.get("ViewConfigContent.iconifyInSteadOfQuitting"), //$NON-NLS-1$
				false);

		//
		creator.createCheckBox(AppPreferences.HIDE_ON_ICONIFIED,//
				mm.get("ViewConfigContent.donotShowInTaskbarWhenIconified"), //$NON-NLS-1$
				false);

		//
		creator.createCheckBox(AppPreferences.IS_MINIMIZED_AT_STARTUP,//
				mm.get("ViewConfigContent.minimizedAtStartup"), //$NON-NLS-1$
				false);

		//
		creator.createCheckBox(AppPreferences.SAVE_SEND_DIALOG_LOCATION,//
				mm.get("ViewConfigContent.saveLocationOfSendDialog"), //$NON-NLS-1$
				false);

		//
		creator.createCheckBox(AppPreferences.SAVE_SEND_DIALOG_SIZE,//
				mm.get("ViewConfigContent.saveSizeOfSendDialog"), //$NON-NLS-1$
				true);

		//
		creator.createCheckBox(AppPreferences.SHOW_STICKY_OPERATION_GUIDE, //
				mm.get("ViewConfigContent.showStickyOperationGuide"), true); //$NON-NLS-1$
		//
		creator.createCheckBox(AppPreferences.MOVE_CURSOR_TO_LAST,
				mm.get("ViewConfigContent.moveCursorAfterQuotation"), true); //$NON-NLS-1$

		// t@C}l[Wst@C
		String[][] menuData = {
				mm.getArray("ViewConfigContent.fileManager.popup.file"), //$NON-NLS-1$
				mm.getArray("ViewConfigContent.fileManager.popup.folder") //$NON-NLS-1$
		};

		creator.createToolConfigurator(AppPreferences.FILE_MANAGER_FOLDER,
				pref.getToolCommand(AppPreferences.FILE_MANAGER_FOLDER),
				pref.getToolParameter(AppPreferences.FILE_MANAGER_FOLDER),
				pref.getToolWorkingFolder(AppPreferences.FILE_MANAGER_FOLDER),
				mm.get("ViewConfigContent.fileManager.openFolder"), menuData); //$NON-NLS-1$

		creator.createToolConfigurator(
				AppPreferences.FILE_MANAGER_CONTAINING_FOLDER,
				pref.getToolCommand(AppPreferences.FILE_MANAGER_CONTAINING_FOLDER),
				pref.getToolParameter(AppPreferences.FILE_MANAGER_CONTAINING_FOLDER),
				pref.getToolWorkingFolder(AppPreferences.FILE_MANAGER_CONTAINING_FOLDER),
				mm.get("ViewConfigContent.fileManager.openContainingFolder"), //$NON-NLS-1$
				menuData);
	}

	/**
	 * \2^u
	 * @param creator Rg[NG[^
	 */
	private void createView2ConfigContent(SWTControlCreator creator) {

		// bZ[WEBhẺƍ
		creator.createSquare(AppPreferences.MESSAGE_WIDTH,
				pref.getMessageWidth(), AppPreferences.MESSAGE_HEIGHT,
				pref.getMessageHeight(),
				mm.get("ViewConfigContent.initialSizeOfSimpleMessage")); //$NON-NLS-1$

		//
		creator.createRadioButton(
				AppPreferences.MESSAGE_ARRANGEMENT,
				mm.get("ViewConfigContent.messageArrangement"), //$NON-NLS-1$
				new String[] {
						mm.get("ViewConfigContent.messageArrangement.fromTopLeft"), //$NON-NLS-1$
						mm.get("ViewConfigContent.messageArrangement.fromBottomCenter"), //$NON-NLS-1$
						mm.get("ViewConfigContent.messageArrangement.mousePosition") }, //$NON-NLS-1$
				new String[] { AppPreferences.MESSAGE_ARRANGEMENT_TOP_LEFT,
						AppPreferences.MESSAGE_ARRANGEMENT_BOTTOM_CENTER,
						AppPreferences.MESSAGE_ARRANGEMENT_MOUSE },
				AppPreferences.MESSAGE_ARRANGEMENT_TOP_LEFT);

		//
		creator.createCheckBox(AppPreferences.USE_INDEPENDENT_MESSAGE_WINDOW,//
				mm.get("ViewConfigContent.openMessageAsIndependentWindow"), //$NON-NLS-1$
				false);

		//
		if (Os.isWindows())
			creator.createCheckBox(AppPreferences.WARN_EXCESSIVE_WINDOWS,
					mm.get("ViewConfigContent.warnIfThereAreTooManyWindows"), //$NON-NLS-1$
					true);

		//
		creator.createCheckBox(AppPreferences.CANCEL_TRANSPARENCY_ON_FOCUS, //
				mm.get("ViewConfigContent.cancelTransparencyOnFocus"), false); //$NON-NLS-1$

		//
		creator.createCheckBox(
				AppPreferences.USE_FIRST_LINE_OF_SELECTION,
				mm.get("ViewConfigContent.useFirstLineOfSelectionForOneLineView"), //$NON-NLS-1$
				false);

		//
		creator.createCheckBox(AppPreferences.USE_TOOLTIP_AT_ONELINE,//
				mm.get("ViewConfigContent.showTooltipAtOneLineView"), //$NON-NLS-1$
				true);

		//
		creator.createCheckBox(
				AppPreferences.ASK_WHETHER_MAKE_FIRST_DROPPED_ITEM_INLINE,
				mm.get("ViewConfigContent.askWhetherMakeFirstDroppedItemInline"), true); //$NON-NLS-1$

		//
		creator.createSeparator();

		//
		creator.createCheckBox(
				AppPreferences.USE_CUSTOM_FONT_IN_SIMPLE_MESSAGE,//
				mm.get("ViewConfigContent.specifyCustomFontInSimpleMessage"), //$NON-NLS-1$
				false);

		//
		creator.createFontDialogButton(AppPreferences.FONT_IN_SIMPLE_MESSAGE,//
				mm.get("ViewConfigContent.fontNameInSimpleMessage")); //$NON-NLS-1$

		//
		creator.createSpinner(AppPreferences.TAB_WIDTH,//
				mm.get("ViewConfigContent.tabWidthInSimpleMessage"), //$NON-NLS-1$
				0, false, 0, 256, 4);
		//
		creator.createSeparator();
		//
		creator.createComboBox(AppPreferences.QRCODE_ERROR_CORRECTION_LEVEL,//
				mm.get("ViewConfigContent.qrCodeErrorCorrectionLevel"),//$NON-NLS-1$
				new String[] { mm.get("ViewConfigContent.qr.L"), //$NON-NLS-1$
						mm.get("ViewConfigContent.qr.M"), //$NON-NLS-1$
						mm.get("ViewConfigContent.qr.Q"), //$NON-NLS-1$
						mm.get("ViewConfigContent.qr.H") }, //$NON-NLS-1$
				new String[] { "L", "M", "Q", "H" }, //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
				"M", false); //$NON-NLS-1$
		//
		creator.createSpinner(AppPreferences.QRCODE_MODULE_SIZE,//
				mm.get("ViewConfigContent.qrModuleSize"), //$NON-NLS-1$
				2, false, 1, 16, 1);
		//
		{
			SortedMap csMap = Charset.availableCharsets();
			String[] css = new String[csMap.size() + 1];
			css[0] = mm.get("ViewConfigContent.qrEncoding.default.label"); //$NON-NLS-1$

			{
				int i = 1;
				Iterator itr = csMap.values().iterator();
				while (itr.hasNext()) {
					Charset cs = (Charset) itr.next();
					css[i] = cs.name();
					i++;
				}
			}
			String[] css2 = new String[css.length];
			System.arraycopy(css, 0, css2, 0, css.length);
			css2[0] = ""; //$NON-NLS-1$
			for (int i = 1; i < css2.length; i++) {
				css2[i] = css[i];
			}

			creator.createComboBox(AppPreferences.QRCODE_ENCODING, //
					mm.get("ViewConfigContent.qrEncoding"), //$NON-NLS-1$
					css, css2, "", false); //$NON-NLS-1$
		}
		{
			String[] values = { "1", "2", "3", "4", "5", "6", "7", "8", "9", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$
					"10", "11", "12", "13", "14", "15", "16", "17", "18", "19", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$ //$NON-NLS-10$
					"20", "21", "22", "23", "24", "25", "26", "27", "28", "29", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$ //$NON-NLS-10$
					"30", "31", "32", "33", "34", "35", "36", "37", "38", "39", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$ //$NON-NLS-10$
					"40" }; //$NON-NLS-1$
			String[] labels = { "1: 9-19 bytes", "2: 16-34 bytes", //$NON-NLS-1$ //$NON-NLS-2$
					"3: 26-55 bytes", "4: 36-80 bytes", "5: 46-108 bytes", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
					"6: 60-136 bytes", "7: 66-156 bytes", "8: 86-194 bytes", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
					"9: 100-232 bytes", "10: 122-274 bytes", //$NON-NLS-1$ //$NON-NLS-2$
					"11: 140-324 bytes", "12: 158-370 bytes", //$NON-NLS-1$ //$NON-NLS-2$
					"13: 180-428 bytes", "14: 197-461 bytes", //$NON-NLS-1$ //$NON-NLS-2$
					"15: 223-523 bytes", "16: 253-589 bytes", //$NON-NLS-1$ //$NON-NLS-2$
					"17: 283-647 bytes", "18: 313-721 bytes", //$NON-NLS-1$ //$NON-NLS-2$
					"19: 341-795 bytes", "20: 385-861 bytes", //$NON-NLS-1$ //$NON-NLS-2$
					"21: 406-932 bytes", "22: 442-1006 bytes", //$NON-NLS-1$ //$NON-NLS-2$
					"23: 464-1094 bytes", "24: 514-1174 bytes", //$NON-NLS-1$ //$NON-NLS-2$
					"25: 538-1276 bytes", "26: 596-1370 bytes", //$NON-NLS-1$ //$NON-NLS-2$
					"27: 628-1468 bytes", "28: 661-1531 bytes", //$NON-NLS-1$ //$NON-NLS-2$
					"29: 701-1631 bytes", "30: 745-1735 bytes", //$NON-NLS-1$ //$NON-NLS-2$
					"31: 793-1843 bytes", "32: 845-1955 bytes", //$NON-NLS-1$ //$NON-NLS-2$
					"33: 901-2071 bytes", "34: 961-2191 bytes", //$NON-NLS-1$ //$NON-NLS-2$
					"35: 986-2306 bytes", "36: 1054-2434 bytes", //$NON-NLS-1$ //$NON-NLS-2$
					"37: 1096-2566 bytes", "38: 1142-2702 bytes", //$NON-NLS-1$ //$NON-NLS-2$
					"39: 1222-2812 bytes", "40: 1276-2956 bytes" }; //$NON-NLS-1$ //$NON-NLS-2$

			//
			creator.createComboBox(AppPreferences.QRCODE_DIVISION,
					mm.get("ViewConfigContent.qrDividingSize"), //$NON-NLS-1$
					labels, values, "40", true); //$NON-NLS-1$

		}
	}

	/**
	 * uOv^u
	 * @param creator Rg[NG[^
	 */
	private void createLogConfigContent(SWTControlCreator creator) {
		//
		creator.createLabel(mm.get("LogConfigContent.storageBox")); //$NON-NLS-1$
		//
		creator.createCheckBox(AppPreferences.SAVE_SENT_MESSAGE,//
				mm.get("LogConfigContent.saveSentMessage"), //$NON-NLS-1$
				false);
		//
		creator.createCheckBox(AppPreferences.SAVE_RECEIVED_MESSAGE,//
				mm.get("LogConfigContent.saveReceivedMessage"), //$NON-NLS-1$
				false);
		//
		creator.createSeparator();
		//
		creator.createLabel(mm.get("LogConfigContent.ipmsgLog")); //$NON-NLS-1$
		//
		creator.createCheckBox(AppPreferences.MESSAGE_LOG_ENABLED,//
				mm.get("LogConfigContent.writeMessageLog"), //$NON-NLS-1$
				false);
		//
		creator.createFileSelector(AppPreferences.LOG_VIEWER,//
				mm.get("LogConfigContent.pathToMessageLogViewer"), //$NON-NLS-1$
				"", "", null, null, true); //$NON-NLS-1$ //$NON-NLS-2$
		//
		creator.createFileSelector(AppPreferences.MESSAGE_LOG_PATH,
				mm.get("LogConfigContent.pathToMessageLogFile"), //$NON-NLS-1$
				"", "", null, null, true); //$NON-NLS-1$ //$NON-NLS-2$
		creator.createLabel(mm.get("LogConfigContent.noteForMessageLog")); //$NON-NLS-1$
		//
		creator.createCheckBox(AppPreferences.MESSAGE_LOG_UTF8,//
				mm.get("LogConfigContent.saveInUTF8"), //$NON-NLS-1$
				false);
		creator.createLabel(mm.get("LogConfigContent.noteForUTF8.1")); //$NON-NLS-1$
		creator.createLabel(mm.get("LogConfigContent.noteForUTF8.2")); //$NON-NLS-1$
	}

	/**
	 * TabItem̏
	 * Composite̐
	 * CAEg̐ݒ
	 * @param item ^uACe
	 * @return Composite
	 */
	private Composite prepareTabItem(TabItem item) {
		Composite c = new Composite(item.getParent(), SWT.NONE);
		GridLayout layout = new GridLayout();
		layout.numColumns = 2;
		c.setLayout(layout);
		item.setControl(c);

		// ςɍL
		GridData data = new GridData();
		data.horizontalAlignment = SWT.FILL;
		data.grabExcessHorizontalSpace = true;

		c.setLayoutData(data);

		return c;
	}

	/**
	 * OKALZ{^𐶐܂B
	 * @param parent eEBWFbg
	 */
	private void createButtons(Composite parent) {
		// OK
		Button okButton = new Button(parent, SWT.PUSH);
		okButton.setText(mm.get("SWTConfigurationDialog.buttons.ok")); //$NON-NLS-1$
		okButton.addSelectionListener(new OkButton());

		// LZ
		Button cancelButton = new Button(parent, SWT.PUSH);
		cancelButton.setText(mm.get("SWTConfigurationDialog.buttons.cancel")); //$NON-NLS-1$
		cancelButton.addSelectionListener(new CancelButton());

		// ftHg̃{^ݒ
		shell.setDefaultButton(okButton);
	}

	/**
	 * _CAOJ܂B
	 */
	public void open() {
		SWTUtils.moveToCenter(shell);
		this.shell.open();
	}

	/**
	 * OK{^NbNƂ̏B
	 * ݒl擾Aۑ܂B
	 */
	private class OkButton extends SelectionAdapter {
		public void widgetSelected(SelectionEvent event) {

			// ͒l`FbN
			CollectableControl control = creator.checkProperties();
			if (control != null) {
				SWTUtils.showError(shell, control.getReason());
				control.setFocus();
				return;
			}

			// N\
			boolean updated = launchConfig.synchCheckbox();
			launchConfigurationUpdated = updated ? true
					: launchConfigurationUpdated;
			// `FbN
			List confs = launchConfigData.getConfigurations();
			int index = facade.getConflictConfiguration(confs);
			if (index != -1) {
				LaunchConfiguration conf = (LaunchConfiguration) confs
						.get(index);
				conf.setIsValid(false);
				launchConfig.updateTable(index);
				launchConfig.setFocus();

				SWTUtils.showWarning(
						shell,
						mm.get("SWTConfigurationDialog.competitionDetected", //$NON-NLS-1$
								new Object[] { conf.getName(),
										conf.getProtocol() }));
				return;
			}

			// ۑ
			launchConfig.saveProperties();

			// Â[U[ƃO[vL
			String oldName = pref.getUserName();
			String oldGroup = pref.getGroupName();

			// vpeBۑ
			creator.saveProperties();

			pref.setLastPreferenceTabIndex(tabFolder.getSelectionIndex());

			// [U[rAύXĂʒm
			String curUserName = pref.getUserName();
			String curGroupName = pref.getGroupName();

			if (!oldName.equals(curUserName) || !oldGroup.equals(curGroupName)) {
				main.setWindowTitle();
				facade.sendRename();
			}

			// CEBhËʒu
			main.saveWindowStatus();

			// zbgL[̍Đݒ
			main.setupHotkeys();

			// ݒۑ
			try {
				pref.save();

			} catch (IOException e) {
				String msg = mm
						.get("SWTConfigurationDialog.couldNotSaveAppPreference"); //$NON-NLS-1$
				logger.error(msg, e);
				SWTUtils.showError(shell, msg);
			}

			if (launchConfigurationUpdated) {
				// ċN
				main.restartMessengers();
			}

			// 
			shell.close();
		}
	}

	/**
	 * LZ{^NbNƂ̏
	 */
	private class CancelButton extends SelectionAdapter {
		public void widgetSelected(SelectionEvent event) {
			shell.close();
		}
	}

	public void addLaunchConfiguration(LaunchConfiguration conf) {
		launchConfig.addConf(conf);
		launchConfigurationUpdated = true;
	}

	public void updateLaunchConfiguration(LaunchConfiguration conf) {
		launchConfig.updateTable(conf);
		launchConfigurationUpdated = true;
	}
}
