package petaclone.swt;

import java.security.NoSuchAlgorithmException;

import jp.takayuki.swt.util.SWTUtils;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Password;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;

/**
 * pX[hύX_CAO
 * @author Takayuki Shoji
 */
public class SWTPasswordChangeDialog {

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/**
	 * eEBhE
	 */
	private Shell parent;

	/**
	 * pX[hύX_CAOShell
	 */
	private Shell shell;

	/**
	 * ݂̃pX[h
	 */
	private Text current;

	/**
	 * VpX[h
	 */
	private Text newPassword;

	/**
	 * VpX[h(mF)
	 */
	private Text verification;

	/**
	 * _CAÕ^Cg
	 */
	private String dialogTitle;

	/**
	 * pX[h
	 */
	private Password password;

	/**
	 * RXgN^
	 * @param parent eEBWFbg
	 * @param dialogTitle _CAÕ^Cg
	 */
	public SWTPasswordChangeDialog(Shell parent, String dialogTitle) {
		this.parent = parent;
		this.dialogTitle = dialogTitle;
	}

	/**
	 * EBhEJ܂B
	 */
	public void open() {

		this.shell = new Shell(parent, SWT.DIALOG_TRIM | SWT.APPLICATION_MODAL);
		shell.setText(dialogTitle);

		GridLayout layout = new GridLayout();
		layout.numColumns = 2;
		layout.makeColumnsEqualWidth = true;

		shell.setLayout(layout);

		// ݂̃pX[h
		current = createRow("SWTPasswordChangeDialog.currentPassword"); //$NON-NLS-1$
		if (password == null || password.isEmptyPassword()) {
			// ݒ肳ĂȂƂ
			current.setEnabled(false);
		}
		// VpX[h
		newPassword = createRow("SWTPasswordChangeDialog.newPassword"); //$NON-NLS-1$
		// VpX[h(mF)
		verification = createRow("SWTPasswordChangeDialog.newPasswordConfirm"); //$NON-NLS-1$

		//
		// {^
		//
		GridData data = new GridData();
		data.horizontalAlignment = SWT.CENTER;
		data.horizontalSpan = 2;

		Composite composite = new Composite(shell, SWT.NULL);
		composite.setLayoutData(data);

		RowLayout buttonLayout = new RowLayout();
		buttonLayout.pack = false;
		composite.setLayout(buttonLayout);

		Button okButton = new Button(composite, SWT.PUSH);
		okButton.setText(mm.get("SWTPasswordChangeDialog.ok")); //$NON-NLS-1$
		okButton.addSelectionListener(new OkButtonSelected());

		Button cancelButton = new Button(composite, SWT.PUSH);
		cancelButton.setText(mm.get("SWTPasswordChangeDialog.cancel")); //$NON-NLS-1$
		cancelButton.addSelectionListener(new CancelButtonSelected());

		shell.setDefaultButton(okButton);

		shell.pack();
		shell.open();
	}

	/**
	 * xƃeLXg{bNX𐶐܂B
	 * @param labelKey x̃L[
	 * @return eLXg{bNX
	 */
	private Text createRow(String labelKey) {
		Label label = new Label(shell, SWT.NONE);
		label.setText(mm.get(labelKey));
		Text text = new Text(shell, SWT.PASSWORD | SWT.BORDER);
		GridData gdata = new GridData();
		gdata.horizontalAlignment = SWT.FILL;
		text.setLayoutData(gdata);
		return text;
	}

	/**
	 * pX[hݒ肵܂B
	 * @param storedPassword vpeBɕۑꂽpX[h
	 */
	public void setPassword(String storedPassword) {
		this.password = new Password();
		this.password.setStoredPassword(storedPassword);
	}

	/**
	 * pX[h擾܂B
	 * @return pX[h(vpeBۑp)
	 */
	public String getPassword() {
		return this.password.getStoredPassword();
	}

	/**
	 * ͂`FbNA_CAO\܂B
	 * @return false 肪
	 */
	private boolean checkInput() {
		String old = current.getText();
		String new1 = newPassword.getText();
		String new2 = verification.getText();

		// VpX[hmFpƈv邩
		if (!new1.equals(new2)) {
			SWTUtils.showWarning(shell, mm
					.get("SWTPasswordChangeDialog.newPasswordsDoNotMatch")); //$NON-NLS-1$

			newPassword.setFocus();
			newPassword.selectAll();
			return false;
		}

		// ݂̃pX[hĂ邩
		try {
			if (!password.equals(old)) {
				SWTUtils.showWarning(shell, mm
						.get("SWTPasswordChangeDialog.passwordIncorrect")); //$NON-NLS-1$

				current.setFocus();
				current.selectAll();
				return false;
			}
		} catch (NoSuchAlgorithmException e) {
			SWTUtils
					.showError(
							shell,
							mm
									.get("SWTPasswordChangeDialog.passwordEncryptionIsIncompatible")); //$NON-NLS-1$
			return false;
		}

		return true;
	}

	/**
	 * OK̏
	 */
	class OkButtonSelected extends SelectionAdapter {
		public void widgetSelected(SelectionEvent arg0) {
			if (!checkInput()) {
				return;
			}

			// VpX[h擾
			password = new Password(newPassword.getText());

			shell.close();
		}
	}

	/**
	 * LZ̏
	 */
	class CancelButtonSelected extends SelectionAdapter {
		public void widgetSelected(SelectionEvent arg0) {
			shell.close();
		}
	}
}
