package petaclone.swt;

import java.io.File;

import jp.takayuki.swt.util.SWTUtils;
import jp.takayuki.tools.MessageMap;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;

/**
 * t@C̕ύX_CAOłB
 * 
 * @author Takayuki Shoji
 */
public class SWTRenameDialog extends SelectionAdapter {

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/**
	 * EBhE
	 */
	private Shell shell = null;

	/**
	 * ݂̃t@C̃x
	 */
	private Label oldNameLabel = null;

	/**
	 * ݂̃t@C̃eLXg{bNX
	 */
	private Text oldFilenameText = null;

	/**
	 * Vt@C̃x
	 */
	private Label newNameLabel = null;

	/**
	 * Vt@C̃eLXg{bNX
	 */
	private Text newFilenameText = null;

	/**
	 * {^̃R|Wbg
	 */
	private Composite buttonComposite = null;

	/**
	 * OK{^
	 */
	private Button okButton = null;

	/**
	 * LZ{^
	 */
	private Button cancelButton = null;

	/**
	 * eEBhE
	 */
	private final Shell parent;

	/**
	 * Ώۂ̃t@C
	 */
	private File file;

	/**
	 * Ăяo
	 */
	private final Renamable renamable;

	/**
	 * RXgN^
	 * 
	 * @param parent eEBhE
	 * @param renamable Ăяo
	 * @param file l[Ώۂ̃t@C
	 */
	public SWTRenameDialog(Shell parent, Renamable renamable, File file) {
		this.parent = parent;
		this.renamable = renamable;
		this.file = file;
		createShell();
		setFile(file);
	}

	/**
	 * l[Ώۂ̃t@Cݒ肵܂B
	 * 
	 * @param f t@C
	 */
	private void setFile(File f) {
		oldFilenameText.setText(f.getName());
		newFilenameText.setText(f.getName());
		if (f.isFile()) {
			int pos = f.getName().lastIndexOf('.');
			if (pos > 0) {
				newFilenameText.setSelection(0, pos);
			} else if (pos != 0) {
				newFilenameText.selectAll();
			}
		} else {
			newFilenameText.selectAll();
		}
	}

	/**
	 * This method initializes buttonComposite	
	 */
	private void createButtonComposite() {
		GridData gridData4 = new GridData();
		gridData4.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		GridData gridData3 = new GridData();
		gridData3.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		GridLayout gridLayout1 = new GridLayout();
		gridLayout1.numColumns = 2;
		gridLayout1.makeColumnsEqualWidth = true;
		GridData gridData2 = new GridData();
		gridData2.horizontalSpan = 2;
		gridData2.horizontalAlignment = org.eclipse.swt.layout.GridData.CENTER;
		gridData2.grabExcessHorizontalSpace = true;
		buttonComposite = new Composite(shell, SWT.NONE);
		buttonComposite.setLayoutData(gridData2);
		buttonComposite.setLayout(gridLayout1);
		okButton = new Button(buttonComposite, SWT.NONE);
		okButton.setText(mm.get("SWTRenameDialog.ok")); //$NON-NLS-1$
		okButton.setLayoutData(gridData3);
		cancelButton = new Button(buttonComposite, SWT.NONE);
		cancelButton.setText(mm.get("SWTRenameDialog.cancel")); //$NON-NLS-1$
		cancelButton.setLayoutData(gridData4);

		okButton.addSelectionListener(this);
		cancelButton.addSelectionListener(this);
	}

	/**
	 * This method initializes sShell
	 */
	private void createShell() {
		GridData gridData1 = new GridData();
		gridData1.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridData1.grabExcessHorizontalSpace = true;
		GridData gridData = new GridData();
		gridData.grabExcessHorizontalSpace = true;
		gridData.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		GridLayout gridLayout = new GridLayout();
		gridLayout.numColumns = 2;
		shell = new Shell(parent, SWT.DIALOG_TRIM | SWT.APPLICATION_MODAL
				| SWT.SHEET);
		shell.setText(mm.get("SWTRenameDialog.renameFile")); //$NON-NLS-1$
		shell.setLayout(gridLayout);
		shell.setSize(new Point(300, 200));
		shell.setMinimumSize(500, -1);
		oldNameLabel = new Label(shell, SWT.NONE);
		oldNameLabel.setText(mm.get("SWTRenameDialog.currentFilename")); //$NON-NLS-1$
		oldFilenameText = new Text(shell, SWT.BORDER);
		oldFilenameText.setEditable(false);
		oldFilenameText.setLayoutData(gridData);
		newNameLabel = new Label(shell, SWT.NONE);
		newNameLabel.setText(mm.get("SWTRenameDialog.newFilename")); //$NON-NLS-1$
		newFilenameText = new Text(shell, SWT.BORDER);
		newFilenameText.setLayoutData(gridData1);
		createButtonComposite();
		shell.setDefaultButton(okButton);
	}

	/**
	 * EBhEJ܂B
	 */
	public void open() {
		shell.pack();
		newFilenameText.setFocus();
		shell.open();
	}

	public void widgetSelected(SelectionEvent event) {
		if (event.getSource() == cancelButton) {
			shell.close();
		} else if (event.getSource() == okButton) {
			String destStr = newFilenameText.getText();
			File dest = new File(file.getParent(), destStr);
			if (dest.exists() || !file.renameTo(dest)) {
				SWTUtils.showError(shell, mm
						.get("SWTRenameDialog.couldNotRenameFile")); //$NON-NLS-1$
				newFilenameText.setFocus();
			} else {
				renamable.setChangedFile(file, dest);
				shell.close();
			}
		}
	}
}
