package petaclone.swt;

import java.net.UnknownHostException;
import java.util.Arrays;
import java.util.Comparator;
import java.util.regex.Matcher;

import jp.takayuki.logging.Logger;
import jp.takayuki.swt.util.ListenerDelegate;
import jp.takayuki.swt.util.SWTMenuUtils;
import jp.takayuki.swt.util.SWTUtils;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Os;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.ShellAdapter;
import org.eclipse.swt.events.ShellEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.Monitor;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.widgets.Text;

import petaclone.AppPreferences;
import petaclone.Destination;
import petaclone.EventBroker;
import petaclone.Events;
import petaclone.MembersListener;
import petaclone.MessageLogger;
import petaclone.MessengerFacade;
import petaclone.Protocol;
import petaclone.SendingResultListener;
import petaclone.TableViewable;
import petaclone.UniqueMember;
import petaclone.UniqueMemberComparator;
import petaclone.swt.logic.AbstractSWTEventListener;

/**
 * M_CAOB
 * @author Takayuki Shoji
 */
public class SWTSendDialog extends AbstractSWTEventListener implements MembersListener, SendingResultListener,
		Findable {

	/**
	 * logger
	 */
	private static final Logger logger = Logger.getLogger(SWTSendDialog.class);

	/** fobOtO */
	private static final boolean DEBUG = logger.isDebugEnabled();

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/**
	 * CxgXi[
	 */
	private Actions actions = new Actions();

	/** MessengerFacade */
	private MessengerFacade facade;

	/** ݒ */
	private AppPreferences pref;

	/** CEBhE */
	private SWTMainWindow main;

	/** EventBroker */
	private EventBroker broker;

	/** M_CAOShell */
	private Shell shell;

	/** e[u */
	private Table members;

	/** ݓYtĂ郁bZ[WEBhE */
	private SWTMessageWindow message;

	// 萔
	///** J萔: [U[ */
	//private static final int USER_NAME = 0;
	//
	///** J萔: Rs[^ */
	//private static final int COMPUTER_NAME = 1;
	//
	///** J萔: O[v */
	//private static final int GROUP_NAME = 2;

	// {^
	/** M{^ */
	private Button sendButton;

	/** ׂđM{^ */
	private Button sendAllButton;

	/** o^{^ */
	private Button registerButton;

	/** o^{^ */
	private Button unregisterButton;

	/** M */
	private int unsentCount;

	/** M̐m肵ĂȂ */
	private int unsuccessCount;

	/** M邠Đ搔 */
	private int destinationCount;

	/**
	 * bZ[WL^
	 */
	private MessageLogger msgLogger;

	/**
	 * peLXg{bNX
	 */
	private Text searchText;

	/** [}(migemo)[h */
	private boolean useMigemo;

	/**
	 * [U[\郉x
	 */
	private Label userCount;

	/**
	 * [U[XV
	 */
	private UpdateUserCountAction UpdateUserCountAction = new UpdateUserCountAction();

	/**
	 * e[ucomparator
	 */
	private UniqueMemberComparator comparator;

	/**
	 * [U[XV
	 */
	private class UpdateUserCountAction implements Runnable {

		/**
		 * XPW[OĂꍇtrue
		 */
		boolean scheduled;

		/**
		 * @see java.lang.Runnable#run()
		 */
		public void run() {
			setUserCount();
			scheduled = false;
		}

		/**
		 * [U[xɐݒ肵܂B
		 */
		private void setUserCount() {
			Display display = Display.getDefault();
			if (display.isDisposed() || userCount.isDisposed()) {
				return;
			}

			StringBuffer sb = new StringBuffer();
			int selected = members.getSelectionCount();
			if (selected > 0) {
				sb.append(selected);
				sb.append('/');
			}
			sb.append(members.getItemCount());
			sb.append(" users"); //$NON-NLS-1$

			userCount.setText(sb.toString());
			// xXV...
			userCount.getParent().layout();
		}
	}

	/**
	 * [U[ݒ肵܂B
	 * @param selected I𒆂̃[U[
	 * @param all S[U[
	 */
	private void updateUserCount() {
		if (this.UpdateUserCountAction.scheduled)
			return;

		Display display = Display.getDefault();
		if (!display.isDisposed()) {
			Display.getDefault().timerExec(200, this.UpdateUserCountAction);
		}
	}

	/**
	 * RXgN^BNɐĔ\ɂĂB
	 * @param main CEBhE
	 * @param pref AvP[Vݒ
	 * @param facade bZW[t@T[h
	 * @param msgLogger bZ[WL^
	 * @param broker Cxg
	 */
	public SWTSendDialog(SWTMainWindow main, AppPreferences pref,
			MessengerFacade facade, MessageLogger msgLogger, EventBroker broker) {

		this.main = main;
		this.pref = pref;
		this.facade = facade;
		this.msgLogger = msgLogger;
		this.broker = broker;
		this.useMigemo = pref.getUseMigemoInSendDialog();

		if (useMigemo) {
			main.attachMigemoDictionaries(true);
		}

		// EBhE̍쐬
		this.shell = new Shell(this.main.getShell(), SWT.RESIZE
				| SWT.DIALOG_TRIM | SWT.APPLICATION_MODAL);
		this.shell.setVisible(false);
		if (!Os.isMacOSX())
			this.shell.setImage(main.getImage("sendto.gif")); //$NON-NLS-1$

		GridLayout gridLayout = new GridLayout();
		gridLayout.numColumns = 2;
		gridLayout.verticalSpacing = 0;
		gridLayout.marginHeight = 0;
		this.shell.setLayout(gridLayout);

		this.shell.setText(mm.get("SWTSendDialog.title")); //$NON-NLS-1$

		// N[Y{bNX(ȂŉB)
		shell.addShellListener(new ShellAdapter() {
			public void shellClosed(ShellEvent event) {
				event.doit = false;
				cancel();
			}
		});

		// j[̐
		createMenuBar();

		// peLXg{bNX
		createTextBox();

		// [U[\x
		createUserCount();

		// EBWFbg̔zu
		// Ƀe[u
		this.members = createTable();

		searchText.addKeyListener(new ArrowDownBehavior(members));

		// {^̍쐬
		createButtons();

		// {^̏Ԃ
		setButtonStatus();

		this.shell.pack();
		this.shell.setMinimumSize(SWT.DEFAULT, this.shell.getSize().y);

		// o^o[e[uɔf
		updateTable();

		// TCY
		if (pref.getSaveSendDialogSize()) {
			int w = pref.getSendDialogWidth();
			int h = pref.getSendDialogHeight();

			if (w != AppPreferences.IMPOSSIBLE
					&& h != AppPreferences.IMPOSSIBLE && w > 0 && h > 0) {
				this.shell.setSize(w, h);
			}
		}

		// ʒu
		if (pref.getSaveSendDialogLocation()) {
			int x = pref.getSendDialogLeft();
			int y = pref.getSendDialogTop();
			if (x != AppPreferences.IMPOSSIBLE
					&& y != AppPreferences.IMPOSSIBLE) {
				Monitor[] monitors = Display.getDefault().getMonitors();
				if (SWTUtils.getMonitorIndexByCoordinate(monitors, x, y) != -1) {
					shell.setLocation(x, y);
				}
			}
		}

		main.members.addListener(this);
	}

	/**
	 * UȈԂXV܂B
	 * @param deselectAll o[̑Iꍇtrue
	 */
	private void updateUIStatus(boolean deselectAll) {
		if (deselectAll) {
			members.deselectAll();
		}
		setButtonStatus();
		updateUserCount();
	}

	/**
	 * [U[\郉x𐶐܂B
	 */
	private void createUserCount() {
		GridData userCountGd = new GridData();
		userCountGd.horizontalAlignment = SWT.RIGHT;

		userCount = new Label(shell, SWT.NONE);
		userCount.setLayoutData(userCountGd);
	}

	/**
	 * peLXg{bNX𐶐܂B
	 */
	private void createTextBox() {
		Composite box = new Composite(shell, SWT.NONE);

		GridLayout boxLayout = new GridLayout();
		boxLayout.numColumns = 3;
		boxLayout.marginHeight = 0;
		boxLayout.horizontalSpacing = 3;

		GridData boxGd = new GridData();
		boxGd.horizontalAlignment = GridData.FILL;
		boxGd.verticalAlignment = GridData.CENTER;

		box.setLayout(boxLayout);
		box.setLayoutData(boxGd);

		searchText = FindUtils.createSearchBox(box, actions, true, this);
	}

	/**
	 * j[𐶐܂B
	 */
	private void createMenuBar() {

		Menu menuBar = SWTMenuUtils.addMenuBar(shell);

		Menu menuFile = SWTMenuUtils.addCascade(menuBar, mm
				.get("SWTSendDialog.menu.file")); //$NON-NLS-1$

		SWTMenuUtils.addMenuItem(menuFile, //
				mm.get("SWTSendDialog.menu.file.close"), //$NON-NLS-1$
				"ESC", 0, null, //$NON-NLS-1$
				actions, "doClose"); //$NON-NLS-1$

		Menu menuEdit = SWTMenuUtils.addCascade(menuBar, mm
				.get("SWTSendDialog.menu.edit")); //$NON-NLS-1$

		SWTMenuUtils.addMenuItem(menuEdit,//
				mm.get("SWTSendDialog.menu.edit.copy"), //$NON-NLS-1$
				"Ctrl+C", SWT.MOD1 | 'C', //$NON-NLS-1$
				null, actions, "doCopy"); //$NON-NLS-1$

		SWTMenuUtils.addMenuItem(menuEdit,//
				mm.get("SWTSendDialog.menu.edit.selectAll"), //$NON-NLS-1$
				"Ctrl+A", SWT.MOD1 | 'A', null, //$NON-NLS-1$
				actions, "doSelectAll"); //$NON-NLS-1$

		SWTMenuUtils.addSeparator(menuEdit);

		FindUtils.createMenu(menuEdit, actions);

		SWTMenuUtils.addSeparator(menuEdit);

		SWTMenuUtils.addMenuItemCheckbox(menuEdit,//
				mm.get("SWTSendDialog.menu.edit.migemo"), //$NON-NLS-1$
				"Ctrl+K", SWT.MOD1 | 'K', null, //$NON-NLS-1$
				actions, "doCheckMigemo", //$NON-NLS-1$
				pref.getUseMigemoInSendDialog());

		Menu menuView = SWTMenuUtils.addCascade(menuBar, mm
				.get("SWTSendDialog.menu.view")); //$NON-NLS-1$

		SWTMenuUtils.addMenuItem(menuView,//
				mm.get("SWTSendDialog.menu.view.refresh"), //$NON-NLS-1$
				"F5", SWT.F5, null, //$NON-NLS-1$
				actions, "doSearch"); //$NON-NLS-1$

		SWTMenuUtils.addMenuItem(menuView,//
				mm.get("SWTSendDialog.menu.view.additionalSearch"), //$NON-NLS-1$
				"Shift+F5", SWT.MOD2 | SWT.F5, null, //$NON-NLS-1$
				actions, "doAdditionalSearch"); //$NON-NLS-1$
	}

	/**
	 * EBhËʒuvpeBɕۑ܂B
	 * ݒɊւ炸ɕۑ܂B
	 */
	public void saveWindowStatus() {
		logger.debug("saveWindowStatus() called."); //$NON-NLS-1$

		Rectangle rect = shell.getBounds();
		pref.setSendDialogLeft(rect.x);
		pref.setSendDialogTop(rect.y);
		pref.setSendDialogWidth(rect.width);
		pref.setSendDialogHeight(rect.height);

		// ̕
		int count = members.getColumnCount();
		TableColumn col = members.getColumn(0);
		StringBuffer sb = new StringBuffer();
		sb.append(col.getWidth());
		for (int i = 1; i < count; i++) {
			TableColumn column = members.getColumn(i);
			sb.append(',');
			sb.append(column.getWidth());
		}
		pref.setSendDialogColumnWidths(sb.toString());

		// migemo
		pref.setUseMigemoInSendDialog(useMigemo);

		// \[g
		TableColumn sortColumn = members.getSortColumn();
		if (sortColumn != null) {
			pref.setSendDialogSortColumn(SWTUtils.getColumnIndex(sortColumn));
			int direction;
			switch (members.getSortDirection()) {
			case SWT.UP:
				direction = 1;
				break;
			case SWT.DOWN:
				direction = -1;
				break;
			case SWT.DEFAULT:
			default:
				direction = 0;
				break;
			}
			pref.setSendDialogSortOrder(direction);
		} else {
			pref.setSendDialogSortColumn(-1);
			pref.setSendDialogSortOrder(0);
		}
	}

	/**
	 * {^̍쐬
	 */
	private void createButtons() {
		// EComposite
		Composite box = new Composite(shell, SWT.NULL);

		GridData data = new GridData();
		data.verticalAlignment = SWT.TOP;
		data.horizontalAlignment = SWT.RIGHT;
		box.setLayoutData(data);

		GridLayout layout = new GridLayout();
		layout.marginHeight = 0;
		layout.marginWidth = 0;
		layout.horizontalSpacing = 0;
		layout.verticalSpacing = 3;
		box.setLayout(layout);

		this.sendButton = new Button(box, SWT.PUSH);
		sendButton.setText(mm.get("SWTSendDialog.button.send")); //$NON-NLS-1$
		sendButton.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL));
		actions.registerAndAddListener(sendButton, "doSend"); //$NON-NLS-1$

		this.sendAllButton = new Button(box, SWT.PUSH);
		sendAllButton.setText(mm.get("SWTSendDialog.button.sendToAll")); //$NON-NLS-1$
		sendAllButton
				.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL));
		actions.registerAndAddListener(sendAllButton, "doSendToAll"); //$NON-NLS-1$

		// Xy[T
		createSpacer(box);

		Button search = new Button(box, SWT.PUSH);
		search.setText(mm.get("SWTSendDialog.button.refresh")); //$NON-NLS-1$
		search.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL));
		actions.registerAndAddListener(search, "doSearch"); //$NON-NLS-1$

		// o^EҏW
		this.registerButton = new Button(box, SWT.PUSH);
		registerButton.setLayoutData(new GridData(
				GridData.HORIZONTAL_ALIGN_FILL));
		actions.registerAndAddListener(registerButton, "doRegister"); //$NON-NLS-1$

		// o^
		this.unregisterButton = new Button(box, SWT.PUSH);
		this.unregisterButton
				.setText(mm.get("SWTSendDialog.button.unregister")); //$NON-NLS-1$
		unregisterButton.setLayoutData(new GridData(
				GridData.HORIZONTAL_ALIGN_FILL));
		actions.registerAndAddListener(unregisterButton, "doUnregister"); //$NON-NLS-1$

		createSpacer(box);

		Button selectGroup = new Button(box, SWT.PUSH);
		selectGroup.setText(mm.get("SWTSendDialog.button.selectGroup")); //$NON-NLS-1$
		selectGroup.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL));
		actions.registerAndAddListener(selectGroup, "doSelectGroup"); //$NON-NLS-1$

		createSpacer(box);

		Button cancel = new Button(box, SWT.PUSH);
		cancel.setText(mm.get("SWTSendDialog.button.close")); //$NON-NLS-1$
		cancel.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL));
		actions.registerAndAddListener(cancel, "doClose"); //$NON-NLS-1$
	}

	/**
	 * 7sNZ̃Xy[X܂B
	 * @param parent eEBWFbg
	 */
	private void createSpacer(Composite parent) {
		Composite spacer = new Composite(parent, SWT.NONE);
		spacer.setVisible(false);
		GridData separatorgd = new GridData();
		separatorgd.heightHint = 7;
		spacer.setLayoutData(separatorgd);
	}

	/**
	 * M{^̗L؂ւ܂B
	 * IĂȂƂɑSɑM\B
	 * o[ȂƂ͑MsB
	 * bZ[ŴƂɑMł邩ǂlB
	 */
	private void setButtonStatus() {
		if (message == null || members.getItemCount() == 0) {
			sendButton.setEnabled(false);

			sendAllButton.setEnabled(false);

			registerButton.setText(mm.get("SWTSendDialog.button.registerNew")); //$NON-NLS-1$
			unregisterButton.setEnabled(false);
		} else if (members.getSelectionCount() == 0) {
			sendButton.setEnabled(false);

			if (!pref.getNeverUseSendAll() && message.isSendableMessage()) {
				sendAllButton.setEnabled(true);
			}

			registerButton.setText(mm.get("SWTSendDialog.button.registerNew")); //$NON-NLS-1$
			unregisterButton.setEnabled(false);
		} else {
			if (message.isSendableMessage()) {
				sendButton.setEnabled(true);
			} else {
				sendButton.setEnabled(false);
			}

			sendAllButton.setEnabled(false);

			TableItem item = members.getItem(members.getSelectionIndex());
			UniqueMember um = (UniqueMember) item.getData();
			if (main.members.isRegistered(um)) {
				// o^ς݂Ȃ"ҏW..."
				registerButton.setText(mm.get("SWTSendDialog.button.edit")); //$NON-NLS-1$
				unregisterButton.setEnabled(true);
			} else {
				// o^Ȃ"o^..."
				registerButton.setText(mm
						.get("SWTSendDialog.button.registerSelection")); //$NON-NLS-1$
				unregisterButton.setEnabled(false);
			}
		}
	}

	/**
	 * e[u̍쐬
	 * @return 쐬ꂽe[u
	 */
	private Table createTable() {
		Table result = new Table(shell, SWT.BORDER | SWT.MULTI
				| SWT.FULL_SELECTION | SWT.H_SCROLL | SWT.V_SCROLL);
		// TCY
		GridData data = new GridData();
		data.horizontalAlignment = SWT.FILL;
		data.verticalAlignment = SWT.FILL;
		data.grabExcessHorizontalSpace = true;
		data.grabExcessVerticalSpace = true;

		result.setLayoutData(data);
		// wb_
		result.setHeaderVisible(true);

		// Comparator
		comparator = new UniqueMemberComparator();
		int sortOrder = pref.getSendDialogSortOrder();
		int sortColumn = pref.getSendDialogSortColumn();
		comparator.setSortDirection(sortOrder);
		comparator.setSortColumn(sortColumn);
		comparator.setGroup(pref.getGroupName());
		TableSortListener listener = new TableSortListener();

		// 
		int[] widths = pref.getSendDialogColumnWidths();
		if (widths == null || widths.length != 4) {
			widths = new int[] { 100, 100, 100, 100 };
		}
		SWTUtils.createTableColumn(result, listener, widths[0],
				"SWTSendDialog.col.name"); //$NON-NLS-1$
		SWTUtils.createTableColumn(result, listener, widths[1],
				"SWTSendDialog.col.computer"); //$NON-NLS-1$
		SWTUtils.createTableColumn(result, listener, widths[2],
				"SWTSendDialog.col.group"); //$NON-NLS-1$
		SWTUtils.createTableColumn(result, listener, widths[3],
				"SWTSendDialog.col.ipaddress"); //$NON-NLS-1$

		if (0 <= sortColumn && sortColumn <= 3
				&& (sortOrder == -1 || sortOrder == 1)) {
			result.setSortColumn(result.getColumn(sortColumn));
			result.setSortDirection(sortOrder > 0 ? SWT.UP : SWT.DOWN);
		}

		// IԂɂă{^̏ԂύX郊Xi[
		result.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent event) {
				updateUIStatus(false);
				if (DEBUG) {
					TableItem item = (TableItem) event.item;
					if (item != null) {
						UniqueMember m = (UniqueMember) item.getData();
						logger.debug(m);
					}
				}
			}
		});
		// ACeȊONbNI
		result.addMouseListener(new MouseAdapter() {
			public void mouseDown(MouseEvent event) {
				if (members.getItem(new Point(event.x, event.y)) == null) {
					updateUIStatus(true);
				}
			}
		});

		result.setMenu(createPopupMenu());

		return result;
	}

	/**
	 * |bvAbvj[𐶐܂B
	 * 
	 * @return Menu
	 */
	private Menu createPopupMenu() {
		Menu popup = new Menu(this.shell, SWT.POP_UP);
		MenuItem item = new MenuItem(popup, SWT.PUSH);
		item.setText(mm.get("SWTSendDialog.menu.edit.copy")); //$NON-NLS-1$
		actions.registerAndAddListener(item, "doCopy"); //$NON-NLS-1$
		return popup;
	}

	/**
	 * w肳ꂽbZ[Wɑ΂đM_CAOJĂ邩ǂԂ܂B
	 * 
	 * @param message bZ[WEBhE
	 * @return JĂꍇtrue
	 */
	public boolean isShown(SWTMessageWindow message) {
		return this.message == message;
	}

	/**
	 * _CAO\܂B
	 * @param message bZ[WEBhE
	 */
	public void show(SWTMessageWindow message) {
		boolean forceActive = Os.isOther();
		if (forceActive || Os.isMacOSX()) {
			/*
			 * Mac OS Xł́ACEBhEŏĂƂ
			 * M_CAO\Ȃ̂ŖB
			 * Ubuntuł́ACEBhEŏĂƂ
			 * M_CAO\悤Ƃƃt[YB
			 */
			main.show(forceActive);
		}
		this.message = message;
		message.setVisibleIfOnTop(false);
		updateUIStatus(false);
		this.shell.setActive(); //SWT2.1ŕKv
		this.shell.setVisible(true);
		// ̓[h
		if (pref.useImInSendDialog()) {
			int mode = SWT.DBCS | SWT.NATIVE;
			if (pref.getImModeIsRoman()) {
				mode |= SWT.ROMAN;
			}
			shell.setImeInputMode(mode);
		}

		this.searchText.setFocus();
	}

	/**
	 * _CAO\ɂ܂B
	 * IԂBbZ[WEBhE̎QƂO܂B
	 * @param closeMessageWindow bZ[WEBhEꍇtrue
	 */
	private void hide(boolean closeMessageWindow) {
		this.shell.setVisible(false);
		this.members.deselectAll();
		this.searchText.setText(""); //$NON-NLS-1$
		this.shell.setEnabled(true);

		if (message == null) {
			return;
		}
		// OɎQƂ͂Ă
		SWTMessageWindow mw = message;
		this.message = null;

		if (closeMessageWindow) {
			mw.close();
		} else {
			mw.setVisibleIfOnTop(true);
			mw.activate();
		}
	}

	/**
	 * M{^BIꂽMɑM܂B
	 * M_CAO\ɂ܂B
	 * M̑IAsM悾IԂɂ܂B
	 */
	private void send() {

		if (!facade.isSendable()) {
			SWTUtils.showError(shell, mm
					.get("SWTSendDialog.messengerIsNotLaunched")); //$NON-NLS-1$

			return;
		}

		// IĂȂƂ͉Ȃ
		if (members.getSelectionCount() == 0) {
			return;
		}
		// MI܂ŃbN
		shell.setEnabled(false);

		// Đ쐬
		TableItem[] items = members.getSelection();
		UniqueMember[] ums = new UniqueMember[items.length];
		for (int i = 0; i < items.length; i++) {
			ums[i] = (UniqueMember) items[i].getData();
		}

		// Đ̐L
		unsentCount = items.length;
		unsuccessCount = unsentCount;
		destinationCount = unsentCount;

		// MBʂ̓Xi[Œʒm
		facade.sendMessage(message.getMessage(), ums, this);
	}

	/**
	 * SɑM{^BݒmĂ邷ׂẴ[U[ɑM܂B
	 * M_CAO\ɂ܂B
	 * 񂷂ׂđIAM̑IA
	 * sM悾IԂɂ܂B
	 */
	private void sendAll() {
		// e[uׂđI
		members.selectAll();
		send();
	}

	/**
	 * 񓯊s\񂵂܂B
	 * 
	 * @param runnable R[obN
	 */
	private void asynchExec(Runnable runnable) {
		Display display = main.getShell().getDisplay();
		if (!display.isDisposed()) {
			display.asyncExec(runnable);
		}
	}

	/**
	 * Mʂ󂯎܂B
	 */
	public void onSendResult(Destination dest, Exception e) {

		if (message == null || message.getMessage() == null || unsentCount <= 0) {
			logger.error("onSendResult(): Illegal callback. unsentCount: " //$NON-NLS-1$
					+ unsentCount + ", message == null ? " + (message == null)); //$NON-NLS-1$
			return;
		}
		asynchExec(new OnSendResult(dest, e));
	}

	/**
	 * Mʂ\ɔf܂B
	 */
	private class OnSendResult implements Runnable {
		/**
		 * ML^
		 */
		private Destination dest;

		/** MɔO */
		private Exception e;

		/**
		 * RXgN^
		 * @param dest ML^
		 * @param e MɔO
		 */
		OnSendResult(Destination dest, Exception e) {
			this.dest = dest;
			this.e = e;
		}

		public void run() {
			unsentCount--;
			if (dest.getResult()) {
				unsuccessCount--;
			}
			message.getMessage().addDestination(dest);

			// MIĐM΃OL^
			if (unsentCount <= 0 && destinationCount != unsuccessCount) {

				// IP MessengerO
				if (pref.getLogEnabled()) {
					msgLogger.writeMessageLog(SWTSendDialog.this.message
							.getMessage());
				}

				// ۊǔ
				if (pref.getSaveSentMessage()) {
					main.savedMessageWindow.add(message.getMessage());
				}
			}

			// Ȃ猟đI
			if (dest.getResult()) {
				final int count = members.getItemCount();
				for (int i = 0; i < count; i++) {
					UniqueMember um = (UniqueMember) members.getItem(i)
							.getData();
					if (um == null || um.getIdentifier() == null
							|| um.getIdentifier().length() == 0) {
						logger.warn("um==null or " //$NON-NLS-1$
								+ "um.getIdentifier() == null or empty i=" + i //$NON-NLS-1$
								+ ", um=" + um); //$NON-NLS-1$
						continue;
					}
					if (um.getIdentifier().equals(dest.getIdentifier())) {
						members.deselect(i);
						break;
					}
				}
				// ׂĐȂ
				if (unsentCount == 0 && unsuccessCount <= 0) {
					// _CAOB
					logger.info("Send dialog closing."); //$NON-NLS-1$
					hide(true);
				}
			} else {
				// G[\
				if (e != null) {
					if (e instanceof UnknownHostException) {
						String msg = mm.get(
								"SWTSendDialog.destinationComputerIsNotFound", //$NON-NLS-1$
								new Object[] { e.getMessage() });
						logger.warn(msg, e);
						SWTUtils.showWarning(shell, msg);
					} else {
						String msg = mm.get("SWTSendDialog.failedToSend", //$NON-NLS-1$
								new Object[] { e.getMessage() });
						logger.warn(msg, e);
						SWTUtils.showWarning(shell, msg);
					}

				} else {

					String msg = mm
							.get("SWTSendDialog.errorOccurredWhileSending"); //$NON-NLS-1$
					logger.warn(msg, e);
					SWTUtils.showWarning(shell, msg);
				}
				// I_CAÕbN
				if (unsentCount <= 0) {
					shell.setEnabled(true);
					members.setFocus();
				}
			}
			updateUIStatus(false);
		}
	}

	/**
	 * bZ[WꍇɌĂяo܂B
	 */
	public void onTooLongMessage(Protocol protocol, UniqueMember member,
			int length) {
		asynchExec(new OnTooLongMessage(shell, protocol, member, length));
	}

	/**
	 * Č{^B
	 * e[uNAUDPT[oɍČNGXg܂B
	 * @param addOnly truêƂXgNA܂B
	 */
	private void search(boolean addOnly) {
		if (DEBUG)
			logger.debug("Refreshing member list. addOnly=" + addOnly); //$NON-NLS-1$
		comparator.setGroup(pref.getGroupName());
		if (!addOnly) {
			// o[Xg
			main.members.initUniqueMembers();
			updateTable();
		}
		// ČNGXg
		facade.sendFind(addOnly);
		updateUIStatus(false);
	}

	/**
	 * o[폜ďԂ܂B
	 */
	public void reset() {
		main.members.initUniqueMembers();
		updateTable();
		setButtonStatus();
		updateUserCount();
	}

	/**
	 * LZ{^B
	 * sendDialog\ɂ܂B
	 */
	private void cancel() {
		hide(false);
	}

	/**
	 * Xe[^XACR擾܂B
	 * @param status UniqueMember.OFFLINE, UniqueMember.ABSENCE, 
	 * 	UniqueMember.ABSENCE_FILE, UniqueMember.ONLINE_FILE
	 * @return ImageIuWFNg
	 */
	private Image getStatusIcon(int status) {
		String file = null;
		switch (status) {
		case UniqueMember.OFFLINE:
			file = "status_offline.png"; //$NON-NLS-1$
			break;
		case UniqueMember.ABSENCE:
			file = "status_absence.png"; //$NON-NLS-1$
			break;
		case UniqueMember.ABSENCE_FILE:
			file = "status_absence_file.png"; //$NON-NLS-1$
			break;
		case UniqueMember.ONLINE:
			file = "status_online.png"; //$NON-NLS-1$
			break;
		case UniqueMember.ONLINE_FILE:
			file = "status_online_file.png"; //$NON-NLS-1$
			break;
		default:
			assert false;
			break;
		}
		return main.getImage(file);
	}

	/**
	 * uniqueMembersɊÂăe[uXV܂B
	 */
	private void updateTable() {
		int count = members.getItemCount();
		UniqueMember[] ums = main.members.getAllMembers();
		Arrays.sort(ums, comparator);
		if (ums.length < count) {
			members.remove(ums.length, count - 1);
		} else if (ums.length > count) {
			for (int i = count; i < ums.length; i++) {
				TableItem item = new TableItem(members, SWT.LEFT, i);
				updateRow(item, ums[i]);
			}
		}
		int min = Math.min(count, ums.length);

		for (int i = 0; i < min; i++) {
			updateRow(i, ums[i]);
		}
	}

	/**
	 * w肳ꂽUniqueMember̕\XV܂B
	 * @param um XV郁o[
	 */
	private void updateRow(UniqueMember um) {
		final int count = members.getItemCount();
		for (int i = 0; i < count; i++) {
			TableItem item = members.getItem(i);
			UniqueMember curMember = (UniqueMember) item.getData();
			if (um.equals(curMember)) {
				updateRow(item, um);
				return;
			}
		}
		if (DEBUG)
			logger.debug("Not exists:" + um); //$NON-NLS-1$
	}

	/**
	 * w肳ꂽs̕\XV܂B
	 * @param row XVe[u̍sԍ
	 * @param newData Vgpf[^B
	 * 	null̏ꍇ͍sɐݒ肳Ăf[^gp܂B
	 */
	private void updateRow(int row, UniqueMember newData) {
		TableItem item = members.getItem(row);
		updateRow(item, newData);
	}

	/**
	 * w肳ꂽs̕\XV܂B
	 * @param item e[u̍s
	 * @param newData Vgpf[^B
	 * 	null̏ꍇ͍sɐݒ肳Ăf[^gp܂B
	 */
	private void updateRow(TableItem item, UniqueMember newData) {
		UniqueMember um;
		if (newData == null) {
			um = (UniqueMember) item.getData();
		} else {
			um = newData;
			item.setData(um);
		}
		// \XV
		// MEMO AhX\
		//if (false) {
		//	String[] info = um.getRowStrings();
		//	info[1] = "(private)"; //$NON-NLS-1$
		//	info[3] = "(private)"; //$NON-NLS-1$
		//	item.setText(info);
		//} else {
		item.setText(um.getRowStrings());
		//}
		// Xe[^XACRXV
		int status = um.getStatus();
		Image img = getStatusIcon(status);
		item.setImage(0, img);
		// F
		if (status == UniqueMember.OFFLINE) {
			item.setForeground(shell.getDisplay().getSystemColor(
					SWT.COLOR_WIDGET_NORMAL_SHADOW));
		} else {
			item.setForeground(shell.getDisplay().getSystemColor(
					SWT.COLOR_LIST_FOREGROUND));
		}
	}

	/**
	 * e[uڂ̃\[g
	 */
	private class TableSortListener extends SelectionAdapter {

		public void widgetSelected(SelectionEvent event) {

			int index = SWTUtils.getColumnIndex((TableColumn) event.widget);

			int currentSortedColumn = SWTUtils.getColumnIndex(members
					.getSortColumn());
			int direction;
			switch (members.getSortDirection()) {
			case SWT.UP:
				direction = 1;
				break;
			case SWT.DOWN:
				direction = -1;
				break;
			case SWT.NONE:
			default:
				direction = 0;
				break;
			}
			if (index == currentSortedColumn) {
				// 0 -> 1, 1 -> -1, -1 -> 0
				direction = ((direction + 2) % 3) - 1;
			} else {
				direction = 1;
			}

			// e[ȕԂɉă\[g߂
			comparator.setSortColumn(index);
			comparator.setSortDirection(direction);
			if (direction == 0) {
				members.setSortDirection(SWT.NONE);
				return;
			}
			// \[g
			members.deselectAll();
			members.setSortColumn((TableColumn) event.widget);
			members.setSortDirection(direction > 0 ? SWT.UP : SWT.DOWN);
			comparator.setGroup(pref.getGroupName());
			TableItem[] items = members.getItems();
			TableViewable[] tvs = new TableViewable[items.length];
			for (int i = 0; i < items.length; i++) {
				tvs[i] = (TableViewable) items[i].getData();
			}
			Arrays.sort(tvs, comparator);
			for (int i = 0; i < tvs.length; i++) {
				updateRow(items[i], (UniqueMember) tvs[i]);
			}
		}
	}

	/**
	 * e[u폜܂B
	 * @see petaclone.MembersListener#deleteMember(UniqueMember)
	 */
	public void deleteMember(UniqueMember um) {
		asynchExec(new DeleteMember(um));
	}

	/** 폜 */
	private class DeleteMember implements Runnable {
		/** o[ */
		private UniqueMember um;

		/**
		 * RXgN^
		 * @param um o[
		 */
		DeleteMember(UniqueMember um) {
			this.um = um;
		}

		public void run() {
			if (members.isDisposed()) {
				return;
			}
			String id = um.getIdentifier();
			final int count = members.getItemCount();
			for (int i = 0; i < count; i++) {
				UniqueMember curMember = (UniqueMember) members.getItem(i)
						.getData();
				if (id.equals(curMember.getIdentifier())) {
					members.remove(i);
					updateUIStatus(false);
					return;
				}
			}
		}
	}

	/**
	 * 	e[uɒǉ܂B
	 * @see petaclone.MembersListener#newMember(UniqueMember)
	 */
	public void newMember(UniqueMember um) {
		asynchExec(new NewMember(um));
	}

	/** ǉ */
	private class NewMember implements Runnable {
		/** o[ */
		private UniqueMember um;

		/**
		 * RXgN^
		 * @param um ǉ悤Ƃ郁o[
		 */
		NewMember(UniqueMember um) {
			this.um = um;
		}

		public void run() {
			if (members.isDisposed()) {
				return;
			}

			// \[g
			int column = comparator.getSortColumn();
			int index;
			if (0 <= column && column <= 3
					&& comparator.getSortDirection() != 0) {
				// }ʒu
				index = binarySearch(members, um, comparator);
				if (index < 0) {
					index = index * -1 - 1;
				}
			} else {
				index = members.getItemCount();
			}

			TableItem item = new TableItem(members, SWT.LEFT, index);
			updateRow(item, um);
			updateUserCount();
		}

		/**
		 * 񕪌s܂B
		 * @param table 
		 * @param um
		 * @param comparator
		 * @return ʒu܂͑}ʒu
		 * @see "http://en.wikipedia.org/wiki/Binary_search_algorithm"
		 */
		private int binarySearch(Table table, UniqueMember um,
				Comparator comparator) {
			int min = 0;
			int max = table.getItemCount() - 1;

			while (max >= min) {
				int mid = (min + max) >>> 1;
				UniqueMember um2 = (UniqueMember) table.getItem(mid).getData();
				int d = comparator.compare(um2, um);
				if (d < 0) {
					min = mid + 1;
				} else if (d > 0) {
					max = mid - 1;
				} else {
					return mid;
				}
			}
			return -(min + 1);
		}
	}

	/**
	 * e[u̕\XV܂B
	 * @see petaclone.MembersListener#updateMember(UniqueMember)
	 */
	public void updateMember(UniqueMember um) {
		asynchExec(new UpdateMember(um));
	}

	/** XV */
	private class UpdateMember implements Runnable {
		/**
		 * o[
		 */
		private UniqueMember um;

		/**
		 * RXgN^
		 * @param um XV悤Ƃ郁o[
		 */
		UpdateMember(UniqueMember um) {
			this.um = um;
		}

		public void run() {
			if (members.isDisposed()) {
				return;
			}

			int sortOrder = comparator.getSortColumn();
			if (sortOrder == 0) {
				updateRow(um);
			} else {
				// TODO \[g - ǂXVꂽ킩Ȃ̂őS\[gKv
				updateRow(um);
			}
		}
	}

	/**
	 * O[ṽo[I܂B
	 * O[vɏĂȂꍇ͑I܂B
	 */
	private void selectSameGroup() {

		String group;

		int selection = members.getSelectionIndex();
		if (selection >= 0) {
			UniqueMember um = (UniqueMember) members.getItem(selection)
					.getData();
			group = um.getGroup();
			if (group == null) {
				group = ""; //$NON-NLS-1$
			}
		} else {
			group = pref.getGroupName();
		}

		members.deselectAll();

		int count = members.getItemCount();
		boolean first = true;
		for (int i = 0; i < count; i++) {
			UniqueMember um = (UniqueMember) members.getItem(i).getData();
			if (group.equals(um.getGroup())) {
				if (first) {
					members.setSelection(i);
					first = false;
				} else {
					members.select(i);
				}
			}
		}
		members.showSelection();
		updateUIStatus(false);
		members.setFocus();
	}

	/**
	 * peLXg{bNXɓ͂ꂽ
	 * v郁o[ׂđI܂B
	 */
	private void selectAllMatchedMembers() {
		if (members.getItemCount() == 0) {
			return;
		}
		String key = searchText.getText();
		if (key == null || key.length() == 0) {
			updateUIStatus(true);
			return;
		}
		Matcher regex = null;
		if (useMigemo) {
			regex = main.compileMigemo(key);
		}
		members.deselectAll();
		final int count = members.getItemCount();
		int index = 0;
		boolean first = true;
		while ((index = find(key, regex, index, 1)) != -1) {
			if (first) {
				members.setSelection(index);
				first = false;
			} else {
				members.select(index);
			}
			index++;
			if (index >= count) {
				break;
			}
		}
		updateUIStatus(false);
		members.setFocus();
	}

	/**
	 * Io[Nbv{[hɃRs[܂B
	 */
	private void copySelectedMembers() {
		TableItem[] items = members.getSelection();
		if (items.length == 0) {
			return;
		}
		final String eol = System.getProperty("line.separator"); //$NON-NLS-1$
		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < items.length; i++) {
			TableItem item = items[i];
			UniqueMember um = (UniqueMember) item.getData();
			sb.append(um.getUserName());
			sb.append('\t');
			sb.append(um.getComputerName());
			sb.append('\t');
			sb.append(um.getGroup());
			sb.append('\t');
			sb.append(um.getIpAddress());
			if (i + 1 < items.length) {
				sb.append(eol);
			}
		}
		SWTUtils.setTextToClipboard(sb.toString());
	}

	public void findItem(String search, int start, boolean reverse) {
		if (search == null || search.length() == 0) {
			updateUIStatus(true);
			return;
		}
		// migemoŐK\ɕϊ
		Matcher regex = null;
		if (useMigemo) {
			regex = main.compileMigemo(search);
		}
		int index;
		if (reverse) {
			index = find(search, regex, start, -1);
		} else {
			index = find(search, regex, start, 1);
		}
		if (index == -1) {
			updateUIStatus(true);
			return;
		}
		members.setSelection(index);
		updateUIStatus(false);
	}

	/**
	 * 
	 * @param search 
	 * @param m K\
	 * @param start Jnʒu
	 * @param direction (1)܂͏(-1)
	 * @return ʒuB݂Ȃꍇ-1B
	 */
	private int find(String search, Matcher m, int start, int direction) {
		String lower = search.toLowerCase();

		final int columnCount = members.getColumnCount();
		final int itemCount = members.getItemCount();
		for (int i = start; 0 <= i && i < itemCount; i += direction) {
			TableItem item = members.getItem(i);
			for (int c = 0; c < columnCount; c++) {
				String s = item.getText(c).toLowerCase();
				if (m != null) {
					m.reset(s);
					if (m.find()) {
						return i;
					}
				} else {
					if (s.indexOf(lower) != -1) {
						return i;
					}
				}
			}
		}
		return -1;
	}

	/**
	 * SelectionListener̃Zbg
	 */
	public final class Actions extends ListenerDelegate {

		/**
		 * O
		 * @see jp.takayuki.swt.util.ListenerDelegate#handleException(java.lang.Exception)
		 */
		protected void handleException(Exception e) {
			broker.notifyEvent(Events.UnexpectedErrorOccured, e);
		}

		/**
		 * I{^
		 * <p>
		 * }b`郁o[ׂđI܂B
		 * @param event Cxg
		 */
		public void doSelectAllMatchedItems(SelectionEvent event) {
			selectAllMatchedMembers();
		}

		/**
		 * o^{^
		 * @param event Cxg
		 */
		public void doRegister(SelectionEvent event) {
			// _CAO̓o^܂͕ҏW
			if (members.getSelectionCount() > 0) {
				// I𒆂uniqueMember
				UniqueMember um = (UniqueMember) members.getItem(
						members.getSelectionIndex()).getData();

				// ҏWUniqueMember
				UniqueMember reg;
				if (main.members.isRegistered(um)) {
					// o^ς݂̂Ƃ
					// o^ꂽregisteredMember𒼐ڕύX
					reg = main.members.getRegistered(um);
					assert reg != um;
				} else {
					// o^̂Ƃ
					// o^o[clone()ĕύX
					reg = (UniqueMember) um.clone();
					assert reg != um;
				}
				new SWTRegisterDialog(main, shell, main.members, reg);
			} else {
				// VK
				new SWTRegisterDialog(main, shell, main.members, null);
			}
			updateUIStatus(false);
		}

		/**
		 * o^{^
		 * 
		 * @param event Cxg
		 */
		public void doUnregister(SelectionEvent event) {
			int index = members.getSelectionIndex();
			UniqueMember um = (UniqueMember) members.getItem(index).getData();
			// o^
			main.members.unregister(um);
			// {^XV
			updateUIStatus(false);
		}

		/**
		 * M{^
		 * @param event Cxg
		 */
		public void doSend(SelectionEvent event) {
			send();
		}

		/**
		 * SɑM{^
		 * @param event Cxg
		 */
		public void doSendToAll(SelectionEvent event) {
			sendAll();
		}

		/** O񌟍 */
		private long lastSearch;

		/**
		 * Č{^
		 * @param event Cxg
		 */
		public void doSearch(SelectionEvent event) {
			long now = System.currentTimeMillis();
			boolean within5sec = lastSearch + 5000 > now;
			lastSearch = now;
			search(within5sec);
		}

		/**
		 * ǉ{^
		 * @param event Cxg
		 */
		public void doAdditionalSearch(SelectionEvent event) {
			search(true);
		}

		/**
		 * O[vI{^
		 * @param event Cxg
		 */
		public void doSelectGroup(SelectionEvent event) {
			selectSameGroup();
		}

		/**
		 * 
		 * @param event Cxg
		 */
		public void doClose(SelectionEvent event) {
			cancel();
		}

		/**
		 * Rs[
		 * @param event Cxg
		 */
		public void doCopy(SelectionEvent event) {
			if (searchText.isFocusControl()) {
				searchText.copy();
			} else {
				copySelectedMembers();
			}
		}

		/**
		 * ׂđI
		 * @param event Cxg
		 */
		public void doSelectAll(SelectionEvent event) {
			if (searchText.isFocusControl()) {
				searchText.selectAll();
				return;
			}
			members.selectAll();
			updateUIStatus(false);
			members.setFocus();
		}

		/**
		 * 
		 * @param event Cxg
		 */
		public void doFindNext(SelectionEvent event) {
			int count = members.getItemCount();
			if (count == 0) {
				return;
			}
			int index = members.getSelectionIndex();
			if (0 > index || index >= count - 1) {
				index = 0;
			} else {
				index++;
			}
			findItem(searchText.getText(), index, false);
		}

		/**
		 * O
		 * @param event Cxg
		 */
		public void doFindPrevious(SelectionEvent event) {
			int count = members.getItemCount();
			if (count == 0) {
				return;
			}
			int index = members.getSelectionIndex();
			if (0 >= index || index >= count) {
				index = count - 1;
			} else {
				index--;
			}
			findItem(searchText.getText(), index, true);
		}

		/**
		 * [}(migemo)
		 * @param event Cxg
		 */
		public void doCheckMigemo(SelectionEvent event) {
			MenuItem item = (MenuItem) event.widget;
			useMigemo = item.getSelection();
			if (useMigemo) {
				main.attachMigemoDictionaries(false);
			} else {
				main.detachMigemoDictionaries();
			}
		}
	}

	protected void handleEvent(String eventName, Object event) {
		if (eventName.equals(Events.Resumed)
				|| eventName.equals(Events.NetworkChanged)) {
			search(false);
		}
	}
}
