package petaclone.swt;

import java.io.File;
import java.io.IOException;
import java.util.Properties;

import jp.takayuki.logging.Logger;
import jp.takayuki.swt.util.ListenerDelegate;
import jp.takayuki.swt.util.SWTUtils;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.RestartHelper;
import jp.takayuki.tools.Version;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.ShellAdapter;
import org.eclipse.swt.events.ShellEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;

import petaclone.AppPreferences;
import petaclone.EventBroker;
import petaclone.Events;

/**
 * o[WAbv[U[ɒʒmďJn邽߂̃EBhEłB
 * @author Takayuki Shoji
 */
public class SWTUpdateWindow extends ShellAdapter {

	/** logger */
	private static final Logger logger = Logger
			.getLogger(SWTUpdateWindow.class);

	/** bZ[W\[X */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/** CEBhE */
	private final SWTMainWindow main;

	/** AvP[Vݒ */
	private final AppPreferences pref;

	/** EBhE */
	private Shell shell;

	/** eEBhE */
	private final Shell parent;

	/** EventBroker */
	private EventBroker broker;

	/** Cxg */
	private Actions actions = new Actions();

	/** XV */
	private final Version newVersion;

	/**
	 * RXgN^
	 * @param main CEBhE
	 * @param parent eEBhE
	 * @param broker Cxgʒm
	 * @param pref AvP[Vݒ
	 * @param newVersion XV
	 */
	public SWTUpdateWindow(SWTMainWindow main, Shell parent,
			EventBroker broker, AppPreferences pref, Version newVersion) {
		this.main = main;
		this.parent = parent;
		this.broker = broker;
		this.pref = pref;
		this.newVersion = newVersion;
	}

	/**
	 * EBhEJ܂B
	 */
	public void open() {
		this.shell = new Shell(parent, SWT.DIALOG_TRIM | SWT.RESIZE);
		shell.setText(mm.get("SWTUpdateWindow.updateNotice")); //$NON-NLS-1$
		shell.addShellListener(this);

		GridLayout layout = new GridLayout();
		layout.numColumns = 2;

		this.shell.setLayout(layout);

		Label messageLabel = new Label(shell, SWT.NONE);
		messageLabel.setText(mm.get("SWTUpdateWindow.readyToUpdate")); //$NON-NLS-1$
		messageLabel.setLayoutData(createSpan2());

		Text text = new Text(shell, SWT.MULTI | SWT.H_SCROLL | SWT.V_SCROLL
				| SWT.READ_ONLY | SWT.WRAP | SWT.BORDER);

		text.setBackground(Display.getDefault().getSystemColor(
				SWT.COLOR_LIST_BACKGROUND));
		GridData textGd = new GridData();
		textGd.grabExcessVerticalSpace = true;
		textGd.grabExcessHorizontalSpace = true;
		textGd.verticalAlignment = SWT.FILL;
		textGd.horizontalAlignment = SWT.FILL;
		textGd.horizontalSpan = 2;
		text.setLayoutData(textGd);

		Version v = Version.loadFromResource();
		String str = mm.get(
				"CheckForUpdates.msgBody", //$NON-NLS-1$
				new Object[] { newVersion.getVersionText(),
						newVersion.getBuildNumberStr(), v.getVersionText(),
						v.getBuildNumberStr() });
		text.setText(str);

		Label warnLabel = new Label(shell, SWT.WRAP);
		warnLabel.setText(mm.get("SWTUpdateWindow.note")); //$NON-NLS-1$
		warnLabel.setLayoutData(createSpan2());

		Button laterButton = new Button(shell, SWT.PUSH);
		laterButton.setText(mm.get("SWTUpdateWindow.later")); //$NON-NLS-1$
		actions.registerAndAddListener(laterButton, "doButtonLater"); //$NON-NLS-1$

		Button updateButton = new Button(shell, SWT.PUSH);
		updateButton.setText(mm.get("SWTUpdateWindow.installUpdate")); //$NON-NLS-1$
		actions.registerAndAddListener(updateButton, "doButtonUpdate"); //$NON-NLS-1$

		GridData updateButtonGd = new GridData();
		updateButtonGd.horizontalAlignment = SWT.END;
		updateButtonGd.verticalAlignment = SWT.END;
		updateButtonGd.grabExcessHorizontalSpace = true;

		updateButton.setLayoutData(updateButtonGd);

		shell.setDefaultButton(updateButton);

		this.shell.pack();
		this.shell.open();
	}

	/**
	 * span=2GridData𐶐܂B
	 * @return GridData
	 */
	private GridData createSpan2() {
		GridData span2 = new GridData();
		span2.horizontalSpan = 2;
		return span2;
	}

	public void shellClosed(ShellEvent e) {
		if (!SWTUtils.showConfirm(shell, mm.get("SWTUpdateWindow.confirm"), //$NON-NLS-1$
				mm.get("SWTUpdateWindow.closeWithoutInstall"))) { //$NON-NLS-1$
			e.doit = false;
		} else {
			saveUpdateInfo();
		}
	}

	/**
	 * _E[hXVۑ܂B
	 */
	void saveUpdateInfo() {
		File f = newVersion.getFile();
		if (f != null) {
			pref.merge(Version.PREFIX, newVersion.getProperties());
		}
	}

	/**
	 * CxgXi[
	 */
	public class Actions extends ListenerDelegate {

		protected void handleException(Exception e) {
			if (broker == null) {
				logger.error("Error occurred.", e); //$NON-NLS-1$
				return;
			}
			broker.notifyEvent(Events.UnexpectedErrorOccured, e);
		}

		/**
		 * ŃAbvf[g
		 * @param ev
		 */
		public void doButtonLater(SelectionEvent ev) {
			saveUpdateInfo();
			shell.dispose();
		}

		/**
		 * ӂăAbvf[g
		 * @param ev
		 */
		public void doButtonUpdate(SelectionEvent ev) {

			RestartHelper helper = main.createRestartHelper();
			if (helper == null) {
				logger.warn("Unsupported platform."); //$NON-NLS-1$
				return;
			}
			File file = newVersion.getFile();
			if (file == null) {
				logger.warn("Update file is not valid"); //$NON-NLS-1$
				return;
			}

			try {
				helper.update(newVersion.getFile());

				// ʏ̏IV[PXɓ
				main.quit();
			} catch (IOException e) {
				logger.warn("Error occurred while updating application.", e); //$NON-NLS-1$
				// G[ʒm
				SWTUtils.showError(shell,
						mm.get("SWTUpdateWindow.errorOccurredWhileUpdating")); //$NON-NLS-1$
			}
		}
	}

	/**
	 * eXgpCNX
	 * @param args Ȃ
	 */
	public static void main(String[] args) {
		Display display = Display.getDefault();
		Shell parent = new Shell(display);
		parent.setSize(150, 150);
		parent.open();
		Properties prop = new Properties();
		prop.setProperty(Version.BUILD_NUMBER, "18000"); //$NON-NLS-1$
		prop.setProperty(Version.VERSION_STAGE, "alpha"); //$NON-NLS-1$
		prop.setProperty(Version.VERSION_TEXT, "1.8.0 Alpha"); //$NON-NLS-1$

		Version v = new Version(prop);
		SWTUpdateWindow win = new SWTUpdateWindow(null, parent, null, null, v);
		win.open();

		while (!parent.isDisposed()) {
			if (!display.readAndDispatch()) {
				display.sleep();
			}
		}
		display.dispose();
	}

}
