package petaclone.swt.logic;

import java.util.Timer;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.Task;
import jp.takayuki.tools.Tool;
import petaclone.AppPreferences;
import petaclone.CheckForUpdates;
import petaclone.EventBroker;
import petaclone.EventListener;
import petaclone.ServerListener;

/**
 * \tgEFAXV̊mFs܂B
 * 
 * @author Takayuki Shoji
 */
public class SWTCheckForUpdates implements EventListener, Runnable {

	/**
	 * `FbNԊu(20)
	 */
	private static final long INTERVAL = 20 * 60 * 60 * 1000;

	/**
	 * `FbNJnx(7)
	 */
	private static final long DELAY = 7 * 60 * 1000;

	/**
	 * logger
	 */
	private static final Logger logger = Logger
			.getLogger(SWTCheckForUpdates.class);

	/**
	 * AvP[Vݒ
	 */
	private final AppPreferences pref;

	/**
	 * [U[ւ̍XVʒmp
	 */
	private final ServerListener serverListener;

	/**
	 * Cxgʒm
	 */
	private final EventBroker broker;

	/**
	 * xsp^C}[
	 */
	private Timer timer;

	/**
	 * RXgN^
	 * 
	 * @param pref AvP[Vݒ
	 * @param serverListener [U[ւ̍XVʒmp
	 * @param broker Cxgʒm
	 */
	public SWTCheckForUpdates(AppPreferences pref,
			ServerListener serverListener, EventBroker broker) {

		this.pref = pref;
		this.serverListener = serverListener;
		this.broker = broker;
	}

	/**
	 * o[WmF̗\
	 */
	public void scheduleVersionCheck() {
		logger.info("scheduleVersionCheck() start"); //$NON-NLS-1$
		// Java Web Startł̓`FbNȂ
		if (pref.getRuntime().isSandBox()) {
			logger.debug("cancel: sandbox"); //$NON-NLS-1$
			return;
		}

		// C^[lbgڑpolicyŐĂꍇ̓`FbNȂ
		if (!pref.getPolicy().canUseInternet()) {
			logger.debug("cancel: policy"); //$NON-NLS-1$
			return;
		}

		// ݒŃ`FbNȂꍇ
		if (pref.checkForUpdates() == 0) {
			logger.debug("cancel: pref"); //$NON-NLS-1$
			return;
		}

		if (timer != null) {
			logger.debug("cancel: timer != null"); //$NON-NLS-1$
			return;
		}

		// 1̋NƂ1񓮍삳
		long delay;
		if (System.currentTimeMillis() < pref.getLastCheckTime() + INTERVAL) {
			delay = pref.getLastCheckTime() + INTERVAL + DELAY
					- System.currentTimeMillis();
		} else {
			delay = DELAY;
		}

		if (Tool.javaVersion1_5_orHigher) {
			this.timer = new Timer("CheckForUpdates", true); //$NON-NLS-1$
		} else {
			this.timer = new Timer(true);
		}
		Task.schedule(timer, this, delay);
		logger.info("scheduleVersionCheck() CheckForUpdates scheduled: delay=" + delay); //$NON-NLS-1$
	}

	public void run() {
		logger.info("CheckForUpdates thread start"); //$NON-NLS-1$
		CheckForUpdates up = new CheckForUpdates(pref, serverListener, broker);
		up.startThread();
		timer.cancel();
		timer = null;
	}

	public void listen(String eventName, Object event) {
		scheduleVersionCheck();
	}

}
